/*
 *	FarSync SDK for Linux
 *
 *	Simple port monitoring evaluation tool
 *
 *	Copyright (C) 2001-2021 FarSite Communications Ltd.
 *	www.farsite.co.uk
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	File:	farmon.c
 *	Author:	Kevin Curtis	<kevin.curtis@farsite.co.uk>
 */
/*	Revision History
 *	
 *      1.10  30 Apr 2012   Fixed bugs 2318 & 2319 to allow monitoring
 *                          of char interfaces.	
 *      1.11  19 Oct 2012   Buffer first 10ms of data
 *      1.12  11 Dec 2012   Fix missing first frame (2578)
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>

#include "farsync.h"


/* Define FALSE & TRUE 
 */
#ifndef true
# define true   1
# define false  0
typedef int bool;
#endif

#define NUM_INTERFACES	16	/* Max no of interfaces we will support */
/*
 * Change these #defines to suit your interface requirements
 */
#define FST_MAX_MTU     32*1024 	/* Max frame size */
#define FST_LINE_TYPE   X21     	/* E1, V24 or V35 as alternatives */
#define FST_LINE_SPEED  8192000 	/* 8 Meg */

#define BUF_SIZE	FST_MAX_MTU	/* Size of receive buffer - must be able to hold your largest frame */
#define MAX_STORE_SIZE	BUF_SIZE	/* For storing out of sequence frames */

#define RX_OK        0
#define RX_ERROR     1
#define RX_CARRIER   2

/*
 * We're passing data around in global structures for simplicity.
 * OK for simple programs but we should probably be doing a lot more parameter
 * passing in larger programs.
 */
static char *progname;
static unsigned char rxbuf[ BUF_SIZE ];
static unsigned char store[ MAX_STORE_SIZE + sizeof(struct fstioc_mon)];   // add a copy of the header, so we don't have trouble derefencing pointers later on

static unsigned char *store_rd_ptr;
static unsigned char *store_wr_ptr     = store;
static unsigned char *store_start_ptr  = store;
static unsigned char *store_end_ptr    = store + MAX_STORE_SIZE;
static unsigned int   store_count = 0;
static unsigned int   store_size = 0;
static unsigned int   exp_seq_no = 0;

static unsigned int init_seq_no = false;
static int init_timestamp = false;
static struct timeval init_tv;
static struct timeval now_tv;

static int network_needed = false;


/* Structure to contain per interface information */
struct if_info {
        int     index;
	int     sock;		/* Socket handle */
	char   *ifname;		/* Interface name */

} if_info[ NUM_INTERFACES ];

long elapsedMilliSecs(struct timeval *current, struct timeval *last)
{
  long msec;
  
  msec = (current->tv_sec - last->tv_sec) * 1000;    
  msec += (current->tv_usec - last->tv_usec) / 1000;
  return msec;
}

void dump_hex(void* Data, unsigned int NumBytes)
{
  unsigned char* pCharDataH=(unsigned char*)Data;
  unsigned char* pCharDataA=(unsigned char*)Data;
  unsigned int  i,j;
    
  for (i = 0; i < NumBytes; i += 16)
    {
      
      
      printf("%08x: ", 0+i);
      // Output the hex bytes
      for (j = i; j < (i+16); j++)
        {
          if (j < NumBytes)
            { 
	      if(pCharDataH+j == store_end_ptr)
	        pCharDataH = store - j;
	      
	      printf("%02x ",(unsigned int)((unsigned char)*(pCharDataH+j)));
            }
          else
            {
              printf("   ");
            }
          /*fi*/
        }
      /*rof*/  

      printf("  ");

      // Output the ASCII bytes
      for (j = i; j < (i+16); j++)
        {
          if (j < NumBytes)
            {
	      if(pCharDataA+j == store_end_ptr)
	        pCharDataA = store - j;
              
	      char c = *(pCharDataA+j);

              if (c < ' ' || c > 'z')
                {
                  c = '.';
                }
              /*fi*/
              printf("%c", (unsigned int)c);
            }
          else
            {
              printf(" ");
            }
          /*fi*/
        }
      /*rof*/
      printf("\r\n");
    }
  /*rof*/
}

/*
 *      Print an error message and exit
 */
static void
error_exit ( char *msg, ...)
{
  va_list ap;
  
  va_start ( ap, msg );
  fprintf ( stderr,"%s: ", progname );
  vfprintf ( stderr, msg, ap );
  va_end ( ap );
  
  exit ( 1 );
}



/*
 *	Close an interface down
 */
static void
fst_uninit ( struct if_info *iface )
{
  struct fstioc_info info;
  struct ifreq req;
  int monitor_flag = 0;

  /* 
   * It's important to set monitoring off in the driver
   * to avoid performance impact
   */
  memset ( &info, 0, sizeof ( info ));
  strcpy ( req.ifr_name, iface->ifname );
  monitor_flag = 0;
  req.ifr_data = (char *) &monitor_flag;
  if ( ioctl ( iface->sock, FSTSETMON, &req ))
    {
      error_exit ("Problem setting monitor mode %s.\n"
                  "%s\n", iface->ifname, strerror ( errno ));
    }

  /* Bring the interface DOWN if we set it to UP */
  if (network_needed)
  {
    if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
      {
        error_exit ("Error getting interface %s flags.\n%s\n", iface->ifname,
	  	  strerror ( errno ));
      }
    if (( req.ifr_flags & IFF_UP ))
      {
        req.ifr_flags &= !IFF_UP;
        if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	  {
	    error_exit ("Failed to bring interface %s down.\n%s\n", 
		        iface->ifname, strerror ( errno ));
	  }
      }
  }
  if ( close ( iface->sock ) < 0 )
    {
      fprintf ( stderr,"%s: Problem closing %s\n%s\n", progname,
		iface->ifname, strerror ( errno ));
    }
}

/*
 *      Check Signal State
 */
static int
check_state ( int sock, char * name )
{
  /*
   * Select has signalled an exception on the socket
   * Check the signal states
   */
  struct ifreq req;
  struct fstioc_status state;

  strcpy ( req.ifr_name, name );
  req.ifr_data = (char *)&state;
  if ( ioctl ( sock, FSTGSTATE, &req ))
    {
      printf ("Problem getting status for interface"
	      " %s.\n%s\n", name, strerror ( errno ));
    }
  printf("%s: Signal state is %x\n", name, state.carrier_state); 
  printf("%s: Tx Queue Length is %x\n", name, state.txq_length);   
  return state.carrier_state;
}

static void
fst_init ( struct if_info *iface, char *name )
{
  struct ifreq req;
  int monitor_flag = 0;

  
  /* Sanity check and record the interface name */
  if ( strlen ( name ) >= IFNAMSIZ )
    error_exit ("Interface name %s too long\n", name );
  
  iface->ifname = name;
  
  if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
    error_exit ("Unable to create socket. %s\n", strerror ( errno ));
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, name );

  printf("Set monitoring mode\n");      
  monitor_flag = 1;
  req.ifr_data = (char *) &monitor_flag;
  if ( ioctl ( iface->sock, FSTSETMON, &req ))
    {
      error_exit ("Problem setting monitor mode %s.\n"
                  "%s\n", name, strerror ( errno ));
    }

  /* Bring the interface UP if necessary */
  if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
    {
      error_exit ("Error getting interface %s flags.%d\n%s\n", name, iface->sock,
		  strerror ( errno ));
    }
  
  if ( (network_needed == 1) && (! ( req.ifr_flags & IFF_UP )))
    {
      printf("%s: Preparing interface for monitoring\n", name);
      req.ifr_flags |= IFF_UP;
      if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	{
	  error_exit ("Failed to bring interface %s up.\n%s\n", name,
		      strerror ( errno ));
	}
    }
  else
    {
      network_needed = 0;
      printf ("%s: Interface already up\n", name);
    }

  if ( close ( iface->sock ) < 0 )
    {
      error_exit ("Error closing interface %s\n.%s\n", name,
		  strerror ( errno ));
    }
}

/*
 * 	Open and Bind interface in raw sockets mode.
 *
 *	Print an error message and exit on first error (real programs may wish
 *	to do something better here).
 */
static void
fst_open ( struct if_info *iface )
{
  struct ifreq req;
  struct sockaddr_ll sll;
  int ifindex;
  
  /* 
   * Now that the driver is set into "raw" mode open the packet interface
   */
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, iface->ifname );
  
  /* The FarSync driver marks monitor packets with ETH_P_DIAG */
  if (( iface->sock = socket ( PF_PACKET, SOCK_RAW,
			       htons(ETH_P_DIAG) )) < 0 )
    {
      error_exit ("Failed to create raw socket for interface %s.\n"
		  "%s\n", iface->ifname, strerror ( errno ));
    }
  
  /* Bind socket to a single interface. Start by getting the index */
  if ( ioctl ( iface->sock, SIOCGIFINDEX, &req ) < 0 )
    {
      error_exit ("Error getting interface %s index.\n%s\n", 
		  iface->ifname, strerror ( errno ) );
    }
  ifindex = req.ifr_ifindex;
  
  /* Now bind it */
  bzero ( &sll, sizeof ( sll ));
  sll.sll_family = AF_PACKET;
  sll.sll_hatype = ARPHRD_RAWHDLC;
  sll.sll_ifindex = ifindex;
  if ( bind ( iface->sock, (struct sockaddr *)&sll, sizeof ( sll )) < 0 )
    {
      error_exit ("Problem binding to interface %s.\n%s\n", 
		  iface->ifname, strerror ( errno ));
    }
 
}

/*
 *	Read a raw packet from an open socket returning the length of data
 *      received
 */
static int
read_rawhdlc ( struct if_info *iface, void *buf, int len )
{
  int retval;
  int saved_errno;
  
  retval = recvfrom ( iface->sock, buf, len, 0, NULL, 0 );
  
  if ( retval == -1 )
    {
      saved_errno = errno;
      if (saved_errno != EAGAIN)
        {
          fprintf ( stderr,"%s: %s recvfrom error.%d\n%s\n", progname,
                    iface->ifname,  errno, strerror ( errno ));
          
          /* Restore errno destroyed by the printf in case parent wants to
           * check it */
          errno = saved_errno;

        }
    }

  return retval;
}

/*
 *   Update pointer to beginning of the window
 */
static void update_store_start(void)
{
  struct fstioc_mon *header;
  int size;  
  header = (struct fstioc_mon *) store_start_ptr; 


  while ((store_count > 0) && (header->sequence < exp_seq_no))
  {    
    size = sizeof(struct fstioc_mon) + header->length;
    store_count--;
    store_size-= size;
    
    if (store_start_ptr + size >= store + MAX_STORE_SIZE) 
      store_start_ptr = store + (store_start_ptr + size - store_end_ptr);
    else	
      store_start_ptr += size; 
    
    header = (struct fstioc_mon *) store_start_ptr;
  } 
}

/*
 * Search through the store and display all frames we can
 */

static void search_store(void)
{
  struct fstioc_mon *header;
  static unsigned int last_searched;
  int size;
  int i;
  
  while (last_searched != exp_seq_no)
  {
    last_searched = exp_seq_no;
	      
    store_rd_ptr = store_start_ptr;

    /*
     * Search for this frame in the store. If we find it, display it, 
     * and we keep going looking for subsequent frames in the store.
     * If our frame was at the beginning of the store, we can 
     * call update_store_start to remove it 
     */

    for(i=1; i<=store_count; i++) 
    {
      header = (struct fstioc_mon *)store_rd_ptr;

      if (header->sequence == exp_seq_no)  // found it
      {
        exp_seq_no++;
        printf("\n%s %s: Seq: %d: Time %lu: Len %d\n", if_info->ifname, 
               header->tx_rx_ind == FST_MON_TX ? "TX" : "RX",
	       header->sequence, header->timestamp, header->length);	
	dump_hex((char *) header+sizeof(struct fstioc_mon), header->length);

        if(store_rd_ptr == store_start_ptr)
          update_store_start(); 
      
        break;		  
      }
      size = sizeof(struct fstioc_mon) + header->length;
      if(store_rd_ptr + size >= store_end_ptr)
        store_rd_ptr = store + (store_rd_ptr + size - store_end_ptr);
      else
        store_rd_ptr += size;
    }
  }//while
}
/*
 * Find the lowest sequence number that's stored
 */
static void find_lowest(void)
{
  struct fstioc_mon *header;
  int lowest = UINT_MAX;
  int size;
  int i;

  store_rd_ptr = store_start_ptr;
 
  for(i=1; i<=store_count; i++)
  {
    header = (struct fstioc_mon *)store_rd_ptr;

    if ((header->sequence < lowest) && (header->sequence >= exp_seq_no))
      lowest = header->sequence;

    size = sizeof(struct fstioc_mon) + header->length;

    if(store_rd_ptr + size >= store_end_ptr)
      store_rd_ptr = store + (store_rd_ptr + size - store_end_ptr);	      
    else
      store_rd_ptr += size;
  } //for	    
	  
  if(lowest != UINT_MAX)
  {
    if(exp_seq_no == lowest-1)
      fprintf(stderr, "lost frame %d\n", exp_seq_no);
    else if (exp_seq_no != 0)
      fprintf(stderr, "lost frames %d to %d\n", exp_seq_no, lowest-1);    
    exp_seq_no = lowest;
  }
  else
    exp_seq_no++;

  update_store_start();
}

/*
 *	get any monitor frames
 */
static int
get_one_message ( struct if_info *port )
{
  static unsigned char *rd_ptr = rxbuf;
     
  int size;
  int xfer;
  int sizeA;
  int extra;
 
  struct fstioc_mon *header;

 
  /* Read all queued frames */
  while (( xfer = read_rawhdlc ( port, rxbuf, BUF_SIZE )) >= 0)
  {     
      header = (struct fstioc_mon*) &rxbuf;

      if(!init_seq_no)
      {
        if(!init_timestamp)
        {
          gettimeofday(&init_tv, NULL);
          init_timestamp = true;	
        }
        else
        {	
          gettimeofday(&now_tv, NULL);
          if(elapsedMilliSecs(&now_tv, &init_tv) > 10) // 10ms
          {	  
	    init_seq_no = true;
	    find_lowest();
	    search_store();	  
	  }
        }
      } // init_seq_no
      
      if ((init_seq_no) && (header->sequence == exp_seq_no))
      {
	/*
	 *  If the sequence number was as expected, just print it out
	 */	    
        exp_seq_no++;
        printf("\n%s %s: Seq: %d: Time %lu: Len %d\n", if_info->ifname, 
               header->tx_rx_ind == FST_MON_TX ? "TX" : "RX",
	       header->sequence, header->timestamp, header->length);	
	dump_hex((char *) header+sizeof(struct fstioc_mon), header->length);	   
        // To display later frames, if any are stored
        search_store();   
      }
      else if (header->sequence > exp_seq_no)
      {
	/*
	 *   Is the sequence we just read in, now the expected sequence number ?
	 */
	if (header->sequence == exp_seq_no)
	{
          exp_seq_no++;
          printf("\n%s %s: Seq: %d: Time %lu: Len %d\n", if_info->ifname, 
                 header->tx_rx_ind == FST_MON_TX ? "TX" : "RX",
	         header->sequence, header->timestamp, header->length);	
	  dump_hex((char *) header + sizeof(struct fstioc_mon), header->length);
	  // To display later frames, if any are stored
	  search_store();
	}
	else
	{
	  /*
	   *   Add it to the store, when it fits
	   */
	  
	  size = sizeof(struct fstioc_mon) + header->length;
	  
	  while	(store_size + size > MAX_STORE_SIZE)
	  {
	    find_lowest();
	    search_store();
	  }
	  
	  /*
	   *   If store was updated, is the sequence now the expected sequence number ?
	   */	  
	  if (header->sequence == exp_seq_no)
	  {
            exp_seq_no++;
            printf("\n%s %s: Seq: %d: Time %lu: Len %d\n", if_info->ifname, 
                 header->tx_rx_ind == FST_MON_TX ? "TX" : "RX",
	         header->sequence, header->timestamp, header->length);	
	    dump_hex((char *) header + sizeof(struct fstioc_mon), header->length);
	  }		
	
	  sizeA=0;
	  extra=0;
	  
	  // if it doesn't fit at end of buffer, copy it in two parts
	  if (store_wr_ptr + size > store_end_ptr)
	  {  // how much can we write for partA
	
	    sizeA = store_end_ptr - store_wr_ptr;
	    size -= sizeA;
	    if(store_wr_ptr + sizeof(struct fstioc_mon) >= store_end_ptr) // do we need to make a copy of header
	      extra = store_wr_ptr + sizeof(struct fstioc_mon) - store_end_ptr;
	    
	    memcpy(store_wr_ptr, rd_ptr, sizeA+extra);
	    store_wr_ptr=store;
	    store_size+=sizeA;
	  }

	  memcpy(store_wr_ptr, rd_ptr+sizeA, size);
	  store_wr_ptr+=size;
	  store_size+=size;
	  store_count++;		           
        } 
      }  
      else if (header->sequence < exp_seq_no)
      {
	fprintf(stderr, "Unexpected sequence number: %d, expected: %d \n",header->sequence, exp_seq_no);
      }      
    
  }// while read_rawhdlc
  return RX_OK;
}
    
       



static void
test_xfer ( struct if_info *port)
{
  fd_set read_set;
  fd_set excpt_set;

  int retval;
  int flags;
  struct timeval timeout;
  int i =1;
    
  /*
   * Set the sockets into non blocking mode
   */
  flags = fcntl(port->sock, F_GETFL, 0);
  flags |= O_NONBLOCK;
  retval = fcntl(port->sock, F_SETFL, flags);
  flags = fcntl(port->sock, F_GETFL, 0);
  
  while (i)
    {
      FD_ZERO(&read_set);
      FD_ZERO(&excpt_set);
      FD_SET(port->sock, &read_set);
      FD_SET(port->sock, &excpt_set);

      timeout.tv_sec = 0;
      timeout.tv_usec = 10000;  // every 10ms
      retval = select(10, &read_set, NULL, &excpt_set, &timeout);
      if (retval > 0)
      {
        if (FD_ISSET(port->sock, &read_set))
        {
          //printf("Select says readable\n");
          if (get_one_message(port) == RX_CARRIER)
	    break;
        }
        if(FD_ISSET(port->sock, &excpt_set))
        {
          printf("Select says exception\n");
          check_state(port->sock, port->ifname);
        }
      }
      else if (retval == 0)
      {
        if(init_timestamp == true)
        {
          if(!init_seq_no)
          {
            gettimeofday(&now_tv, NULL);
            if(elapsedMilliSecs(&now_tv, &init_tv) > 10) // 10ms
            {	      
	      init_seq_no = true;
	      find_lowest();
	      search_store();	  
            }        
          } // init_seq_no
	}  
      }
      else
      {
         printf("%s: Error in select %d %d\n", 
		     port->ifname, retval ,errno);
      }
    } //while    
}


/*
 *	Print a simple usage message
 */
static void
usage ( void )
{
  fprintf ( stderr,"usage: %s device [char]\n", progname );
  fprintf ( stderr,"e.g. %s sync0\n", progname);
  fprintf ( stderr,"     %s sync0 char\n", progname);
  fprintf ( stderr,"where 'char' is required for monitoring a char interface\n");
}
static void cleanup(void)
{
  printf("Exit handler\n");
  if (if_info[0].ifname != NULL)
    {
      fst_uninit ( &if_info[0]);
    }
}

void signal_terminate(int signo)
{
  /*
   * This will force the exit handler to run
   */
  printf("Signal handler\n");
  sleep(1);
  exit(1);
}


/*
 *	Monitor port 0 and display any frames received 	
 */
int
main ( int argc, char *argv[])
{
  /* Strip and record the program name for later use */
  printf("FarMon Version 1.12 - 11 Dec 2012\n");
  if (( progname = strrchr ( argv[0],'/')) != NULL )
    ++progname;
  else
    progname = argv[0];

  if ((argc < 2) || (argc > 3))
    {
      usage();
      exit(1);
    }
  if (argc == 3)
    {
      if (strncmp("char", argv[2], 4) != 0)
	{
	  usage();
	  exit(1);
	}
      else
	{
	  network_needed = 1;
	}
    }
  /* Initialise and Open the interface */
  fst_init ( &if_info[0], argv[1]);
  if_info[0].index = 0;
  printf("Opening port %s to monitor\n", if_info[0].ifname);
  fst_open( &if_info[0]);
  sleep (1);
  
  /*
   * Set up signal handlers for program termination
   */
  printf("Setting up Signal and Exit handlers\n");
  if (signal(SIGTERM, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  if (signal(SIGINT, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  
  /*
   * Set up exit handlers to clean up
   */
  
  if (atexit(cleanup) != 0)
    {
      printf("Can't register exit handler\n");
      exit (1);
    }
  
  /*
   * Monitor Line zero only for now 
   */
  test_xfer(&if_info[0]); 
  exit ( 0 );
}

