 /*
 *      FarSync Flex driver for Linux (2.6.x kernel syncppp version)
 *
 *      Copyright (C) 2001-2023 FarSite Communications Ltd.
 *      www.farsite.co.uk
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *      Author:      Kevin Curtis  <kevin.curtis@farsite.co.uk>
 */
/*      Revision History
 *      $Id: fsflex-2.6-sync.c 4596 2019-06-04 08:49:07Z kevinc $
 *		
 */

/* 400: PnP fixes for flex removal/reinsertion */
/* 401a: Support for kernels <= 5.19 */
/* 408: gen_mon_packet updates to not be skb-based */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <linux/netdevice.h>
#include <linux/init.h>
#include <linux/if_arp.h>
#include <linux/interrupt.h>
#include <linux/aio.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/usb.h>
#ifdef FST_NET_DEV
#include <net/syncppp.h>
#endif
#include <asm/delay.h>
#include <linux/delay.h>
#include <linux/workqueue.h>
/* TTY includes */
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,8,0)
#include <crypto/hash.h>
#else
#include <linux/crypto.h>
#endif
#include <linux/err.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
#include <linux/scatterlist.h>
#endif
#define FST_BUILD_NO "-b408"
#include "farsync.h"
#include "uss_cmn.h"

/* 400: ioctl changes for >= 5.3 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(5,3,0)
#ifndef HAVE_UNLOCKED_IOCTL
#define HAVE_UNLOCKED_IOCTL	1
#endif
#endif

#ifndef RHEL_RELEASE_VERSION
#define RHEL_RELEASE_VERSION(a,b) 0
#endif

#define FST_PORT_NAME "sync"
#define FST_DRIVER_TYPE "OEM"
//

/*      Number of ports (per card) and cards supported
 */
#define FST_MAX_PORTS           1
#define FST_MAX_CARDS           32
#define FST_MAX_USB_BUFFER      (32*1024)+8

/* Enable this #define if you want an oem driver without hooks into syncppp
 * This means the driver can be loaded without the presence of the
 * syncppp module
 */
//#define FST_NET_DEV             1

/*
 *      Module info
 */
MODULE_AUTHOR("K. P. Curtis <kevin.curtis@farsite.co.uk>");
MODULE_DESCRIPTION("FarSync Flex OEM driver. FarSite Communications Ltd.");

#define FST_ALLOW_IOCTL_GET  1
#define FST_ALLOW_IOCTL_SET  2
#define FST_ALLOW_IOCTL_CARD 4
#define FST_ALLOW_IOCTL_ALL  FST_ALLOW_IOCTL_GET + FST_ALLOW_IOCTL_SET + FST_ALLOW_IOCTL_CARD

/*
 * Modules parameters and associated varaibles
 */
#define FST_HIGH_WATER_MARK     12      /* Point at which we flow control
					 * network layer */
#define FST_LOW_WATER_MARK      8       /* Point at which we remove flow
					 * control from network layer */
#define MAX_URBS_OUTSTANDING    1
#define FST_MIN_DMA_LEN 64              /* do DMA on transfers > 64 bytes*/

#define FIRST_FLEX_V2_MAJOR_ID  0      /* A Major id of >0 means flex V2 */
#define FIRST_FLEX_V2_MINOR_ID  3      /* A Minor id of 3 means flex V2 */

#define FIRST_FLEX_V3_MAJOR_ID  2      /* A Major id of >=2 means flex V3 */

int   fst_txq_low=FST_LOW_WATER_MARK;
int   fst_txq_high=FST_HIGH_WATER_MARK;
int   fst_min_dma_len=FST_MIN_DMA_LEN;
int   fst_dmathr=0xdd00dd00;
int   fst_iocinfo_version=FST_VERSION_CURRENT;
int   fst_max_reads=7;
int   fst_allow_ioctl=0;
int   fst_excluded_cards=0;
int   fst_excluded_list[FST_MAX_CARDS];
int   fst_excluded_list_len = FST_MAX_CARDS;
int   fst_num_serials=0;
char *fst_serials[FST_MAX_CARDS];


module_param (fst_txq_low,int,S_IRUGO);
module_param (fst_txq_high,int,S_IRUGO);
module_param (fst_min_dma_len,int,S_IRUGO);
module_param (fst_dmathr,int,S_IRUGO);
module_param (fst_iocinfo_version,int,S_IRUGO);
module_param (fst_max_reads,int,S_IRUGO);
module_param (fst_allow_ioctl,int,S_IRUGO);
module_param (fst_excluded_cards, int, S_IRUGO);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
module_param_array (fst_excluded_list, int, fst_excluded_list_len, S_IRUGO);
module_param_array (fst_serials, charp, fst_num_serials, S_IRUGO);
#else
module_param_array (fst_excluded_list, int, &fst_excluded_list_len, S_IRUGO);
module_param_array (fst_serials, charp, &fst_num_serials, S_IRUGO);
#endif
MODULE_LICENSE("GPL");

/*      Driver configuration and global parameters
 *      ==========================================
 */

/*      Default parameters for the link
 */
#define FST_TX_QUEUE_LEN        100     /* At 8Mbps a longer queue length is
                                         * useful, the syncppp module forces
                                         * this down assuming a slower line I
                                         * guess.
                                         */
#define FST_TXQ_DEPTH           16      /* This one is for the buffering
                                         * of frames on the way down to the card
                                         * so that we can keep the card busy
                                         * and maximise throughput
					 */
#define FST_MAX_MTU             32*1024 /* Limitation of hdlc controller on
                                         * card */
#define FST_DEF_MTU             1500    /* Common sane value */
#define FST_MAX_ATM_MTU         (32*53) /* 1500 bytes should fit in 32 cells */
#define MAX_PPP_HEADER          10      /* Allowance for PPP overhead when
                                           allocating skbs or checking lengths*/
#define FST_TX_TIMEOUT          (10*HZ)
#define FST_MAX_SEGMENTS        FST_MAX_MTU/128 /* Maximum tx/rx desriptors per frame */

#ifdef ARPHRD_RAWHDLC
#define ARPHRD_MYTYPE   ARPHRD_RAWHDLC  /* Raw frames */
#else
#define ARPHRD_MYTYPE   ARPHRD_HDLC     /* Cisco-HDLC (keepalives etc) */
#endif

/* Mailbox values. (portMailbox) */
#define NOP             0       /* No operation */
#define ACK             1       /* Positive acknowledgement to PC driver */
#define NAK             2       /* Negative acknowledgement to PC driver */
#define STARTPORT       3       /* Start an HDLC port */
#define STOPPORT        4       /* Stop an HDLC port */
#define ABORTTX         5       /* Abort the transmitter for a port */
#define SETV24O         6       /* Set V24 outputs */
#define ASSERTSIGNALS   7
#define RELEASESIGNALS  8
#define SETRXCONFIG     9
#define SETTXCONFIG     10
#define GETSIGNALS      11
#define SETSIGNALS      SETV24O
#define RECONFIG        12
#define SETPRESERVE     13
#define GETPRESERVE     14
#define FLUSHBUFFERS    15
#define SETXON          16
#define SETXOFF         17
#define GETPORTSTATS    18
#define RESETPORTSTATS  19
#define SETLINESPEED    20

#define FST_MEMSIZE 0x100000    /* Size of card memory (1Mb) */

#define REQUIRED_TX_BUFFERS 8
#define REQUIRED_RX_BUFFERS 8
#define REQUIRED_TX_BUFFER_SIZE 8*1024
#define REQUIRED_RX_BUFFER_SIZE 8*1024

/*
 * Rx and Tx status byte bit values
 */

#define RX_FRAM         0x80            /* Rx: framing error */
#define RX_OFLO         0x40            /* Rx: overflow error */
#define RX_CRC          0x20            /* Rx: CRC error */
#define RX_ABRT         0x10            /* Rx: buffer error */
#define TX_URUN         0x40            /* Tx: Tx underrun */
#define TX_ABRT         0x10            /* Tx: Tx abort */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
typedef struct class class_t;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
#define CLASS_DEVICE_CREATE(class, dev, device, fmt, rest) device_create(class, device, dev, NULL, fmt, rest)
#else
#define CLASS_DEVICE_CREATE(class, dev, device, fmt, rest) device_create(class, device, dev, fmt, rest)
#endif
#define class_device_destroy(class, dev) device_destroy(class, dev)
#else
#define CLASS_DEVICE_CREATE(cls, devt, device, fmt, arg...) class_device_create(cls, NULL, devt, device, fmt, ## arg)
#endif
#else /* LINUX 2.6.0 - 2.6.14 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13) /* LINUX 2.6.13 - 2.6.14 */
typedef struct class class_t;
#define CLASS_DEVICE_CREATE class_device_create
#else /* LINUX 2.6.0 - 2.6.12, class_simple */
typedef struct class_simple class_t;
#define CLASS_DEVICE_CREATE class_simple_device_add
#define class_create class_simple_create
#define class_destroy class_simple_destroy
#define class_device_destroy(a, b) class_simple_device_remove(b)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13) */
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15) */

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,34)
#define usb_buffer_alloc usb_alloc_coherent
#define usb_buffer_free usb_free_coherent
#endif

typedef struct fst_fifo {
        u16            fifo_length;
        u16            read_idx;
        u16            write_idx;
        short          overflow_idx;
        u8             data[4];
} FIFO, *PFIFO;

#define INFO_RESERVED 64
/*      Device driver private information
 *      =================================
 */

/*
 * Something to count interrupts
 */
struct fst_ints {
  unsigned long int_186;
  unsigned long int_txdma;
  unsigned long int_rxdma;
  unsigned long int_no_work;
};

struct fst_ints_32 {
  uint32_t int_186;
  uint32_t int_txdma;
  uint32_t int_rxdma;
  uint32_t int_no_work;
};

struct fst_ints fst_int_counter[FST_MAX_CARDS];

struct fst_tx_buckets {
  unsigned long txb0;
  unsigned long txb1;
  unsigned long txb2;
  unsigned long txb4;
  unsigned long txb8;
};

struct fst_rx_buckets {
  unsigned long rxb0;
  unsigned long rxb1;
  unsigned long rxb2;
  unsigned long rxb4;
  unsigned long rxb8;
};

struct fst_tx_buckets tx_buckets[FST_MAX_CARDS];
struct fst_rx_buckets rx_buckets[FST_MAX_CARDS];

/*
 * A structure to queue received frame at the new char interface
 */
struct fst_char_rx_frame {
  struct sk_buff            *frame;
  struct fst_char_rx_frame  *next;
  struct fst_char_rx_frame  *prev;
};

/*
 * A multi-part tx/rx queue structure
 * For long frame support a frame can be upto 32*1024 - 2 bytes in length
 * This would occupy 4 tx/rx desriptors (each of 8k)
 */
struct fst_queue_info {
        int                     count;       /* Total size of frame */
        int                     segment_cnt; /* Number of descriptors 
					      * required/used */
        int                     current_seg; /* Which segment we are currently doing */
        unsigned char           flags;       /* Start and End frame flags */
        unsigned char           error_recovery;
        struct sk_buff         *frame;       /* The complete skb */
};

/*      Per port (line or channel) information
 */
struct fst_port_info {
        void                   *if_ptr; /* Some drivers describe this as a
                                         * general purpose pointer. However if
                                         * using syncPPP it has a very specific
                                         * purpose: must be the first item in
                                         * the structure pointed to by dev->priv
                                         * and must in turn point to the
                                         * associated ppp_device structure.
                                         */
        struct fst_card_info   *card;   /* Card we're associated with */
        int                     index;  /* Port index on the card */
        int                      mode;  /* Are we transparent or not */
        int                     proto;  /* Protocol we are running */
        int                     hwif;   /* Line hardware (lineInterface copy) */
        int                   v24IpSts;
        int                   v24OpSts;
        int                     run;    /* Port is running */
        int             ignoreCarrier;  /* Allow tx regardless of carrier state */
        int             num_tx_buffers; /* No of tx buffers in card window */
        int             num_rx_buffers; /* No of rx buffers in card window */
        int             tx_buffer_size; /* Size of tx buffers in card window */
        int             rx_buffer_size; /* Size of rx buffers in card window */
        int           transmitMSBFirst; /* Transmit MSB First */
        int            receiveMSBFirst; /* Receive MSB First */
        int                     rxpos;  /* Next Rx buffer to use */
        int                     txpos;  /* Next Tx buffer to use */
        int                     txipos; /* Next Tx buffer to check for free */
        int                     start;  /* Indication of start/stop to network */
        int                notify_mode; /* Application has selected the notify event */
        int               monitor_mode; /* Application has selected monitor */
        unsigned int          sequence; /* Monitor sequence no */
        int                  port_mode; /* DSL normal or active */
        struct net_device      *dev;    /* Kernel network device entry */
        struct net_device_stats stats;  /* Standard statistics */
#ifdef FST_NET_DEV
        struct ppp_device       pppdev; /* Link to syncPPP */
#endif
  /*
   * A sixteen entry transmit queue
   */
        int                     txqs;  /* index to get next buffer to tx */
        int                     txqe;  /* index to queue next packet */
        struct fst_queue_info   txq[FST_TXQ_DEPTH]; /* The sent segments info */
        struct fst_queue_info   rxq;   /* The received segments info */
        struct file            *char_file;
        int                     char_opens;
        spinlock_t              rxf_lock;   /* Lock for queue head access */
        struct fst_char_rx_frame   *char_inq;
        struct fst_char_rx_frame   *char_inq_end;
        unsigned char char_inq_max_len;   /* Only allow this many frames */
        unsigned char char_inq_threshold; /* Generate a notify if */
        wait_queue_head_t       pollq;   /* poll() support */
        wait_queue_head_t       readq; /* Blocking read support (char) */
        wait_queue_head_t       writeq; /* Blocking write support (cahr) */
        FLEX_SERCONF            usb_config;
        FLEX_CARD_INFO          usb_card_info;
        int                     minor_dev_no;
        char             lowLatency;
        char        fstioc_info_ver;
        char             readv_mode;
        unsigned char flow_controlled;
        unsigned char exception;
        struct fst_fifo *fifo_rxdata;
        struct fst_fifo fifo_txdata;
        char compat;
        unsigned int   features;
        IREX           scratch;
};


/*      Per card information
 */
struct fst_card_info {
        char          *mem;             /* Card memory mapped to kernel space */
        char          *ctlmem;          /* Control memory for PCI cards */
        unsigned int   phys_mem;        /* Physical memory window address */
        unsigned int   phys_ctlmem;     /* Physical control memory address */
        unsigned int   irq;             /* Interrupt request line number */
        unsigned int   nports;          /* Number of serial ports */
        unsigned int   type;            /* Type index of card */
        unsigned int   state;           /* State of card */
        spinlock_t     card_lock;       /* Lock for SMP access */
        spinlock_t     fifo_lock;
        wait_queue_head_t fifo_waitq;   /* A queue to wait for a fifo response */
        int            fifo_complete;
        unsigned short pci_conf;        /* PCI card config in I/O space */
                                        /* Per port info */
        struct fst_port_info *ports[ FST_MAX_PORTS ];
        struct usb_device    *udev;	/* usb device  */
        struct usb_interface *interface;/* usb interface  */
        unsigned char *bulk_in_buffer;		/* buffer to receive data */
        size_t         bulk_in_size;		/* size of receive buffer */
        unsigned char *bulk_out_buffer;		/* buffer to xmit data */
        size_t         bulk_out_size;		/* size of xmit buffer */
        __u8           bulk_in_endpointAddr;	/* addr of bulk in endpoint */
        __u8           bulk_out_endpointAddr;	/* addr of bulk out endpoint */
        unsigned char *int_in_buffer;		/* buffer to receive data */
        size_t         int_in_size;		/* size of receive buffer */
        int            actual_length;
        __u8           int_in_endpointAddr;	/* addr of bulk in endpoint */
        __u8           int_out_endpointAddr;	/* addr of bulk out endpoint */
        __u8           int_interval;
        __u8           tx_urb_cnt;      /* Outstanding Tx urbs allowed */
        struct semaphore limit_sem;	/* limiting # of writes in progress */
        struct urb    *int_urb;
        struct urb    *bulk_urb;
        struct urb    *tx_bulk_urb;
        char           usb_trace_buffer[128];
        int            usb_trace_ptr;
        int            card_no;         /* Inst of the card on the system */
        int            family;          /* TxP or TxU */
        int            dmarx_in_progress;
        int            dmatx_in_progress;
        unsigned long  int_count;
        unsigned long  int_time_ave;
	void          *rx_dma_handle_host;
	dma_addr_t     rx_dma_handle_card;  
	void          *tx_dma_handle_host;
	dma_addr_t     tx_dma_handle_card;  
        struct sk_buff *dma_skb_rx;
        struct fst_port_info *dma_port_rx;
        struct fst_port_info *dma_port_tx;
        int            dma_len_rx;
        int            dma_len_tx;
        int            dma_txpos;
        int            dma_rxpos;
        int            dma_tx_flags;
        int            last_tx_port;
        int            card_mode; /* What we think the identify mode state is*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
	struct device *sync_cdev;
#else
	struct class_device *sync_cdev;
#endif
};

/*
 *      Debug support
 */
#if FST_DEBUG

int fst_debug_mask = { FST_DEBUG };

/* Most common debug activity is to print something if the corresponding bit
 * is set in the debug mask. Note: this uses a non-ANSI extension in GCC to
 * support variable numbers of macro parameters. The inverted if prevents us
 * eating someone else's else clause.
 */
#define fst_dbg(F,fmt,A...) if ( ! ( fst_debug_mask & (F))) \
                                ; \
                        else \
                                printk ( KERN_DEBUG FST_NAME ": " fmt, ## A )

#else
#define fst_dbg(X...)      /* NOP */
#endif


/*      Printing short cuts
 */
#define printk_err(fmt,A...)    printk ( KERN_ERR     FST_NAME ": " fmt, ## A )
#define printk_warn(fmt,A...)   printk ( KERN_WARNING FST_NAME ": " fmt, ## A )
#define printk_info(fmt,A...)   printk ( KERN_INFO    FST_NAME ": " fmt, ## A )

/*
 *      Device Driver Work Queues
 *
 *      So that we don't spend too much time processing events in the 
 *      Interrupt Service routine, we will declare a work queue per Card 
 *      and make the ISR schedule a task in the queue for later execution.
 */

static void do_bottom_half_tx (struct fst_card_info *card);
static void do_bottom_half_rx (struct fst_card_info *card);
static void fst_process_tx_work_q(unsigned long work_q);
static void fst_process_int_work_q(unsigned long work_q);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static void fst_process_threshold_work_q ( struct work_struct *work);
static void fst_process_port_stats_work_q ( struct work_struct *work);
#else
static void fst_process_threshold_work_q ( void *work);
static void fst_process_port_stats_work_q ( void *work);
#endif
static void fst_process_rx_status ( int rx_status , char *name);
/*
static int gen_mon_packet(struct sk_buff *skb, struct fst_port_info *port,
                          int tx_rx_ind);
*/
static int gen_mon_packet(char *pkt, int pkt_len, struct fst_port_info *port,
                          int tx_rx_ind);                         
static void fst_q_work_item (u64 *queue, int card_index);
static void fst_issue_cmd ( struct fst_port_info *port, unsigned short cmd );
#ifndef FST_TTY_CHAR
static int  fst_get_tx_queue_len (struct fst_port_info *port);
#endif

/* 400: taskletchanges for >= 5.11 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static void fst_process_tx_work(struct tasklet_struct *t);
static void fst_process_int_work(struct tasklet_struct *t);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0)
DECLARE_TASKLET(fst_tx_task, fst_process_tx_work);
DECLARE_TASKLET(fst_int_task, fst_process_int_work);
#else
DECLARE_TASKLET(fst_tx_task, fst_process_tx_work_q, 0);
DECLARE_TASKLET(fst_int_task, fst_process_int_work_q, 0);
#endif

#ifdef FST_TTY_CHAR
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static void fst_process_tty_work (struct work_struct *work);
static void fst_process_async_tty_work (struct work_struct *work);
#else
static void fst_process_tty_work (void *work);
static void fst_process_async_tty_work (void *work);
#endif
#endif

static struct work_struct thresh_work_q;
static struct work_struct port_stats_work_q;

unsigned int          fst_ncards = 0;
struct fst_card_info *fst_cards_list[FST_MAX_CARDS];
struct fst_port_info *fst_ports_list[FST_MAX_CARDS*FST_MAX_PORTS];
struct net_device    *fst_net_list[FST_MAX_CARDS*FST_MAX_PORTS];
spinlock_t            fst_work_q_lock;
u64                   fst_work_txq;
u64                   fst_work_intq;
u64                   fst_threshq;
u64                   fst_pstatsq;
int                   last_segment_cnt =0;

/*
 * We need to send an idle cell when using mpoa
 * encapsulation
 */
char                  atm_idle_cell [53];

static unsigned int fst_minor = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
static struct class *farsync_class;
#endif

static unsigned int fst_major = 0;     /* The char driver major device number.
				        * Setting it to 0 will cause us to use
				        * dynamic allocation during module load
				        */
extern struct file_operations fst_sync_fops; 


static char *type_strings[]  = {
        "no hardware",                  /* Should never be seen */
        "FarSync T2P   ",
        "FarSync T4P   ",
	"FarSync T1U   ",
	"FarSync T2U   ",
	"FarSync T4U   ",
        "FarSync TE1   ",
        "FarSync DSL-S1",
        "FarSync T4E   ",
        "FarSync Flex-1     ",
        "FarSync T4Ue   ",
        "FarSync T2Ue   ",
        "FarSync T4E+   ",
        "FarSync T2U-PMC",
        "FarSync TE1e   ",
        "FarSync T2Ee   ",
        "FarSync T4Ee   ",
        "FarSync Flex-1 (v2)",
        "FarSync Flex-1 (v3)"
};

static char *state_strings[]  = {
        "Uninitialised",                  /* Should never be seen */
        "Reset",
        "Downloading",
	"Starting",
	"Running",
	"Bad Version",
        "Halted",
        "I Failed"
};

static void fst_openport( struct fst_port_info *port);
static void fst_closeport ( struct fst_port_info *port );
static int fst_start_xmit ( struct sk_buff *skb, struct net_device *dev );

#define port_to_dev(P)     (P)->dev
#define dev_to_port(D)     netdev_priv(D)
#define port_to_stats(P,S) (P)->stats.S

/* Sooner or later you can't avoid a forward declaration */
static int fst_ioctl ( struct net_device *dev, struct ifreq *ifr, int cmd );
static void fst_notify_state_change ( struct fst_card_info *card, 
                                      struct fst_port_info *port );
static int post_a_read_bulk_urb( struct fst_card_info *usb_dev);
static int post_a_read_int_urb( struct fst_card_info *usb_dev);
static int flex_get_config ( struct fst_card_info *usb_dev,
                             struct fst_port_info *port,
                             struct fstioc_info *info);
static int flex_set_config ( struct fst_card_info *usb_dev, 
                             struct fstioc_info *info);
static void flex_delete(struct fst_card_info *dev)
{

	usb_put_dev(dev->udev);
	kfree (dev);
}

static void flex_serconf_to_le(FLEX_SERCONF *usb_config);
static void flex_serconf_to_cpu(FLEX_SERCONF *usb_config);
static void flex_irex_to_le(IREX *stats);
static void flex_irex_to_cpu(IREX *stats);
static void flex_fstioc_req_to_le(struct fstioc_req *sys_request);
static void flex_fstioc_req_to_cpu(struct fstioc_req *sys_request);
static void flex_card_info_to_le(FLEX_CARD_INFO *usb_card_info);
static void flex_card_info_to_cpu(FLEX_CARD_INFO *usb_card_info);
static void flex_txrx_preamble_to_cpu(FSCMN_TXRX_PREAMBLE *header);

#define FIFO_ASY_CMD    0
#define FIFO_ASY_TX     1
#define FIFO_ASY_MSG    2
#define FIFO_ASY_RX     3
#define FIFO_ASY_RSP    4
#define FIFO_ASY_RXEVNT 5
#define FIFO_PORT_RX    6
#define FIFO_PORT_TX    7

#ifndef FST_TTY_CHAR
static char *fifo_strings[] = {
  "Command Fifo",
  "Transmit Fifo",
  "Message Fifo",
  "Receive Fifo",
  "Response Fifo",
  "Rx Event Fifo",
  "Port Rx Fifo",
  "Port Tx Fifo"
};
#endif
struct fst_device_stats_32   
{
	uint32_t	rx_packets;	/* total packets received	*/
	uint32_t	tx_packets;	/* total packets transmitted	*/
	uint32_t	rx_bytes;	/* total bytes received 	*/
	uint32_t	tx_bytes;	/* total bytes transmitted	*/
	uint32_t	rx_errors;	/* bad packets received		*/
	uint32_t	tx_errors;	/* packet transmit problems	*/
	uint32_t	rx_dropped;	/* no space in linux buffers	*/
	uint32_t	tx_dropped;	/* no space available in linux	*/
	uint32_t	multicast;	/* multicast packets received	*/
	uint32_t	collisions;

	/* detailed rx_errors: */
	uint32_t	rx_length_errors;
	uint32_t	rx_over_errors;	/* receiver ring buff overflow	*/
	uint32_t	rx_crc_errors;	/* recved pkt with crc error	*/
	uint32_t	rx_frame_errors;/* recv'd frame alignment error */
	uint32_t	rx_fifo_errors;	/* recv'r fifo overrun		*/
	uint32_t	rx_missed_errors;	/* receiver missed packet	*/

	/* detailed tx_errors */
	uint32_t	tx_aborted_errors;
	uint32_t	tx_carrier_errors;
	uint32_t	tx_fifo_errors;
	uint32_t	tx_heartbeat_errors;
	uint32_t	tx_underrun_errors;
	
	/* for cslip etc */
	uint32_t	rx_compressed;
	uint32_t	tx_compressed;
};

struct fstioc_info_32 {
        unsigned int   valid;           /* Bits of structure that are valid */
        unsigned int   nports;          /* Number of serial ports */
        unsigned int   type;            /* Type index of card */
        unsigned int   state;           /* State of card */
        unsigned int   index;           /* Index of port ioctl was issued on */
        unsigned int   smcFirmwareVersion;
        uint32_t       kernelVersion;   /* What Kernel version we are working with */
        unsigned short lineInterface;   /* Physical interface type */
        unsigned char  proto;           /* Line protocol */
        unsigned char  internalClock;   /* 1 => internal clock, 0 => external */
        unsigned int   lineSpeed;       /* Speed in bps */
        unsigned int   estLineSpeed;    /* Estimated speed in bps */
        unsigned int   v24IpSts;        /* V.24 control input status */
        unsigned int   v24OpSts;        /* V.24 control output status */
        unsigned short clockStatus;     /* lsb: 0=> present, 1=> absent */
        unsigned short cableStatus;     /* lsb: 0=> present, 1=> absent */
        unsigned short cardMode;        /* lsb: LED id mode */
        unsigned short debug;           /* Debug flags */
        unsigned char  transparentMode; /* hdlc, transparent, bisync, async */
        unsigned char  invertClock;     /* Invert clock feature for syncing */
        unsigned char  asyncAbility ;   /* The ability to do async */
        unsigned char  synthAbility;    /* The ability to syth a clock */
        unsigned char  extendedClocking;/* New T4e clock modes */
        unsigned char  startingSlot;    /* Time slot to use for start of tx */
        unsigned char  clockSource;     /* External or internal */
        unsigned char  framing;         /* E1, T1 or J1 */
        unsigned char  structure;       /* unframed, double, crc4, f4, f12, */
                                        /* f24 f72 */
        unsigned char  interface;       /* rj48c or bnc */
        unsigned char  coding;          /* hdb3 b8zs */
        unsigned char  lineBuildOut;    /* 0, -7.5, -15, -22 */
        unsigned char  equalizer;       /* short or lon haul settings */
        unsigned char  loopMode;        /* various loopbacks */
        unsigned char  range;           /* cable lengths */
        unsigned char  txBufferMode;    /* tx elastic buffer depth */
        unsigned char  rxBufferMode;    /* rx elastic buffer depth */
        unsigned char  losThreshold;    /* Attenuation on LOS signal */
        unsigned char  idleCode;        /* Value to send as idle timeslot */
        unsigned int   receiveBufferDelay; /* delay thro rx buffer timeslots */
        unsigned int   framingErrorCount; /* framing errors */
        unsigned int   codeViolationCount; /* code violations */
        unsigned int   crcErrorCount;   /* CRC errors */
        int            lineAttenuation; /* in dB*/
        unsigned short lossOfSignal;
        unsigned short receiveRemoteAlarm;
        unsigned short alarmIndicationSignal;
        unsigned short _reserved[64];
        unsigned char  ignoreCarrier;  /* If set transmit regardless of carrier state */
        unsigned char  numTxBuffers;   /* No of tx buffers in card window */
        unsigned char  numRxBuffers;   /* No of rx buffers in card window */
        unsigned int   txBufferSize;   /* Size of tx buffers in card window */
        unsigned int   rxBufferSize;   /* Size of rx buffers in card window */
        unsigned char  terminalType;  /* Additional hdsl */
        unsigned char  annexType;
        unsigned char  encap;
        unsigned char  testMode;
        unsigned char  backoff;
        unsigned char  bLineProbingEnable;
        unsigned char  snrth;
        unsigned char  lpath;
        unsigned short vpi;
        unsigned short vci;
        unsigned char  activationStatus;
        unsigned char  noCommonModeStatus;
        unsigned char  transceiverStatus1;
        unsigned char  transceiverStatus2;
        unsigned char  lineLoss;
        char           signalQuality;
        unsigned char  nearEndBlockErrorCount;
        char           signalToNoiseRatio;
        unsigned char  erroredSecondCount;
        unsigned char  severelyErroredSecondCount;
        unsigned char  lossOfSyncWordSecondCount;
        unsigned char  unavailableSecondCount;
        char           frequencyDeviation;
        char           negotiatedPowerBackOff;
        unsigned char  negotiatedPSD;
        unsigned char  negotiatedBChannels;
        unsigned char  negotiatedZBits;
        unsigned short negotiatedSyncWord;
        unsigned char  negotiatedStuffBits;
        unsigned char  chipVersion;
        unsigned char  firmwareVersion;
        unsigned char  romVersion;
        unsigned short atmTxCellCount;
        unsigned short atmRxCellCount;
        unsigned short atmHecErrorCount;
        unsigned int   atmCellsDropped;
        unsigned char  transmitMSBFirst;
        unsigned char  receiveMSBFirst;
        unsigned char  xpldVersion;
        unsigned char  farEndCountryCode[2]; 
        unsigned char  farEndProviderCode[4];
        unsigned char  farEndVendorInfo[2];
        unsigned char  utopiaAtmStatus;
        unsigned int   termination;
        unsigned int   txRxStart;
        unsigned char  enableNRZIClocking;   /* Ver 1 addition */
        unsigned char  lowLatency;           /* Ver 1 addition */
        struct fst_device_stats_32 stats;    /* Ver 1 addition */
        struct fstioc_async_conf async_conf; /* Ver 2 addition */
        unsigned char  iocinfo_version;      /* Ver 2 addition */
        unsigned char  extSyncClockEnable;   /* Ver 3 addition */
        unsigned char  extSyncClockOffset;   /* Ver 3 addition */
        unsigned long  extSyncClockRate;     /* Ver 3 addition */
        unsigned char  ppsEnable;            /* Ver 3 addition */  
        unsigned char  ppsOffset;            /* Ver 3 addition */
        unsigned char  cardRevMajor;         /* Ver 4 addition */
        unsigned char  cardRevMinor;         /* Ver 4 addition */
        unsigned char  cardRevBuild;         /* Ver 4 addition */
        unsigned int   features;             /* Ver 4 addition */
        unsigned int   oneBitInsert;         /* Ver 5 addition */
};

struct fstioc_status_32 {
  int carrier_state;
  int txq_length;
  int rxq_length;
  struct fst_device_stats_32 stats;
};

struct fstioc_cmd_32 {
  unsigned int version;
  unsigned int command;
  unsigned int status;
  unsigned int input_data_len;
  unsigned int output_data_len;
  uint32_t     data_ptr;      
};

#define FST_RX_DATA_FIFO_LEN 4096
#define FST_RX_DATA_FIFO_HIGH 3*1024
#define FST_RX_DATA_FIFO_LOW 1*1024

#ifdef FST_TTY_CHAR
#define FST_MAX_FARSYNC_PORTS   4
#define FST_L1_DEV_NAME         "hdlc"

int find_last_hdlc_device (void)
{
  char devname[16];
  char devnum[16];
  struct net_device *dev;
  int i;

  /*
   * Ask the system how many hdlcX cards there are ...
   */

  for (i=0; i< FST_MAX_CARDS*FST_MAX_FARSYNC_PORTS; i++)
    {
      strcpy(devname, FST_L1_DEV_NAME);
      sprintf(devnum,"%d", i);
      strcat(devname, devnum);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
      if ((dev = dev_get_by_name(devname)) == NULL)
#else
      if ((dev = dev_get_by_name(&init_net, devname)) == NULL)
#endif
	{
	  fst_dbg(DBG_TTY, "Missing device at %d\n", i);
	  break; /* End loop early */
	}
      else
	{
	  dev_put(dev);
	}
    }
  fst_dbg(DBG_TTY, "There were %d farsync tty ports, starting flex tty ports at %d\n", i, i);
  return i;
}

#endif

#ifndef FST_TTY_CHAR
static int check_fifo_space( char *actual_fifo)
{
  int h_size = 0;
  struct fst_fifo fifo;
  u16 avail_fifo_space = 0;
  u16 used_space = 0;
  u16 free_space = 0;
  u16 next_write_idx = 0;

  h_size = sizeof(struct fst_fifo) - 4;
  fst_dbg(DBG_FIFO, "check_fifo_space\n");
  /*
   * Make a local copy of the fifo while we update it 
   */
  memcpy_fromio ( &fifo, actual_fifo, h_size);

  next_write_idx = (u16) ((fifo.write_idx+1) % fifo.fifo_length);
  fst_dbg(DBG_FIFO, "next_write_idx = %d\n", next_write_idx);
  free_space = (u16) ((fifo.read_idx - next_write_idx + 
                                  fifo.fifo_length) % fifo.fifo_length);
  used_space = (u16) (fifo.fifo_length - free_space - 1);
  avail_fifo_space = fifo.fifo_length - used_space;
  fst_dbg(DBG_FIFO, "fifo size = %d, free_space = %d, used_space = %d, avail_fifo_space = %d\n", 
	 fifo.fifo_length, free_space, used_space, avail_fifo_space);
  avail_fifo_space = (u16) min(avail_fifo_space, 
                               (u16)fifo.fifo_length);
  return used_space;
}

static int check_rx_fifo_threshold( struct fst_port_info *port)
{
  u16 used_space =0 ;
  int outputs;
  /*
   * flow control has been enabled, so take some action if the thresholds have
   * been crossed.
   */

  used_space = (u16) check_fifo_space((char *) port->fifo_rxdata);
  if ((used_space > (u16) FST_RX_DATA_FIFO_HIGH) && (!port->flow_controlled))
    {
      fst_dbg(DBG_ASY, "%s: Asserting flow control %d %x\n", port->dev->name, 
	     used_space, port->v24OpSts);
      port->flow_controlled++;
      if (port->usb_config.flowControl)
	{
	  outputs = (port->v24OpSts) & (~OPSTS_RTS);
	  port->v24OpSts = outputs;
	  if ( port->run )
	    {
	      /*
	       * We are probably in interrupt context so must defer any signal
	       * changes to a bh
	       */
	      fst_q_work_item(&fst_threshq, port->card->card_no);
	      schedule_work(&thresh_work_q); 
	    }
	}
      return 0;
    }
  if (used_space < (u16) FST_RX_DATA_FIFO_LOW)
    {
      if (port->flow_controlled)
	{
	  fst_dbg(DBG_ASY, "%s: Releasing flow control %d %x \n", port->dev->name, 
		 used_space, port->v24OpSts);
	  if (port->usb_config.flowControl)
	    {
	      outputs = OPSTS_RTS | port->v24OpSts;
	      port->v24OpSts = outputs;
	      if ( port->run )
		{
		  /*
		   * We are probably in interrupt context so must defer any signal
		   * changes to a bh
		   */
		  fst_q_work_item(&fst_threshq, port->card->card_no);
		  schedule_work(&thresh_work_q); 
		}
	    }
	  port->flow_controlled = 0;
	}
    }
    return 0;
}

static int write_into_fifo( char *actual_fifo, int fifo_select, 
                            char *request, int msg_len)
{
  int retval = -1;
  int h_size = 0;
  struct fst_fifo fifo;
  u16 avail_fifo_space = 0;
  u16 used_space = 0;
  u16 free_space = 0;
  u16 next_write_idx = 0;
  u16 first_write_amount = 0;
  u16 second_write_amount = 0;

  h_size = sizeof(struct fst_fifo) - 4;
  fst_dbg(DBG_FIFO, "write_into_fifo [%s]  a message of %d bytes\n", 
          fifo_strings[fifo_select], msg_len);
  /*
   * Make a local copy of the fifo while we update it 
   */
  memcpy ( &fifo, actual_fifo, h_size);

  next_write_idx = (u16) ((fifo.write_idx+1) % fifo.fifo_length);
  fst_dbg(DBG_FIFO, "next_write_idx = %d\n", next_write_idx);
  free_space = (u16) ((fifo.read_idx - next_write_idx + 
                                  fifo.fifo_length) % fifo.fifo_length);
  used_space = (u16) (fifo.fifo_length - free_space - 1);
  avail_fifo_space = fifo.fifo_length - used_space;
  fst_dbg(DBG_FIFO, "fifo size = %d, free_space = %d, used_space = %d, avail_fifo_space = %d\n", 
	 fifo.fifo_length, free_space, used_space, avail_fifo_space);
  avail_fifo_space = (u16) min(avail_fifo_space, 
                               (u16)fifo.fifo_length);

  if (fifo.overflow_idx == -1)
    {
      fst_dbg(DBG_FIFO, "msg_fifo.overflow_idx = -1\n");
      if (avail_fifo_space >= msg_len)
        {
	  fst_dbg(DBG_FIFO, "And there is space to write data\n");
          /* How much of it can we write into the end fragment */
          first_write_amount = min(msg_len, 
                                   (fifo.fifo_length - next_write_idx));
	  fst_dbg(DBG_FIFO, "First write amount is %d\n", first_write_amount);
	  fst_dbg(DBG_FIFO, "Next write index is %d\n", next_write_idx);
	  fst_dbg(DBG_FIFO, "Copying to offset %p\n", actual_fifo+h_size+next_write_idx);
	  if (first_write_amount > fifo.fifo_length)
	    {
	      printk("Fifo length error %d %d\n", first_write_amount, fifo.fifo_length);
	      return retval;
	    }
          memcpy(actual_fifo+h_size+next_write_idx, request, first_write_amount);
          fifo.write_idx = (u16) ((fifo.write_idx + first_write_amount)
                                  % fifo.fifo_length);
	  fst_dbg(DBG_FIFO, "Write Index updated to %d\n", fifo.write_idx);
          if (first_write_amount != msg_len)
            {
	      fst_dbg(DBG_FIFO, "We need a second write chunk because of wrap\n");
              /* were do we start writing the next chunk to */
              next_write_idx = (u16) ((fifo.write_idx + 1) % 
                                      fifo.fifo_length);
              
              /* write the rest into the starting fragment */
              second_write_amount = (u16) (msg_len - 
                                           first_write_amount);
	      if (second_write_amount > fifo.fifo_length)
		{
		  printk("Fifo length error on 2nd write %d %d\n", 
			 second_write_amount, fifo.fifo_length);
		  return retval;
		}
	      fst_dbg(DBG_FIFO, "Second write amount is %d\n", second_write_amount);
	      fst_dbg(DBG_FIFO, "Next write index is %d\n", next_write_idx);
	      fst_dbg(DBG_FIFO, "Copying to offset %p\n", actual_fifo+h_size+next_write_idx);
              memcpy(actual_fifo+h_size+next_write_idx, 
		     request+first_write_amount, 
		     second_write_amount);
              fifo.write_idx = (u16) ((fifo.write_idx + 
                                       second_write_amount)
                                      % fifo.fifo_length);
            }
          retval = 0;
        }
      else
	{
	  fst_dbg(DBG_ASS, "Not enough room in fifo (%d) for %d bytes of data\n",
		 avail_fifo_space, msg_len);
	}
    }
  else
    {
      printk("write_into_fifo: [%s] fifo overflow\n", 
             fifo_strings[fifo_select]);
      fifo.overflow_idx = fifo.write_idx;
    }

  fst_dbg(DBG_FIFO, "Updating the fifo header\n");
  /*
   * Write the local copy back (of write_idx and overflow idx)
   * to the fifo
   */
  memcpy( actual_fifo+offsetof(struct fst_fifo, write_idx), 
	  &fifo.write_idx, 
	  sizeof(fifo.write_idx)+sizeof(fifo.overflow_idx));
  return retval;
}

static int read_from_fifo( char * actual_fifo, int fifo_select, 
                           char *data, int msg_len)
{
  int h_size = 0;
  struct fst_fifo fifo;
  u16   next_read_idx = 0;
  u16   amount_to_read = 0;
  char *data1 = NULL;
  u16   length1;
  char *data2 = NULL;
  u16   length2;
  u16   data_len;

  fst_dbg(DBG_FIFO, "read_from_fifo [%s] into buffer of size %d\n", 
          fifo_strings[fifo_select], msg_len);
  h_size = sizeof(struct fst_fifo) - 4;
  /*
   * Get the fifo header
   */
  memcpy(&fifo, actual_fifo, h_size);
  if (fifo.fifo_length > (FST_RX_DATA_FIFO_LEN+1))
    {
      printk("fifo size error! %d\n", fifo.fifo_length);
      return 0;
    }
  /*
   * Work out the length of data in the fifo
   */
  data_len = (fifo.write_idx - fifo.read_idx + fifo.fifo_length) 
              % fifo.fifo_length;
  fst_dbg(DBG_FIFO, "Actual amount of data in the fifo is %d\n", data_len);
  /*
   * But we will only take the smaller of the length of data
   * in the fifo or the length of the buffer we have been given
   * to copy the data into
   */
  amount_to_read = (u16) min(data_len, (u16) msg_len);
  fst_dbg(DBG_FIFO, "amount to read is %d\n", amount_to_read);
  if (amount_to_read > 0)
    {
      /* where should we start reading from ? */
      next_read_idx = (fifo.read_idx + 1) % fifo.fifo_length;
      fst_dbg(DBG_FIFO, "Next read index = %d\n", next_read_idx);
      if (next_read_idx > (FST_RX_DATA_FIFO_LEN+1))
	{
	  printk("read index size error!\n");
	  return 0;
	}
      data1 = &actual_fifo[next_read_idx+h_size];
      length1 = (u16) min(amount_to_read, 
                          (u16) (fifo.fifo_length - next_read_idx));
      fst_dbg(DBG_FIFO, "length1 = %d\n", length1);
      if (amount_to_read > (FST_RX_DATA_FIFO_LEN+1))
	{
	  printk("amount to read size error!\n");
	  return 0;
	}
      data2 = actual_fifo+h_size;
      length2 = amount_to_read - length1;
      fst_dbg(DBG_FIFO, "length2 = %d\n", length2);
      fst_dbg(DBG_FIFO, "Copying 1st blob from offset %p of length %d\n", data1, length1);
      memcpy(data, data1, length1);

      if (length2)
        {
	  fst_dbg(DBG_ASS, "Copying from offset %p\n", data2);
          memcpy(data+length1, data2, length2);
        }
      fifo.read_idx = (fifo.read_idx + amount_to_read) % 
                       fifo.fifo_length;
      /*
       * Write the fifo header back as we have taken some data
       */
      memcpy ( actual_fifo+offsetof(struct fst_fifo, read_idx), 
		    &fifo.read_idx, sizeof(fifo.read_idx));
      fst_dbg(DBG_FIFO, "read_from_fifo: Amount read was %d\n", length1+length2);
      return amount_to_read;
    }
  return 0;
}
#endif

static void fst_reset_rx_fifo( struct fst_port_info *port)
{
  int used_space;

  used_space = ((port->fifo_rxdata->write_idx - 
		 port->fifo_rxdata->read_idx));
  if (used_space < 0)
    {
      fst_dbg(DBG_FIFO, "adjusting used_space %d\n", used_space);
      used_space = used_space + FST_RX_DATA_FIFO_LEN;
    }
  fst_dbg(DBG_FIFO, "%s: Resetting Rx Fifo %d %d \n", port->dev->name,
	 port->fifo_rxdata->write_idx, port->fifo_rxdata->read_idx);
  fst_dbg(DBG_FIFO, "%s: %d bytes left in fifo\n", port->dev->name, used_space);
  port->fifo_rxdata->fifo_length  = (u16)(FST_RX_DATA_FIFO_LEN+1); 
  port->fifo_rxdata->read_idx = (u16)(port->fifo_rxdata->fifo_length-1);
  port->fifo_rxdata->write_idx = (u16)(port->fifo_rxdata->fifo_length-1);
  port->fifo_rxdata->overflow_idx = -1;
}

static int fst_alloc_rx_fifo( struct fst_port_info *port)
{

  /*
   * Allocate the data for the fifo
   */
  port->fifo_rxdata = kmalloc(FST_RX_DATA_FIFO_LEN 
			      + sizeof(struct fst_fifo), GFP_ATOMIC);
  if (port->fifo_rxdata == NULL)
    {
      printk("fifo_init: can't allocate %u bytes for Fifo\n", 
	     (unsigned int)(FST_RX_DATA_FIFO_LEN+sizeof(struct fst_fifo)));
      return 1;
    }
  fst_dbg(DBG_ASY, "%s: Fifo memory allocated at %p\n", port->dev->name,
	 port->fifo_rxdata);
  /*
   * Now initialise it
   */
  /* must setup the fifo's length */
  port->fifo_rxdata->fifo_length  = (u16)(FST_RX_DATA_FIFO_LEN+1); 
  port->fifo_rxdata->read_idx = (u16)(port->fifo_rxdata->fifo_length-1);
  port->fifo_rxdata->write_idx = (u16)(port->fifo_rxdata->fifo_length-1);
  port->fifo_rxdata->overflow_idx = -1;
  port->fifo_rxdata->data[0] = 'F';
  port->fifo_rxdata->data[1] = '1';
  port->fifo_rxdata->data[2] = 'D';
  port->fifo_rxdata->data[3] = '0';
  return 0;
}
struct net_device *get_net_device( int if_index )
{
        struct net_device *dev;

	dev = fst_ports_list[if_index]->dev;
	fst_dbg(DBG_ASS, "get_net_device: returning %s\n", dev->name);
	return dev;
}

/* Sooner or later you can't avoid a forward declaration */
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
static int fst_ifs_ioctl ( struct net_device *dev, struct if_settings *ifs );
#endif
static int fst_ioctl ( struct net_device *dev, struct ifreq *ifr, int cmd );

static const speed_t baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 230400, 460800,
	500000, 576000, 921600, 1000000, 1152000, 1500000, 2000000,
	2500000, 3000000, 3500000, 4000000
};
static int n_baud_table = ARRAY_SIZE(baud_table);

#ifdef HAVE_COMPAT_IOCTL
void deviceStatsTo32(struct fst_device_stats_32 *info32, struct fst_device_stats *info64 ) {

	info32->rx_packets		= info64->rx_packets;
	info32->tx_packets		= info64->tx_packets;
	info32->rx_bytes		= info64->rx_bytes;
	info32->tx_bytes		= info64->tx_bytes;
	info32->rx_errors		= info64->rx_errors;
	info32->tx_errors		= info64->tx_errors;
	info32->rx_dropped		= info64->rx_dropped;
	info32->tx_dropped		= info64->tx_dropped;
	info32->multicast		= info64->multicast;
	info32->collisions		= info64->collisions;

	info32->rx_length_errors	= info64->rx_length_errors;
	info32->rx_over_errors		= info64->rx_over_errors;
	info32->rx_crc_errors		= info64->rx_crc_errors;
	info32->rx_frame_errors		= info64->rx_frame_errors;
	info32->rx_fifo_errors		= info64->rx_fifo_errors;
	info32->rx_missed_errors	= info64->rx_missed_errors;
	
	info32->tx_aborted_errors	= info64->tx_aborted_errors;
	info32->tx_carrier_errors	= info64->tx_carrier_errors;
	info32->tx_fifo_errors		= info64->tx_fifo_errors;
	info32->tx_heartbeat_errors	= info64->tx_heartbeat_errors;
	info32->tx_underrun_errors	= info64->tx_underrun_errors;
	
	info32->rx_compressed		= info64->rx_compressed;
	info32->tx_compressed		= info64->tx_compressed;
}

void convertTo64(struct fstioc_info *info64, struct fstioc_info_32 *info32) {
	int i; 

	info64->valid 			= info32->valid;
	info64->nports			= info32->nports;          
	info64->type			= info32->type;
	info64->state			= info32->state;
	info64->index			= info32->index;
	info64->smcFirmwareVersion	= info32->smcFirmwareVersion;
	info64->kernelVersion		= info32->kernelVersion;
	info64->lineInterface		= info32->lineInterface;
	info64->proto			= info32->proto;
	info64->internalClock		= info32->internalClock;
	info64->lineSpeed		= info32->lineSpeed;
	info64->estLineSpeed		= info32->estLineSpeed;
	info64->v24IpSts		= info32->v24IpSts;
	info64->v24OpSts		= info32->v24OpSts;
	info64->clockStatus		= info32->clockStatus;
	info64->cableStatus		= info32->cableStatus;
	info64->cardMode		= info32->cardMode;
	info64->debug			= info32->debug;
	info64->transparentMode		= info32->transparentMode;
	info64->invertClock		= info32->invertClock;
	info64->asyncAbility		= info32->asyncAbility;
	info64->synthAbility		= info32->synthAbility;
	info64->extendedClocking	= info32->extendedClocking;
	info64->startingSlot		= info32->startingSlot;
	info64->clockSource		= info32->clockSource;
	info64->framing			= info32->framing;
	info64->structure		= info32->structure;
                         
	info64->interface		= info32->interface;
	info64->coding			= info32->coding;
	info64->lineBuildOut		= info32->lineBuildOut;
	info64->equalizer		= info32->equalizer;
	info64->loopMode		= info32->loopMode;
	info64->range			= info32->range;
	info64->txBufferMode		= info32->txBufferMode;
	info64->rxBufferMode		= info32->rxBufferMode;
	info64->losThreshold		= info32->losThreshold;
	info64->idleCode		= info32->idleCode;
	info64->receiveBufferDelay	= info32->receiveBufferDelay;
	info64->framingErrorCount	= info32->framingErrorCount;
	info64->codeViolationCount	= info32->codeViolationCount;
	info64->crcErrorCount		= info32->crcErrorCount;
	info64->lineAttenuation		= info32->lineAttenuation;
	info64->lossOfSignal		= info32->lossOfSignal;
	info64->receiveRemoteAlarm	= info32->receiveRemoteAlarm;
	info64->alarmIndicationSignal	= info32->alarmIndicationSignal;

	for (i=0; i<INFO_RESERVED; i++)
		info64->_reserved[i] = info32->_reserved[i];

	info64->ignoreCarrier		= info32->ignoreCarrier;
	info64->numTxBuffers		= info32->numTxBuffers;
	info64->numRxBuffers		= info32->numRxBuffers;
	info64->txBufferSize		= info32->txBufferSize;
	info64->rxBufferSize		= info32->rxBufferSize;
	info64->terminalType		= info32->terminalType;
	info64->annexType		= info32->annexType;
	info64->encap			= info32->encap;
	info64->testMode		= info32->testMode;
	info64->backoff			= info32->backoff;
	info64->bLineProbingEnable	= info32->bLineProbingEnable;
	info64->snrth			= info32->snrth;
	info64->lpath			= info32->lpath;
	info64->vpi			= info32->vpi;
	info64->vci			= info32->vci;
	info64->activationStatus	= info32->activationStatus;
	info64->noCommonModeStatus	= info32->noCommonModeStatus;
	info64->transceiverStatus1	= info32->transceiverStatus1;
	info64->transceiverStatus2	= info32->transceiverStatus2;
	info64->lineLoss		= info32->lineLoss;
	info64->signalQuality		= info32->signalQuality;
	info64->nearEndBlockErrorCount	= info32->nearEndBlockErrorCount;
	info64->signalToNoiseRatio	= info32->signalToNoiseRatio;
	info64->erroredSecondCount	= info32->erroredSecondCount;
	info64->severelyErroredSecondCount	= info32->severelyErroredSecondCount;
	info64->lossOfSyncWordSecondCount	= info32->lossOfSyncWordSecondCount;
	info64->unavailableSecondCount	= info32->unavailableSecondCount;
	info64->frequencyDeviation	= info32->frequencyDeviation;
	info64->negotiatedPowerBackOff	= info32->negotiatedPowerBackOff;
	info64->negotiatedPSD		= info32->negotiatedPSD;
	info64->negotiatedBChannels	= info32->negotiatedBChannels;
	info64->negotiatedZBits		= info32->negotiatedZBits;
	info64->negotiatedSyncWord	= info32->negotiatedSyncWord;
	info64->negotiatedStuffBits	= info32->negotiatedStuffBits;
	info64->chipVersion		= info32->chipVersion;
	info64->firmwareVersion		= info32->firmwareVersion;
	info64->romVersion		= info32->romVersion;
	info64->atmTxCellCount		= info32->atmTxCellCount;
	info64->atmRxCellCount		= info32->atmRxCellCount;
	info64->atmHecErrorCount	= info32->atmHecErrorCount;
	info64->atmCellsDropped		= info32->atmCellsDropped;
	info64->transmitMSBFirst	= info32->transmitMSBFirst;
	info64->receiveMSBFirst		= info32->receiveMSBFirst;
	info64->xpldVersion		= info32->xpldVersion;
	info64->farEndCountryCode[0]	= info32->farEndCountryCode[0];  
	info64->farEndCountryCode[1]	= info32->farEndCountryCode[1];  

	info64->farEndProviderCode[0]	= info32->farEndProviderCode[0]; 
	info64->farEndProviderCode[1]	= info32->farEndProviderCode[1];
	info64->farEndProviderCode[2]	= info32->farEndProviderCode[2];
	info64->farEndProviderCode[3]	= info32->farEndProviderCode[3];

	info64->farEndVendorInfo[0]	= info32->farEndVendorInfo[0];  
	info64->farEndVendorInfo[1]	= info32->farEndVendorInfo[1];
	info64->utopiaAtmStatus		= info32->utopiaAtmStatus;
	info64->termination		= info32->termination;
	info64->txRxStart		= info32->txRxStart;

	info64->enableNRZIClocking	= info32->enableNRZIClocking;   
	info64->lowLatency		= info32->lowLatency;		

	info64->stats.rx_packets	= info32->stats.rx_packets;			
	info64->stats.tx_packets	= info32->stats.tx_packets;		
	info64->stats.rx_bytes		= info32->stats.rx_bytes;		
	info64->stats.tx_bytes		= info32->stats.tx_bytes;		
	info64->stats.rx_errors		= info32->stats.rx_errors;		
	info64->stats.tx_errors		= info32->stats.tx_errors;		
	info64->stats.rx_dropped	= info32->stats.rx_dropped;
	info64->stats.tx_dropped	= info32->stats.tx_dropped;		
	info64->stats.multicast		= info32->stats.multicast;		
	info64->stats.collisions	= info32->stats.collisions;

	info64->stats.rx_length_errors	= info32->stats.rx_length_errors;
	info64->stats.rx_over_errors	= info32->stats.rx_over_errors;		
	info64->stats.rx_crc_errors	= info32->stats.rx_crc_errors;		
	info64->stats.rx_frame_errors	= info32->stats.rx_frame_errors;
	info64->stats.rx_fifo_errors	= info32->stats.rx_fifo_errors;
	info64->stats.rx_missed_errors	= info32->stats.rx_missed_errors;	
	
	info64->stats.tx_aborted_errors	= info32->stats.tx_aborted_errors;
	info64->stats.tx_carrier_errors	= info32->stats.tx_carrier_errors;
	info64->stats.tx_fifo_errors	= info32->stats.tx_fifo_errors;
	info64->stats.tx_heartbeat_errors	= info32->stats.tx_heartbeat_errors;
	info64->stats.tx_underrun_errors	= info32->stats.tx_underrun_errors;
	
	info64->stats.rx_compressed	= info32->stats.rx_compressed;
	info64->stats.tx_compressed	= info32->stats.tx_compressed;

	info64->async_conf		= info32->async_conf;   // copy whole structure in one go
	info64->iocinfo_version		= info32->iocinfo_version;      

	info64->extSyncClockEnable	= info32->extSyncClockEnable;   
	info64->extSyncClockOffset	= info32->extSyncClockOffset;
	info64->extSyncClockRate	= info32->extSyncClockRate;     
	info64->ppsEnable		= info32->ppsEnable;
	info64->ppsOffset		= info32->ppsOffset;            
	info64->cardRevMajor		= info32->cardRevMajor;
	info64->cardRevMinor		= info32->cardRevMinor;
	info64->cardRevBuild            = info32->cardRevBuild;
	info64->features		= info32->features;
	info64->oneBitInsert		= info32->oneBitInsert;
}

void convertTo32(struct fstioc_info_32 *info32, struct fstioc_info *info64) {
	int i; 

	info32->valid			= info64->valid;
	info32->nports			= info64->nports;
	info32->type			= info64->type;          
	info32->state			= info64->state;
	info32->index			= info64->index;
	info32->smcFirmwareVersion	= info64->smcFirmwareVersion;
	info32->kernelVersion		= info64->kernelVersion;
	info32->lineInterface		= info64->lineInterface;
	info32->proto			= info64->proto;
	info32->internalClock		= info64->internalClock;
	info32->lineSpeed		= info64->lineSpeed;
	info32->estLineSpeed		= info64->estLineSpeed;
	info32->v24IpSts		= info64->v24IpSts;
	info32->v24OpSts		= info64->v24OpSts;
	info32->clockStatus		= info64->clockStatus;
	info32->cableStatus		= info64->cableStatus;
	info32->cardMode		= info64->cardMode;
	info32->debug			= info64->debug;   
	info32->transparentMode		= info64->transparentMode;
	info32->invertClock 		= info64->invertClock;
	info32->asyncAbility  		= info64->asyncAbility;
	info32->synthAbility		= info64->synthAbility;
	info32->extendedClocking	= info64->extendedClocking;
	info32->startingSlot		= info64->startingSlot;
	info32->clockSource		= info64->clockSource;
	info32->framing			= info64->framing;
	info32->structure		= info64->structure;
                         
	info32->interface		= info64->interface;
	info32->coding			= info64->coding;
	info32->lineBuildOut		= info64->lineBuildOut;
	info32->equalizer		= info64->equalizer;
	info32->loopMode		= info64->loopMode;
	info32->range			= info64->range;
	info32->txBufferMode		= info64->txBufferMode;
	info32->rxBufferMode		= info64->rxBufferMode;
	info32->losThreshold		= info64->losThreshold;
	info32->idleCode		= info64->idleCode;
	info32->receiveBufferDelay	= info64->receiveBufferDelay;
	info32->framingErrorCount	= info64->framingErrorCount;
	info32->codeViolationCount	= info64->codeViolationCount;
	info32->crcErrorCount		= info64->crcErrorCount;
	info32->lineAttenuation		= info64->lineAttenuation;
	info32->lossOfSignal		= info64->lossOfSignal;
	info32->receiveRemoteAlarm	= info64->receiveRemoteAlarm;
	info32->alarmIndicationSignal	= info64->alarmIndicationSignal;

	for (i=0; i<INFO_RESERVED; i++)
		info32->_reserved[i] = info64->_reserved[i];

	info32->ignoreCarrier		= info64->ignoreCarrier;
	info32->numTxBuffers		= info64->numTxBuffers;
	info32->numRxBuffers		= info64->numRxBuffers;
	info32->txBufferSize		= info64->txBufferSize;
	info32->rxBufferSize		= info64->rxBufferSize;
	info32->terminalType		= info64->terminalType;
	info32->annexType		= info64->annexType;
	info32->encap			= info64->encap;
	info32->testMode		= info64->testMode;
	info32->backoff			= info64->backoff;
	info32->bLineProbingEnable	= info64->bLineProbingEnable;
	info32->snrth			= info64->snrth;
	info32->lpath			= info64->lpath;
	info32->vpi			= info64->vpi;
	info32->vci			= info64->vci;
	info32->activationStatus	= info64->activationStatus;
	info32->noCommonModeStatus	= info64->noCommonModeStatus;
	info32->transceiverStatus1	= info64->transceiverStatus1;
	info32->transceiverStatus2	= info64->transceiverStatus2;
	info32->lineLoss		= info64->lineLoss;
	info32->signalQuality		= info64->signalQuality;
	info32->nearEndBlockErrorCount	= info64->nearEndBlockErrorCount;
	info32->signalToNoiseRatio	= info64->signalToNoiseRatio;
	info32->erroredSecondCount	= info64->erroredSecondCount;
	info32->severelyErroredSecondCount	= info64->severelyErroredSecondCount;
	info32->lossOfSyncWordSecondCount	= info64->lossOfSyncWordSecondCount;
	info32->unavailableSecondCount	= info64->unavailableSecondCount;
	info32->frequencyDeviation	= info64->frequencyDeviation;
	info32->negotiatedPowerBackOff	= info64->negotiatedPowerBackOff;
	info32->negotiatedPSD		= info64->negotiatedPSD;
	info32->negotiatedBChannels	= info64->negotiatedBChannels;
	info32->negotiatedZBits		= info64->negotiatedZBits;
	info32->negotiatedSyncWord	= info64->negotiatedSyncWord;
	info32->negotiatedStuffBits	= info64->negotiatedStuffBits;
	info32->chipVersion		= info64->chipVersion;
	info32->firmwareVersion		= info64->firmwareVersion;
	info32->romVersion		= info64->romVersion;
	info32->atmTxCellCount		= info64->atmTxCellCount;
	info32->atmRxCellCount		= info64->atmRxCellCount;
	info32->atmHecErrorCount	= info64->atmHecErrorCount;
	info32->atmCellsDropped		= info64->atmCellsDropped;
	info32->transmitMSBFirst	= info64->transmitMSBFirst;
	info32->receiveMSBFirst		= info64->receiveMSBFirst;
	info32->xpldVersion		= info64->xpldVersion;
	info32->farEndCountryCode[0]	= info64->farEndCountryCode[0];
	info32->farEndCountryCode[1]	= info64->farEndCountryCode[1];

	info32->farEndProviderCode[0]	= info64->farEndProviderCode[0];
	info32->farEndProviderCode[1]	= info64->farEndProviderCode[1];
	info32->farEndProviderCode[2]	= info64->farEndProviderCode[2];
	info32->farEndProviderCode[3]	= info64->farEndProviderCode[3];

	info32->farEndVendorInfo[0]	= info64->farEndVendorInfo[0];
	info32->farEndVendorInfo[1]	= info64->farEndVendorInfo[1];
	info32->utopiaAtmStatus		= info64->utopiaAtmStatus;
	info32->termination		= info64->termination;
	info32->txRxStart		= info64->txRxStart;

	info32->enableNRZIClocking	= info64->enableNRZIClocking;
	info32->lowLatency		= info64->lowLatency;

	deviceStatsTo32(&info32->stats, &info64->stats);

	info32->async_conf		= info64->async_conf;   // copy whole structure in one go
	info32->iocinfo_version		= info64->iocinfo_version;

	info32->extSyncClockEnable	= info64->extSyncClockEnable;
	info32->extSyncClockOffset	= info64->extSyncClockOffset;
	info32->extSyncClockRate	= info64->extSyncClockRate;
	info32->ppsEnable		= info64->ppsEnable;
	info32->ppsOffset		= info64->ppsOffset;
	info32->cardRevMajor		= info64->cardRevMajor;
	info32->cardRevMinor		= info64->cardRevMinor;
	info32->cardRevBuild		= info64->cardRevBuild;
	info32->features		= info64->features;
	info32->oneBitInsert		= info64->oneBitInsert;
}
#endif /* HAVE_COMPAT_IOCTL */

static unsigned int
get_baud (tcflag_t c_cflag)
{
        unsigned int baud;

        baud = c_cflag & CBAUD;
	if (baud & CBAUDEX) {
		baud &= ~CBAUDEX;

		if (baud < 1 || baud + 15 > n_baud_table)
			c_cflag &= ~CBAUDEX;
		else
			baud += 15;
	}
        baud = baud_table[baud];
        fst_dbg(DBG_IOCTL, "Baud rate was %d\n", baud);
        return baud;
}

static unsigned int
get_rtscts (tcflag_t c_cflag)
{
        unsigned int rtscts;

        rtscts = c_cflag & CRTSCTS;
        fst_dbg(DBG_IOCTL, "rtscts was %x\n", rtscts);
	if (rtscts)
	  return 1;
	else
	  return 0;
}

static unsigned int
get_xonxoff (tcflag_t c_iflag)
{
       unsigned int xon, xoff;

        xon = c_iflag & IXON;
	xoff = c_iflag & IXOFF;
        fst_dbg(DBG_IOCTL, "xon was %x xoff was %x\n", xon, xoff);
	if (xon && xoff)
	  return 2;
	else
	  return 0;
}

static unsigned int
get_char_size (tcflag_t c_cflag)
{
        unsigned int char_size;

        char_size = c_cflag & CSIZE;
        fst_dbg(DBG_IOCTL,"Character size = %d\n", char_size);
        switch (char_size)
        {
          case CS5:
            fst_dbg(DBG_IOCTL, "Character size set as 5 bits\n");
              return 5;

          case CS6:
            fst_dbg(DBG_IOCTL, "Character size set as 6 bits\n");
              return 6;

          case CS7:
            fst_dbg(DBG_IOCTL, "Character size set as 7 bits\n");
              return 7;

          case CS8:
          default:
            fst_dbg(DBG_IOCTL, "Character size set as 8 bits\n");
              return 8;

        }
}

static unsigned int
get_stop_bits (tcflag_t c_cflag)
{
        unsigned int stop_bits;

        stop_bits = c_cflag & CSTOPB;
        fst_dbg(DBG_IOCTL, "Stop bits = %d\n", stop_bits);
        if (stop_bits)
            {
              fst_dbg(DBG_IOCTL, "2 stop bits selected\n");
                return 2;
            }
        fst_dbg(DBG_IOCTL, "1 stop bit selected\n");
        return 1;
}

static unsigned int
get_parity(tcflag_t c_cflag)
{
        unsigned int parity_enable, parity;

        parity_enable = c_cflag & PARENB;
        parity = c_cflag & PARODD;
        fst_dbg(DBG_IOCTL,"Parity enable is %d and parity odd is %d\n", 
                parity_enable, parity);
	if (parity_enable)
            {
                if (parity)
                    {
                      fst_dbg(DBG_IOCTL,"Returning Odd Parity\n");
                        return 1;
                    }
                else
                    {
                      fst_dbg(DBG_IOCTL,"Returning Even Parity\n");
                        return 2;
                    }
            }
       else
            {
              fst_dbg(DBG_IOCTL, "Parity not enabled\n");
                return 0;
            }
}

static void
init_async_parameters (struct fst_port_info *port, 
		       unsigned int char_size, unsigned int stop_bits, 
		       unsigned int parity, unsigned int flow_control)
{
        int retval;

        fst_dbg(DBG_IOCTL, "%s: Initialising async parameters\n", port->dev->name);
	flex_serconf_to_le(&port->usb_config);	
	retval = usb_control_msg(port->card->udev, 
				 usb_rcvctrlpipe(port->card->udev, 0),
				 USS_CMD_GET_CONFIG, 0xc0, 0, 0, 
                                 &port->usb_config, 
                                 sizeof(FLEX_SERCONF), 10000);
	flex_serconf_to_cpu(&port->usb_config);				 
	if (retval != sizeof(port->usb_config))
	  {
            printk("Get Config error errno = %d\n", retval);
	  }
	port->usb_config.flowControl = flow_control;
	port->usb_config.xonchar = 0x11;
        port->usb_config.xoffchar = 0x13;
        port->usb_config.dataBits = char_size;
	port->usb_config.stopbits = stop_bits;
	port->usb_config.parity = parity;
        fst_dbg(DBG_IOCTL, "%s: Reconfiguring Port\n", port->dev->name);
	flex_serconf_to_le(&port->usb_config);
	retval = usb_control_msg(port->card->udev, 
				 usb_sndctrlpipe(port->card->udev, 0),
				 USS_CMD_SET_CONFIG, 0x40, 0, 0, 
                                 &port->usb_config, 
                                 sizeof(FLEX_SERCONF), 10000);
	flex_serconf_to_cpu(&port->usb_config);			 
	if (retval>0) 
          {
            fst_dbg(DBG_IOCTL, "Set config response received\n");
          }
        else
          {
            printk("Set Config error errno = %d\n", retval);
          }
}

static void
configure_async_parameters (struct fst_port_info *port, unsigned int baud, 
                            unsigned int char_size, unsigned int stop_bits, 
                            unsigned int parity, unsigned int flow_control)
{
        int retval;

        fst_dbg(DBG_IOCTL, "%s: Setting async parameters\n", port->dev->name);
        if (port->run)
          fst_issue_cmd ( port, STOPPORT );
	port->usb_config.flowControl = COM_FLOW_CONTROL_NONE;
	if ((flow_control == 1) || (flow_control == 3))
	  port->usb_config.flowControl = COM_FLOW_CONTROL_RTSCTS;
	if (flow_control == 2)
	  port->usb_config.flowControl = COM_FLOW_CONTROL_XONXOFF;
	fst_dbg(DBG_IOCTL, "Flow control chosen as %d\n", port->usb_config.flowControl);
	port->usb_config.xonchar = 0x11;
        port->usb_config.xoffchar = 0x13;
        port->usb_config.lineSpeed = baud ;
        port->usb_config.clock = 1 ;
        port->usb_config.dataBits = char_size;
        if (stop_bits == 2)
        {
              port->usb_config.stopbits = COM_STOP_BITS_2;
        }
        else
        {
              port->usb_config.stopbits = COM_STOP_BITS_1;
        }

        if (parity)
        {
            if (parity == 1)
                port->usb_config.parity = COM_ODD_PARITY;
            else
                port->usb_config.parity = COM_EVEN_PARITY;
        }
        else
        {
            port->usb_config.parity = COM_NO_PARITY;
        }
        fst_dbg(DBG_IOCTL, "%s: Reconfiguring Port\n", port->dev->name);
        flex_serconf_to_le(&port->usb_config);
	retval = usb_control_msg(port->card->udev, 
				 usb_sndctrlpipe(port->card->udev, 0),
				 USS_CMD_SET_CONFIG, 0x40, 0, 0, 
                                 &port->usb_config, 
                                 sizeof(FLEX_SERCONF), 10000);
	flex_serconf_to_cpu(&port->usb_config);

	if (retval>0) 
          {
            fst_dbg(DBG_IOCTL, "Set config response received\n");
          }
        else
          {
            printk("Set Config error errno = %d\n", retval);
          }
        if (port->run)
          fst_issue_cmd ( port, STARTPORT );
}

#define MAX_TRANSFER		FST_MAX_USB_BUFFER
#define FST_CHAR_INQ_MAX_LEN    100;
#ifdef FST_TTY_CHAR
/* 
 * Char tty interface for async PPP
 */
#define FST_TTY_NPORTS         FST_MAX_CARDS*FST_MAX_PORTS
#define FST_TTY_MAJOR          191
#define FST_TTY_MINOR_START    0
/* tty interface state */
#define	FST_TTY_ST_CLOSED      0  /* Initial state */
#define FST_TTY_ST_OPEN	       1  /* Has at least one open */
#define FST_TTY_ST_DATA        2  /* Open and used for data */
#define FST_TTY_ST_DISC        3  /* Open but line has dropped */

/* TTY functions prototype */
static int  fst_tty_open(struct tty_struct *tty, struct file *flip);
static void fst_tty_close(struct tty_struct *tty, struct file *flip);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
static int  fst_tty_write(struct tty_struct *tty, const unsigned char *buf, 
                          int count);
#else
static int  fst_tty_write(struct tty_struct *tty, int from_user,
                          const unsigned char *buf, int count);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,14,0)
static unsigned int  fst_tty_write_room(struct tty_struct *tty);
static unsigned int  fst_tty_chars_in_buffer(struct tty_struct *tty);
#else
static int  fst_tty_write_room(struct tty_struct *tty);
static int  fst_tty_chars_in_buffer(struct tty_struct *tty);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
int fst_tty_tiocmset(struct tty_struct *, unsigned int, unsigned int);
int fst_tty_tiocmget(struct tty_struct *);
#else
int fst_tty_tiocmset(struct tty_struct *, struct file *, 
                     unsigned int, unsigned int);
int fst_tty_tiocmget(struct tty_struct *, struct file *);
#endif
static void fst_tty_flush_buffer(struct tty_struct *tty);
static void fst_tty_hangup(struct tty_struct *tty);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
static void fst_tty_set_termios (struct tty_struct *tty,
                                 struct ktermios *old);
#else
static void fst_tty_set_termios (struct tty_struct *tty,
                                 struct termios *old);
#endif

static struct tty_driver *serial_drv;

typedef	struct _fst_tty_area {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
        struct tty_port tty_port;
#endif
	int		state;		/* state of the TTY interface */
	int		num_open;	
	unsigned int 	tty_minor;	/* minor this interface */
        struct fst_port_info  *port;	        /* ptr. to port info */
	unsigned char	name[20];	/* interf. name + "-tty" */
	struct tty_struct *tty;
	struct sk_buff     *skb;
	int 		   len;		/* async */
	char		  *data; 	/* async */
	struct work_struct  fst_tty_work;
	struct work_struct  fst_async_tty_work;	
} st_fst_tty_area;

st_fst_tty_area	fst_tty_area[FST_TTY_NPORTS];
int fst_tty_refcount=0;
int fst_tty_unreg_flag = 0;
int fst_tty_cnt = 0;

void fst_tty_uninit (void)
{
  /*
   * Unload the tty driver
   */
  
  printk_info("Unregister the tty driver\n");
  tty_unregister_driver(serial_drv);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)
  tty_driver_kref_put(serial_drv);
#else
  put_tty_driver(serial_drv);
#endif
}

static const struct tty_operations fst_tty_ops = {
	.open = fst_tty_open,
	.close = fst_tty_close,
	.write = fst_tty_write,
	.write_room = fst_tty_write_room,
	.chars_in_buffer = fst_tty_chars_in_buffer,
	.tiocmset = fst_tty_tiocmset,
	.tiocmget = fst_tty_tiocmget,
	.set_termios = fst_tty_set_termios,
	.flush_buffer = fst_tty_flush_buffer,
	.hangup = fst_tty_hangup,
};

int fst_tty_init_body (void)
{
  int base_port_number = 0;

  base_port_number = find_last_hdlc_device();

  /* initialize tty driver struct */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)
  serial_drv = tty_alloc_driver(FST_TTY_NPORTS, 0);
  if (IS_ERR(serial_drv))
    serial_drv = NULL;
#else
  serial_drv = alloc_tty_driver(FST_TTY_NPORTS);
#endif
  if (serial_drv == NULL)
    return 0;
  serial_drv->magic = TTY_DRIVER_MAGIC;
  serial_drv->driver_name = "flex_tty";
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
  serial_drv->name = "tty_hdlc";
  serial_drv->major = 0;
  serial_drv->name_base = base_port_number;
#else
  serial_drv->name = "fsflex";
  serial_drv->major = FST_TTY_MAJOR ;
#endif
  serial_drv->minor_start = 0;     /* Use the whole of the minor range */
  serial_drv->num = FST_TTY_NPORTS;
  serial_drv->type = TTY_DRIVER_TYPE_SERIAL;
  serial_drv->subtype = SERIAL_TYPE_NORMAL;
  serial_drv->init_termios = tty_std_termios;
  serial_drv->init_termios.c_cflag = B9600|CS8|CREAD|HUPCL|CLOCAL;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17)
  serial_drv->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_DYNAMIC_DEV;
#else
  serial_drv->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
#endif

  /* interface routines from the upper tty layer to the tty driver */
  tty_set_operations(serial_drv, &fst_tty_ops);

  /* register the TTY driver */
  if (tty_register_driver(serial_drv)) 
    { 
      printk("fsc: Failed to register serial driver!\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)
      tty_driver_kref_put(serial_drv);
#else
      put_tty_driver(serial_drv);
#endif
      return 0;
    } 
  return 1; /* OK */  
}

int fst_tty_init (void)
{
  /*
   * Load the tty driver
   */
  int retval=0;

  printk_info("Initialising flex tty driver\n");
  retval = fst_tty_init_body();
  if (retval)
    memset((void *)fst_tty_area, 0,
           sizeof(st_fst_tty_area) * FST_TTY_NPORTS);
  return retval; 
}

static int fst_tty_open(struct tty_struct *tty, struct file *file)
{
  /*
   * Open a farsync port 
   * It must not already be open by the network stack
   */
  int port_num;
  st_fst_tty_area *fst_tty;
  struct net_device *dev;
  struct fst_port_info *port;
  char dev_name[16];
  
  fst_dbg(DBG_ASS, "fst_tty_open\n");
  if (!tty) 
    { 
      return -ENODEV;
    } 

  port_num = tty->index; 
  
  if ((port_num < 0) || (port_num >= FST_TTY_NPORTS))
    { 
      fst_dbg(DBG_ASS, "fst_tty: open invalid flex minor %i\n", port_num);
      return -ENODEV;
    } 
  
  fst_tty = &fst_tty_area[port_num];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
  INIT_WORK(&fst_tty->fst_tty_work, fst_process_tty_work);
  INIT_WORK(&fst_tty->fst_async_tty_work, fst_process_async_tty_work);
#else
  INIT_WORK(&fst_tty->fst_tty_work, fst_process_tty_work, fst_tty);
  INIT_WORK(&fst_tty->fst_async_tty_work, fst_process_async_tty_work, fst_tty);
#endif  
  
  if (fst_tty->num_open == 0) 
    { 
      /*
       * We need to find the dev device and see if it is open
       */
      if ((dev = get_net_device(port_num)) == NULL)
        {
          fst_dbg(DBG_ASS, "Cannot open char device %s because net device not found\n", dev->name);
          return -ENXIO;
        }
      port = dev_to_port(dev);
      if (port == NULL)
        {
          fst_dbg(DBG_ASS, "Cannot open char device %s because port is not created\n", dev->name);
          return -EACCES;
        }
      if (port->run)
	{
	  /*
           * It is likely that the network device is active
	   * so refuse this open
	   */
	  return -EBUSY;
	}
      /* first open of this tty */
      fst_tty->state = FST_TTY_ST_OPEN;
      fst_tty->tty = tty;
      tty->driver_data = &fst_tty_area[port_num];
      fst_tty->num_open= 0;
      fst_tty->tty_minor = port_num + FST_TTY_MINOR_START;

      strcpy(dev_name, "tty_");
      strcat(dev_name, dev->name);
      strcpy(fst_tty->name, dev->name);
      fst_dbg(DBG_ASS, "%s: Initializing Flex TTY Sync Driver, tty major#%d minor#%i\n",
          fst_tty->name,FST_TTY_MAJOR,fst_tty->tty_minor); 
      

      /*
       * Set a flag to indicate we are now using the char interface for 
       * data read and write
       */
      port->char_file = file;
      /*
       * Open the physical port if not already open
       */
      if (!port->port_mode)
        fst_openport(port);
      fst_tty->port = port;
      try_module_get(THIS_MODULE);
      fst_tty_cnt++;
      printk_info("TTY driver flex port %s is now open\n", fst_tty->name);
    } 
  fst_tty->num_open++;        
  return 0; 
}

static void fst_tty_close(struct tty_struct *tty, struct file *flip)
{
  /*
   * Close a farsync port
   */
  st_fst_tty_area    *fst_tty;
  unsigned long flags;

	if (!tty || !tty->driver_data ) 
          {
            fst_dbg(DBG_ASS, "fst_tty: no Flex TTY in close \n");
            return;
          }

	fst_tty = (st_fst_tty_area *) tty->driver_data;

	if ((fst_tty->tty != tty)|| (fst_tty->state < FST_TTY_ST_OPEN)) 
          {
            fst_dbg(DBG_ASS, "%s: TTY is not opened\n",fst_tty->name);
            return;
          }
   	
	if (!fst_tty->num_open) 
          {
            fst_dbg(DBG_ASS, "%s: TTY is closed\n",fst_tty->name);
            return;
          }

	if (--fst_tty->num_open > 0) 
          {
            fst_dbg(DBG_ASS, "%s: TTY closing the second open\n",
                fst_tty->name);
            return;
          }

        fst_tty_cnt--;
        if (fst_tty->state != FST_TTY_ST_CLOSED)
          {
            spin_lock_irqsave(&fst_tty->port->card->card_lock, flags);
            fst_tty->tty = NULL;
            fst_tty->state = FST_TTY_ST_CLOSED;
            spin_unlock_irqrestore(&fst_tty->port->card->card_lock, flags);
            fst_closeport(fst_tty->port);
	    fst_tty->port->run = 0;
            fst_tty->port->char_file = NULL;
          }
        module_put(THIS_MODULE);
	printk_info("TTY driver flex port %s is now closed\n",fst_tty->name);
	return; 
} 

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
static int fst_tty_write(struct tty_struct *tty, const unsigned char *buf, 
                         int count)
#else
static int fst_tty_write(struct tty_struct *tty, int from_user,
			const unsigned char *buf, int count)
#endif
{
  /*
   * Send data to the farsync port
   */
  st_fst_tty_area    *fst_tty; 
  struct sk_buff     *skb;
  int                 retval;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
  int from_user = 0;
#endif

  if (!tty || !tty->driver_data ) 
    { 
      fst_dbg(DBG_ASS, "Could not find Flex TTY device in write\n");
      return -ENODEV;
    } 

  fst_tty = (st_fst_tty_area *) tty->driver_data; 

  if ((fst_tty->tty != tty) ||  (fst_tty->state < FST_TTY_ST_OPEN)) 
    { 
      fst_dbg(DBG_ASS, "%s: TTY device is not opened\n", fst_tty->name);
      return -ENODEV; 
    }

  if (count > FST_MAX_MTU) 
    { 
      fst_dbg(DBG_ASS, "%s: Size of write is too large\n",fst_tty->name);
      return -EINVAL;        /* frame too big */ 
    }

  if (fst_tty->state == FST_TTY_ST_CLOSED)
    {
      fst_dbg(DBG_TTY, "%s: discarding %d bytes of write data on a closed device\n", 
              fst_tty->name, count);
      return -ENODEV;
    }
  fst_dbg(DBG_TTY, "%s: fst_tty_write %s data len=%i\n",fst_tty->name,
      (from_user)?"from user" : "from kernel",count);
  /*
   * No special treatment for async mode, unlike farsync driver
   */
  if ( (skb = dev_alloc_skb ( count )) == NULL )
    {
      fst_dbg( DBG_ASS,"fst_tty_write: can't allocate skb\n");
      return -EFAULT;
    }

  /* Read it in */
  if (from_user)
    {
      fst_dbg(DBG_TTY, "Value of from user is %d \n", from_user);
      if (__copy_from_user ( skb_put(skb, count), buf, count ))
        {
          dev_kfree_skb(skb);
          return -EFAULT;
        }
    }
  else
    {
      memcpy(skb_put(skb, count), buf, count);
    }
  fst_tty->state = FST_TTY_ST_DATA;
  retval = fst_start_xmit( skb, port_to_dev(fst_tty->port));
  if (retval != count)
    {
      fst_dbg(DBG_TTY, "%s: tty_write asked to transmit %d, transmitted %d\n",
             fst_tty->name, count, retval);
    }
  return retval;
} 

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,14,0)
static unsigned int fst_tty_write_room(struct tty_struct *tty)
#else
static int fst_tty_write_room(struct tty_struct *tty)
#endif
{
  st_fst_tty_area    *fst_tty; 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,14,0)
  unsigned int room = FST_MAX_MTU;
#else
  int room = FST_MAX_MTU;
#endif

  if (!tty || !tty->driver_data ) { 
    fst_dbg(DBG_ASS, "Could not find TTY device to write room\n");
    return -ENODEV;
  }
  
  fst_tty = (st_fst_tty_area *) tty->driver_data; 
  
  if ((fst_tty->tty != tty) ||  (fst_tty->state < FST_TTY_ST_OPEN)) 
    { 
      fst_dbg(DBG_ASS, "%s: TTY device is not opened\n",fst_tty->name);
      return -ENODEV; 
  }
  
  fst_dbg(DBG_ASS, "%s: write room\n",fst_tty->name);
  if (fst_tty->port->start)
    room = 0;
  fst_dbg(DBG_TTY, "%s: tty_write_room returning %d\n", fst_tty->name, room);
  return room;
} 

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,14,0)
static unsigned int fst_tty_chars_in_buffer(struct tty_struct *tty)
#else
static int fst_tty_chars_in_buffer(struct tty_struct *tty)
#endif
{
  st_fst_tty_area    *fst_tty; 

  if (!tty || !tty->driver_data ) 
    {
      fst_dbg(DBG_ASS, "Could not find TTY device for chars in buffer\n");
      return -ENODEV; 
    }
  
  fst_tty = (st_fst_tty_area *) tty->driver_data; 
  
  if ((fst_tty->tty != tty) || (fst_tty->state < FST_TTY_ST_OPEN)) 
    { 
      fst_dbg(DBG_ASS, "%s: TTY device is not opened\n",fst_tty->name);
      return -ENODEV; 
    }
  fst_dbg(DBG_TTY, "%s: tty_chars_in_buffer returning 0\n", fst_tty->name);
  return 0; 
} 

static void
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
fst_tty_set_termios (struct tty_struct *tty, struct ktermios *old)
#else
fst_tty_set_termios (struct tty_struct *tty, struct termios *old)
#endif
{
	st_fst_tty_area    *fst_tty; 
        unsigned int        baud;
        unsigned int        stop_bits;
        unsigned int        char_size;
        unsigned int        parity;
	unsigned int        rtscts;
	unsigned int        xonxoff;

	fst_dbg(DBG_TTY, "tty_set_termios\n");
	if (!tty || !tty->driver_data ) {
	   	fst_dbg(DBG_ASS, "Could not find TTY for tiocmset\n");	
		return; 
	}
	fst_tty = (st_fst_tty_area *) tty->driver_data; 
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0)
	old = fst_tty->tty->termios;
#else
	old = &fst_tty->tty->termios;
#endif
	fst_dbg(DBG_TTY, "%s: Setting %x\n", fst_tty->name, old->c_cflag);
	baud = get_baud(old->c_cflag);
	char_size = get_char_size(old->c_cflag);
	stop_bits = get_stop_bits(old->c_cflag);
	parity = get_parity(old->c_cflag);
	rtscts = get_rtscts(old->c_cflag);
        xonxoff = get_xonxoff(old->c_iflag);
	/*
	 * If we are in async mode we need to configure the async parameters
	 */
	if (fst_tty->port->mode == FST_MODE_ASYNC)
	  {
	    configure_async_parameters(fst_tty->port, baud, char_size, 
                                       stop_bits, parity, rtscts|xonxoff);
	  }
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
int fst_tty_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear)
#else
int fst_tty_tiocmset(struct tty_struct *tty, struct file *file,
			unsigned int set, unsigned int clear)
#endif
{
	st_fst_tty_area    *fst_tty;
	unsigned long outputs;

	fst_dbg(DBG_TTY, "%s: set:%x clear:%x\n", __FUNCTION__, set, clear);

	if (!tty || !tty->driver_data ) {
	   	fst_dbg(DBG_ASS, "Could not find TTY for tiocmset\n");	
		return -ENODEV; 
	}

	fst_tty = (st_fst_tty_area *) tty->driver_data; 

	if (set & TIOCM_RTS)
          {
            fst_dbg(DBG_TTY, "Set RTS\n");
	    outputs = OPSTS_RTS | fst_tty->port->v24OpSts;
	    fst_tty->port->v24OpSts = outputs;
	    if ( fst_tty->port->run )
	      fst_issue_cmd ( fst_tty->port, SETV24O );
          }

	if (set & TIOCM_DTR)
          {
            fst_dbg(DBG_TTY, "Set DTR\n");
	    outputs = OPSTS_DTR | fst_tty->port->v24OpSts;
	    fst_tty->port->v24OpSts = outputs;
	    if ( fst_tty->port->run )
	      fst_issue_cmd ( fst_tty->port, SETV24O );
          }

	if (clear & TIOCM_RTS)
          {
            fst_dbg(DBG_TTY, "Clear RTS\n");
	    outputs = OPSTS_RTS ^ fst_tty->port->v24OpSts;
	    fst_tty->port->v24OpSts = outputs;
	    if ( fst_tty->port->run )
	      fst_issue_cmd ( fst_tty->port, SETV24O );
          }
	if (clear & TIOCM_DTR)
          {
            fst_dbg(DBG_TTY, "Clear DTR\n");
	    outputs = OPSTS_DTR ^ fst_tty->port->v24OpSts;
	    fst_tty->port->v24OpSts = outputs;
	    if ( fst_tty->port->run )
	      fst_issue_cmd ( fst_tty->port, SETV24O );
          }
	return 0;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
int fst_tty_tiocmget(struct tty_struct *tty)
#else
int fst_tty_tiocmget(struct tty_struct *tty, struct file *file)
#endif
{
	unsigned int signals = 0;
        unsigned long out = 0;

	st_fst_tty_area  *fst_tty = (st_fst_tty_area *) tty->driver_data;
	fst_tty = (st_fst_tty_area *) tty->driver_data;

	fst_dbg(DBG_TTY, "%s: tiocmget\n",fst_tty->name);

	signals = fst_tty->port->v24OpSts;
	if (signals & OPSTS_RTS)
	  out |= TIOCM_RTS;
	if (signals & OPSTS_DTR)
	  out |= TIOCM_DTR;
	fst_issue_cmd (fst_tty->port, GETSIGNALS);
	signals = fst_tty->port->v24IpSts;
	if (signals & IPSTS_CTS)
	  out |= TIOCM_CTS;
	if (signals & IPSTS_DSR)
	  out |= TIOCM_DSR;
	if (signals & IPSTS_DCD)
	  out |= TIOCM_CD;
        fst_dbg(DBG_TTY, "Returning signals as %lx\n", out);
	return out;
}

static void fst_tty_flush_buffer(struct tty_struct *tty)
{ 
  st_fst_tty_area    *fst_tty; 
	
  if (!tty || !tty->driver_data ) 
    {
      fst_dbg(DBG_ASS, "Could not find TTY to flush buffer\n");	
      return; 
    }
  
  fst_tty = (st_fst_tty_area *) tty->driver_data; 
  
  if ((fst_tty->tty != tty) ||  (fst_tty->state < FST_TTY_ST_OPEN)) 
    { 
      fst_dbg(DBG_ASS, "%s: TTY device is not opened\n",fst_tty->name);
      return; 
    }
  
  fst_dbg(DBG_TTY, "%s: call wake_up_interruptible\n",fst_tty->name);
  
  wake_up_interruptible(&tty->write_wait); 
  
  return; 
} 

static void fst_tty_hangup(struct tty_struct *tty)
{ 
  st_fst_tty_area    *fst_tty; 
  
  if (!tty || !tty->driver_data ) 
    {
      fst_dbg(DBG_ASS, "Could not find TTY device to hangup\n");	
      return ; 
    }
  
  fst_tty = (st_fst_tty_area *) tty->driver_data; 

  if ((fst_tty->tty != tty) || (fst_tty->state < FST_TTY_ST_OPEN)) 
    {
      fst_dbg(DBG_ASS, "%s: TTY device is not opened\n",fst_tty->name);
      return ;
    }
  fst_dbg(DBG_ASS, "fst_tty_hangup\n");
}

//static int
//tty_signal_async_rx_frame ( struct fst_port_info *port, char * data, int len)
static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
fst_process_async_tty_work ( struct work_struct *work )
#else
fst_process_async_tty_work ( void *work )
#endif
{

  st_fst_tty_area *fst_tty;
  struct fst_port_info *port; 
  char *data;
  int len;  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  struct tty_ldisc *ld;
#endif
  
  if (fst_tty_cnt == 0)
    return;
  
  /* 
   * ToDo: Note that the third parameter to the ldisc receive_buf function is
   * a pointer to an array of status bytes which is the same length as
   * the data.  When we move on to getting the status bytes from the 
   * async interface we can implement this.  For the moment it is a NULL
   * pointer
   */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
  fst_tty = container_of(work, st_fst_tty_area, fst_async_tty_work);
#else
  fst_tty= work;
#endif
  port = fst_tty->port;
  len  = fst_tty->len;
  data = fst_tty->data; 

  fst_tty->state = FST_TTY_ST_DATA;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  ld = tty_ldisc_ref(fst_tty->tty);
  if (ld) 
    {
      if (fst_tty->tty && (ld->ops->receive_buf)) 
	{ 
          fst_dbg(DBG_TTY, "%s: call line disc. receive_buf of length %d first byte is %x\n", 
                  fst_tty->name, len, data[0]);
	  ld->ops->receive_buf(fst_tty->tty, data, 
			       NULL, len);
          tty_ldisc_deref(ld);
	  port_to_stats(port,rx_packets)++;
	  port_to_stats(port,rx_bytes) += len;
	}
      else
	{
	  fst_dbg(DBG_ASS, "%s: frame dropped. receive_buf of length %d\n", 
		  fst_tty->name, len);
	  port_to_stats(port,rx_dropped)++;
	}
    }
#else
  if (fst_tty->tty && (fst_tty->tty->ldisc.receive_buf)) 
    { 
      fst_dbg(DBG_TTY, "%s: call line disc. receive_buf of length %d\n", 
          fst_tty->name, len);
      fst_tty->tty->ldisc.receive_buf(fst_tty->tty, data, 
                                      NULL, len);
      port_to_stats(port,rx_packets)++;
      port_to_stats(port,rx_bytes) += len;
    }
  else
    {
      fst_dbg(DBG_ASS, "%s: frame dropped. receive_buf of length %d\n", 
          fst_tty->name, len);
      fst_dbg(DBG_ASS, "tty address = %p, ldisc routine = %p\n",
          fst_tty->tty, fst_tty->tty->ldisc.receive_buf);
      port_to_stats(port,rx_dropped)++;
    }
#endif
  
}

static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
fst_process_tty_work ( struct work_struct *work )
#else
fst_process_tty_work ( void *work )
#endif
{
  struct fst_port_info *port;
  st_fst_tty_area *fst_tty;
  struct sk_buff *skb;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  struct tty_ldisc *ld;
#endif
  if (fst_tty_cnt == 0) return;
  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
  fst_tty = container_of(work, st_fst_tty_area, fst_tty_work);
#else
  fst_tty = work;
#endif
  port = fst_tty->port;
  skb  = fst_tty->skb;
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  
  ld = tty_ldisc_ref(fst_tty->tty);
  if (ld) 
    {
      if ((fst_tty->tty) && (ld->ops->receive_buf)) 
	{ 
	  fst_dbg(DBG_ASS, "%s: call line disc. receive_buf of length %d\n", 
		  fst_tty->name, skb->len);

	  ld->ops->receive_buf(fst_tty->tty, skb->data, 
			       NULL, skb->len);
	}
      else
	{
	  fst_dbg(DBG_ASS, "%s: frame dropped. receive_buf of length %d\n", 
		  fst_tty->name, skb->len);
	}
    tty_ldisc_deref(ld);
    }    
#else
  if (fst_tty->tty && (fst_tty->tty->ldisc.receive_buf)) 
    { 
      fst_dbg(DBG_TTY, "%s: call line disc. receive_buf of length %d\n", 
          fst_tty->name, skb->len);
      fst_tty->tty->ldisc.receive_buf(fst_tty->tty, skb->data, 
                                      NULL, skb->len);
    }
  else
    {
      fst_dbg(DBG_ASS, "%s: frame dropped. receive_buf of length %d\n", 
          fst_tty->name, skb->len);
      fst_dbg(DBG_ASS, "tty address = %p, ldisc routine = %p\n",
          fst_tty->tty, fst_tty->tty->ldisc.receive_buf);
    }
#endif
  dev_kfree_skb(skb);
}

/* Card memory image access */
static const struct file_operations fst_mem_fops = {
};

#else
/*
 * Char interface for RAW API
 */
#define FST_SYNC_NPORTS  FST_MAX_CARDS*FST_MAX_PORTS
#define UNIT(K)		(MINOR(K) & 0x3F)	/* Port number */

static struct termios fst_serial_termios[FST_SYNC_NPORTS];
static int char_check_rx_queue_length (struct fst_port_info *port);
static int char_purge_one (struct fst_port_info *port);
static int char_queue_rx_frame ( struct fst_port_info *port, 
                                 struct sk_buff *skb);

static void
fst_sync_init (void)
{
    int i;

    fst_dbg(DBG_ASS, "Initialising serial parameters\n");
    for (i=0; i<FST_SYNC_NPORTS; i++)
    {
        fst_serial_termios[i].c_cflag = B9600|CS8|CREAD|HUPCL|CLOCAL;
    }
}

static int
fst_sync_open ( struct inode *inode, struct file *file )
{
	int unit;
        struct net_device *net_dev;
        struct fst_card_info *usb_dev;
        struct fst_port_info *port;

        unit = UNIT ( inode->i_rdev );
	fst_dbg ( DBG_OPEN, "fst_sync_open: %d on device %d\n", 
                 inode->i_rdev, unit );
        /*
         * We need to find the dev device and see if it is open
         */
        if ((net_dev = get_net_device(unit)) == NULL)
          {
            fst_dbg(DBG_ASS, "Cannot open char device sync%d because net device not found\n", unit);
            return -ENXIO;
          }
#if 0
	printk("The default flags for the file are %x\n", file->f_flags);
	if (file->f_flags & O_RDONLY)
	  printk("    file is read only %x\n", O_RDONLY);
	if (file->f_flags & O_WRONLY)
	  printk("    file is write only %x\n", O_WRONLY);
	if (file->f_flags & O_RDWR)
	  printk("    file is read/write %x\n", O_RDWR);
	if (file->f_flags & O_APPEND)
	  printk("    file is to be appended to %d\n", O_APPEND);
	if (file->f_flags & O_CREAT)
	  printk("    file is to be created if it doesn't exist %x\n", O_CREAT);
	if (file->f_flags & O_EXCL)
	  printk("    must be used with O_CREAT and tests for the file %x\n", O_EXCL);
	if (file->f_flags & O_TRUNC)
	  printk("    file is truncated on open %x\n", O_TRUNC);
	if (file->f_flags & O_NOCTTY)
	  printk("    file is not a controlling terminal %x\n", O_NOCTTY);
	if (file->f_flags & O_NONBLOCK)
	  printk("    file is accessed in non blocking mode %x\n", O_NONBLOCK);
	if (file->f_flags & O_SYNC)
	  printk("    file must wait for writes to complete %x\n", O_SYNC);
#endif
        port = dev_to_port(net_dev);
        usb_dev = port->card;
        /*
         * Set a flag to indicate we are now using the char interface for 
         * data read and write
         */
        port->char_file = file;
        file->private_data = port;
	fst_dbg(DBG_OPEN, "Setting mtu to %d\n", port->rx_buffer_size);
        port_to_dev(port)->mtu = port->rx_buffer_size;
        /*
         * Open the physical port if not already open
         */
        if ((!port->port_mode) && (!port->char_opens))
          fst_openport(port);
        usb_dev->actual_length = 0;
	try_module_get(THIS_MODULE);
        port->char_opens++;
	return 0;		/* OK */
}

/*	Simple close just decrements the module count
 */
static int
fst_sync_release ( struct inode *inode, struct file *file )
{
	int unit;
	struct fst_card_info *dev;
        struct fst_port_info *port;

        unit = UNIT ( inode->i_rdev );
	fst_dbg( DBG_OPEN,"fst_sync_release: %d on device %d\n", 
              inode->i_rdev, unit );
	port = (struct fst_port_info *)file->private_data;
        if (port == NULL)
        {
	  fst_dbg (DBG_ASS, "fst_sync_release: Device gone\n");
          module_put(THIS_MODULE);
	  /*
	   * Was there a delayed release mode set?
	   */
          fst_dbg(DBG_OPEN, "Has net device already been free'd for %d\n", unit);
	  if (fst_net_list[unit])
	    {
	      fst_dbg(DBG_ASS, "Doing the delayed release of %p\n", 
		  fst_net_list[unit]);
	      free_netdev(fst_net_list[unit]);
	      fst_net_list[unit] = NULL;
	    }
	  if (fst_ports_list[unit])
	    {
	      fst_dbg(DBG_OPEN, "Decrementing the port open count for %p\n",
		     fst_ports_list[unit]);
	      fst_ports_list[unit]->char_opens--;
	    }
	  return 0;
        }
        port->char_file = NULL;
        wake_up_interruptible (&port->pollq);
        dev = port->card;
	if (dev == NULL)
		return -ENODEV;
        if (port->char_opens == 1)
          {
	    if (dev->tx_urb_cnt != MAX_URBS_OUTSTANDING)
	    {
              if (dev->tx_bulk_urb)
              {
                fst_dbg(DBG_OPEN, "Cancel a Tx urb\n");
                usb_kill_urb(dev->tx_bulk_urb); /* Stop sending data */
              }
	      fst_dbg(DBG_ASS, "Closing port with urbs outstanding\n");
	    }

	    /*
             * We only want to physically close the port
             * if this is the last close
             */
            fst_closeport(port);
           	    
            /*
             * Better make sure that the rx queue in empty
             */
            while (char_check_rx_queue_length(port) > 0)
              {
                /*
                 * Remove frames one by one
                 */
                char_purge_one(port);
              }
          }
	module_put(THIS_MODULE);
        port->char_opens--;
	return 0;
}

static ssize_t
fst_read_async_data ( struct file *file, struct fst_port_info *port, 
                       char *buf, size_t count, loff_t *offp )
{
        int len=0, i;
	char *my_buff = NULL;
	unsigned long flags;

        /* 
	 * The data is in the port rxfifo so we just need 
	 * to copy it to user space
	 */
	fst_dbg(DBG_RX, "fst_read_async_data: App read on port %s\n",
	       port->dev->name);
	while (check_fifo_space((char *)port->fifo_rxdata) == 0)
	  {
	    /*
	     * Nothing in the fifo yet, so return eagain or block and wait
	     */
	    if (file->f_flags & O_NONBLOCK)
	      return -EAGAIN;
	    if (wait_event_interruptible(port->readq, check_fifo_space((char *)port->fifo_rxdata) > 0))
                  return -ERESTARTSYS;
	  }
        my_buff = kmalloc ( FST_RX_DATA_FIFO_LEN, GFP_ATOMIC );
        if ( my_buff == NULL )
          {
            printk("%s: Cannot allocate memory for user rx buffer\n",
                   port->dev->name);
            return -EAGAIN;
          }
	spin_lock_irqsave(&port->card->fifo_lock, flags);
	len = read_from_fifo((char *)port->fifo_rxdata, FIFO_PORT_RX, my_buff, count);
	spin_unlock_irqrestore(&port->card->fifo_lock, flags);
	if (len > count)
	  {
	    fst_dbg(DBG_TTY, "User buffer not large enough, truncated %d to %d\n",
		   len, (int) count);
	    len = count;
	  }
	i = copy_to_user ( buf, my_buff, len);
    if (port->monitor_mode)
	    gen_mon_packet(my_buff, len, port, FST_MON_RX);
        //*offp += sizeof ( len );
	fst_dbg(DBG_RX, "%s: read_async_data - Returning %d bytes in buffer of size %d\n", 
	       port->dev->name, len, (int) count);
        kfree (my_buff);
	if (len == 0)
	  return -EAGAIN;
	return len;
}

static ssize_t
fst_read_sync_data (  struct file *file, struct fst_port_info *port, char *buf,
                      size_t count, loff_t *offp )
{
        struct fst_card_info     *usb_dev;
        struct fst_char_rx_frame *rx_frame;
        unsigned long             flags;
        int                       len;
        int                       i;

        usb_dev = port->card;
        /*
         * Any data yet
         */
        if (port->char_inq == NULL)
          {
            fst_dbg(DBG_RX, "<<fst_sync_read: %s No data yet!, %p\n", 
                port->dev->name, port->char_inq);
	    if (file->f_flags & O_NONBLOCK)
              return -EAGAIN; /* Nothing ready */
	    while (port->char_inq == NULL)
              {
                if (wait_event_interruptible(port->readq, 
					     port->char_inq != NULL))
                  return -ERESTARTSYS;
              }
          }
        /*
         * Take the next item off the rx queue
         */
        fst_dbg(DBG_RX, "Taking an rx buffer off the read queue %p\n",
            port->char_inq);
        spin_lock_irqsave(&port->rxf_lock, flags);
        rx_frame = port->char_inq;
        port->char_inq = rx_frame->next;
        spin_unlock_irqrestore(&port->rxf_lock, flags);
        len = rx_frame->frame->len;
        /*
         * Copy any data to user space if there is any
         */
        if (len == 0)
          {
            /*
             * must be a status change message
             */
            dev_kfree_skb(rx_frame->frame);
            kfree(rx_frame);
            return 0;
          }
        else
          {
            if ( len > count)
              i = copy_to_user ( buf, rx_frame->frame->data, count);
            else
              i = copy_to_user ( buf, rx_frame->frame->data, len);
            if ( i < 0 ) 
              {
                fst_dbg( DBG_ASS,"fst_sync_read: Couldn't copy data to user\n");
                dev_kfree_skb(rx_frame->frame);
                kfree(rx_frame);
                return -EFAULT;
              }      
          }
  
        dev_kfree_skb(rx_frame->frame);
        kfree(rx_frame);
        usb_dev->actual_length = 0;
        /*
         * Note that we always return the length of the message
         * If the user didn't supply a buffer big enough then he
         * should notice that len > count and detect the error that way
         */
        return (len);
}

static ssize_t
fst_sync_read ( struct file *file, char *buf, size_t count, loff_t *offp )
{
        struct fst_port_info *port;

	fst_dbg(DBG_RX, "In usb read\n");
	port = (struct fst_port_info *)file->private_data;
        if (port == NULL)
        {
	  fst_dbg (DBG_ASS, "fst_sync_read: Device gone\n");
	  return -ENODEV;
        }
	/*
	 * Are we reading async or sync data?
	 */
	if (port->mode == FST_MODE_ASYNC)
	  {
	    if (port->usb_config.flowControl > COM_FLOW_CONTROL_NONE)
	      check_rx_fifo_threshold(port);
	    return fst_read_async_data (file, port, buf, count, offp);
	  }
	else
	  {
	    return fst_read_sync_data (file, port, buf, count, offp);
	  }
}

static ssize_t
fst_sync_write ( struct file *file, const char *buf, size_t count, loff_t *offp )
{
	struct fst_port_info *port;
        struct sk_buff       *skb;
        int                   retval;

        port = file->private_data;
        if (port == NULL)
        {
	  printk ("fst_sync_write: Device gone\n");
	  return -ENODEV;
        }
	fst_dbg ( DBG_TX,"fst_sync_write: %s %d\n", port->dev->name, (int)count );

        if (count == 0)
          {
            printk("%s: Asked to transmit zero bytes\n", port->dev->name);
          }
        /* Make sure we can both read and write it */
#ifndef RHEL_RELEASE_CODE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)
        if ( ! access_ok ( buf, count )) 
#else
        if ( ! access_ok ( VERIFY_READ, buf, count )) 
#endif
#elif (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(8,1)))
        if (!access_ok(buf, count)) 
#else
        if (!access_ok(VERIFY_READ, buf, count))
#endif
          {
            fst_dbg(DBG_ASS, "fst_sync_write: Can't access write buffer\n");
            return -EFAULT;
          }
	/*
	 * No special treatment for async mode, unlike the farsync driver
	 */
	while (fst_get_tx_queue_len(port) >= fst_txq_high)
	  {
	    //printk("%s: Couldn't accept write of %d bytes, no room in txq %d\n",
	    //   port->dev->name, count, fst_get_tx_queue_len(port));
	    if (file->f_flags & O_NONBLOCK)
	      {
		return -EWOULDBLOCK;
	      }
	    //printk("Write thread sleeping\n");
	    if (wait_event_interruptible(port->writeq, fst_get_tx_queue_len(port) < fst_txq_high))
	      {
		printk("Write interrupted\n");
		return -ERESTARTSYS;
	      }
	    //printk("Write thread awake again %d\n", fst_get_tx_queue_len(port));
	  }
	//printk("Proceeding with the write\n");
        if ( (skb = dev_alloc_skb ( count )) == NULL )
          {
            fst_dbg(DBG_INTR,"fst_sync_write: can't allocate skb\n");
            return -EFAULT;
          }
        /* Read it in */
        if (__copy_from_user ( skb_put(skb, count), buf, count ))
          {
            dev_kfree_skb(skb);
            return -EFAULT;
          }
	fst_dbg(DBG_TX, "%s: Sending %d bytes\n", port->dev->name, (int) count);
        retval = fst_start_xmit( skb, port->dev);
        if ((retval == 0) && (file->f_flags & O_NONBLOCK))
          {
            return -EWOULDBLOCK;
          }
        return retval;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,1,0)
static ssize_t
fst_sync_readv ( struct kiocb *iocb, struct iov_iter *iov_iter)
{
  struct file *file = iocb->ki_filp;
  unsigned long count = iov_iter->nr_segs;
  struct iovec *iov = (struct iovec *) iov_iter->iov;
  loff_t *ppos=NULL;
#else
static ssize_t
fst_sync_readv ( struct kiocb *iocb, const struct iovec *iov,
                 unsigned long count, loff_t my_pos)
{
  struct file *file = iocb->ki_filp;
  loff_t pos = my_pos;
  loff_t *ppos = &pos;
#endif
#else
static ssize_t
fst_sync_readv ( struct file *file, struct iovec *iov,
                 unsigned long count, loff_t *ppos)
{
#endif
  struct fst_port_info     *port;
  int total_read = 0;
  int i;
  size_t ret_val = 0;
  int copy_result = 0;
  size_t *length_ptr = NULL;
  int start_buffer = 0;
  int buffer_offset = 0;
  int num_buffers = count;
  size_t buffer_adjust = 0;

  port = file->private_data;
  if(!port)
    {
      printk("fst_sync_readv: port has gone away\n");
      return -ENODEV;
    }

  fst_dbg (DBG_RX, "In fst_sync_readv: %lu buffers provided\n", count);
  /*
   * Some checking to do depending on the readv mode set in the port 
   * structure
   */
  if (port->readv_mode == FST_READV_SYNC2)
    {
      /* 
       *This mode is going to user the first buffer in the vector
       * to store the lengths of each of the subsequent buffers
       * so first check is that there is at least 2 buffers
       */
      if (count < 2)
        {
          fst_dbg(DBG_ASS, "Need at least 2 buffers for SYNC2 mode (%ld)\n", 
                  count);
          return -ENOBUFS;
        }
      start_buffer = 1;
      num_buffers--; /* one is for the buffer lengths */
      /*
       * The length of the first buffer must be large enough to contain
       * the lengths of all the subsequent ones
       */
      if (iov[0].iov_len < (count-1) * sizeof(size_t))
        {
          fst_dbg(DBG_ASS, "First buffer needs to be at least %ld bytes, but is only %d bytes\n",
                  (count-1) * sizeof(size_t), (int)(iov[0].iov_len));
          return -ENOBUFS;
        }
    }
  if (port->readv_mode == FST_READV_SYNC)
    {
      /* 
       * The buffer lengths will be the first few bytes of the buffer
       */
      buffer_offset = sizeof(size_t);
      buffer_adjust = sizeof(size_t);
    }
  /*
   * In it's simplest form, this function can just keep calling the read 
   * function until there is nothing left.
   */
  for (i=0; i< num_buffers; i++)
    {
      fst_dbg(DBG_RX, "call %d fst_sync_read with a buffer of %d bytes\n", 
             i, (int) (iov[0].iov_len));
      ret_val = fst_sync_read(file, iov[start_buffer+i].iov_base+buffer_offset,
                              iov[start_buffer+i].iov_len-buffer_adjust, 
                              ppos);
      fst_dbg(DBG_RX, "fst_sync_read returning %d\n", (int)ret_val);
      if (ret_val == -EAGAIN)
        {
          /*
           * no more to read.  Break out of the loop
           */
          fst_dbg(DBG_RX, "readv loop terminating, i=%d\n", i);
          if (total_read == 0)
            total_read = ret_val;
          break;
        }
      /*
       * Check for any other sort of error
       */
      if (ret_val < 0)
        {
          printk("Unexpected error in readv %d\n", (int)ret_val);
          return ret_val;
        }
      /*
       * If in one of the Sync modes, set the lengths in the designated 
       * place.
       */
      if (port->readv_mode == FST_READV_SYNC)
        {
          /*
           * The length is at the start of the buffer
           */
          length_ptr = iov[i].iov_base;
          copy_result = copy_to_user(length_ptr, &ret_val, sizeof(size_t));
          if ( copy_result < 0 ) 
              {
                fst_dbg( DBG_ASS,"fst_sync_readv: Couldn't copy data lengths to user\n");
                return EIO;
              }
        }
      if (port->readv_mode == FST_READV_SYNC2)
        {
          length_ptr = iov[0].iov_base+(i*sizeof(size_t));
          copy_result = copy_to_user(length_ptr, &ret_val, sizeof(size_t));
          if ( copy_result < 0 ) 
              {
                fst_dbg( DBG_ASS,"fst_sync_readv: Couldn't copy data lengths to user\n");
                return EIO;
              }
        }
      total_read += ret_val;
    }
  /*
   * Set the buffer length to zero on the first unused buffer if in
   * SYNC mode
   */
  ret_val = 0;
  if ( (i <= num_buffers) && (port->readv_mode == FST_READV_SYNC))
    {
      /*
       * The length is at the start of the buffer
       */
      length_ptr = iov[i].iov_base;
      copy_result = copy_to_user(length_ptr, &ret_val, sizeof(size_t));
      if ( copy_result < 0 ) 
        {
          fst_dbg( DBG_ASS,"fst_sync_readv: Couldn't copy data lengths to user\n");
          return EIO;
        }
    }
  if ((i <= num_buffers) && (port->readv_mode == FST_READV_SYNC2))
    {
      length_ptr = iov[0].iov_base+(i*sizeof(size_t));
      copy_result = copy_to_user(length_ptr, &ret_val, sizeof(size_t));
      if ( copy_result < 0 ) 
        {
          fst_dbg( DBG_ASS,"fst_sync_readv: Couldn't copy data lengths to user\n");
          return EIO;
        }
    }
  fst_dbg(DBG_RX, "readv returning %d bytes read in %d buffers\n",
            total_read, i);
  return total_read;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,1,0)
static ssize_t
fst_sync_writev (struct kiocb *iocb , struct iov_iter *iov_iter)
{
  struct file *file = iocb->ki_filp;
  unsigned long count = iov_iter->nr_segs;
  struct iovec *iov = (struct iovec *) iov_iter->iov;
  loff_t *ppos=NULL;
#else
static ssize_t
fst_sync_writev (struct kiocb *iocb , const struct iovec *my_iov,
                 unsigned long count, loff_t my_pos)
{
  struct file *file = iocb->ki_filp;
  loff_t pos = my_pos;
  loff_t *ppos = &pos;
  struct iovec *iov = my_iov;
#endif
#else
static ssize_t
fst_sync_writev (struct file *file, struct iovec *iov,
                 unsigned long count, loff_t *ppos)
{
#endif
  struct fst_port_info     *port;
  int total_written = 0;
  int ret_val = 0;
  int i;

  fst_dbg(DBG_TX, "In fst_sync_writev for %lu buffers\n", count);
  port = file->private_data;
  if(!port)
    {
      printk("fst_sync_writev: port has gone away\n");
      return -ENODEV;
    }

  /*
   * This can simple be a case of repeated calle to the write function
   * Just need to check for error conditions, or a return of zero
   * indicating no more room
   */

  for (i=0; i<count; i++)
    {
      fst_dbg(DBG_TX, "call %d fst_sync_write with a buffer of %d bytes\n", 
             i, (int)(iov[0].iov_len));
      ret_val = fst_sync_write(file, iov[i].iov_base, iov[i].iov_len, ppos);
      fst_dbg(DBG_TX, "fst_sync_write returning %d\n", ret_val);
      if (ret_val == -EAGAIN)
        {
          /*
           * no more space to write.  Break out of the loop
           */
          fst_dbg(DBG_ASS, "writev loop terminating, i=%d\n", i);
          if (total_written == 0)
            total_written = ret_val;
          iov[i].iov_len = ret_val;
          break;
        }
      /*
       * Check for any other sort of error
       */
      if (ret_val < 0)
        {
          printk("Unexpected error in writev %d\n", ret_val);
          return ret_val;
        }
      /*
       * Set the lengths written into each vector and sum up the total so far
       */
      iov[i].iov_len = ret_val;
      total_written += ret_val;
    }
  fst_dbg(DBG_TX, "writev returning %d bytes written\n", total_written);
  return total_written;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
static int
fst_sync_ioctl ( struct inode *inode, struct file *file, unsigned int cmd,
                 unsigned long arg )
#else
static long
fst_sync_ioctl ( struct file *file, unsigned int cmd,
                 unsigned long arg )
#endif
{
        struct fst_port_info     *port;
	int retval;
        struct termios terminal;
        unsigned int baud;
        unsigned int stop_bits;
        unsigned int char_size;
        unsigned int parity = 0;
	unsigned int rtscts;
	unsigned int xonxoff;
	unsigned int signals = 0;
	unsigned int outputs = 0;
	struct ifreq user_ifr;
	struct ifreq *user_ifr_ptr = &user_ifr;

        port = file->private_data;
	if(!port)
	{
	  printk("fst_sync_ioctl: port has gone away\n");
	  return -ENODEV;
	}
	fst_dbg( DBG_IOCTL,"fst_sync_ioctl: %d, %lx\n", cmd-SIOCDEVPRIVATE, arg );
        /*
         * Check for an async configuration request first
         */
	switch (cmd)
	  {
	  case TCGETS:
	    fst_dbg(DBG_IOCTL, "%s: TC Get Serial\n", port->dev->name);
            if ( copy_to_user ( (char *) arg, &fst_serial_termios[port->index], 
                                sizeof(struct termios)))
            {
                    return -EFAULT;
            }
            fst_dbg(DBG_IOCTL, "%s: Returning %x\n", port->dev->name, 
                    fst_serial_termios[port->index].c_cflag);
            return 0;
	    break;

	  case TCSETS:
	  case TCSETSW:
	    fst_dbg(DBG_IOCTL, "%s: TC Set Serial\n", port->dev->name);
            if (copy_from_user(&terminal, (char *) arg, sizeof(struct termios)))
            {
                    return -EFAULT;
            }
            fst_dbg(DBG_IOCTL, "%s: Setting %x\n", port->dev->name, terminal.c_cflag);
            memcpy(&fst_serial_termios[port->index], &terminal, 
                    sizeof(struct termios));
            baud = get_baud(terminal.c_cflag);
            char_size = get_char_size(terminal.c_cflag);
            stop_bits = get_stop_bits(terminal.c_cflag);
            parity = get_parity(terminal.c_cflag);
	    rtscts = get_rtscts(terminal.c_cflag);
	    xonxoff = get_xonxoff(terminal.c_iflag);
            /*
	     * If we are in async mode we need to configure the 
	     * async parameters
	     */
	    if (port->mode == FST_MODE_ASYNC)
	      {
                configure_async_parameters(port, baud, char_size, stop_bits, 
                                           parity, rtscts|xonxoff);
    	      }
	    break;

	  case TCSETSF:
	    fst_dbg(DBG_IOCTL, "TC Set Serial F\n");
	    break;
	  case TIOCGSERIAL:
	    fst_dbg(DBG_IOCTL, "Get serial config\n");
	    break;
	  case TIOCSSERIAL:
	    fst_dbg(DBG_IOCTL, "Set Serial Config\n");
	    break;
	  case TIOCMGET:
            fst_dbg(DBG_IOCTL, "Get Control Signals\n");
	    signals = port->v24OpSts;
	    if (signals & OPSTS_RTS)
	      outputs |= TIOCM_RTS;
	    if (signals & OPSTS_DTR)
	      outputs |= TIOCM_DTR;
	    fst_issue_cmd (port, GETSIGNALS);
	    signals = port->v24IpSts;
	    if (signals & IPSTS_CTS)
	      outputs |= TIOCM_CTS;
	    if (signals & IPSTS_DSR)
	      outputs |= TIOCM_DSR;
	    if (signals & IPSTS_DCD)
	      outputs |= TIOCM_CD;
	    fst_dbg(DBG_IOCTL, "Returning signals as %x\n", outputs);
            if ( copy_to_user ( (unsigned int *) arg, &outputs, 
                                sizeof(unsigned int)))
	      {
		return -EFAULT;
	      }
	  break;
	  case TIOCMBIS:
	    fst_dbg(DBG_IOCTL, "TIOCMBIS - ?\n");
	    break;
	  case TIOCMBIC:
	    fst_dbg(DBG_IOCTL, "TIOCMBIC- ?\n");
	    break;
	  case TIOCMSET:
            fst_dbg(DBG_IOCTL, "Set control signals\n");
            if (copy_from_user(&signals, (unsigned int *) arg, sizeof(unsigned int)))
            {
                    return -EFAULT;
            }
            if (signals & TIOCM_RTS)
              {
                fst_dbg(DBG_IOCTL, "Set RTS\n");
                outputs = OPSTS_RTS | port->v24OpSts;
                port->v24OpSts = outputs;
                if ( port->run )
                  fst_issue_cmd ( port, SETV24O );
              }
            else
              {
                fst_dbg(DBG_IOCTL, "Clear RTS\n");
                outputs = port->v24OpSts  & ~OPSTS_RTS;
                port->v24OpSts = outputs;
                if ( port->run )
                  fst_issue_cmd ( port, SETV24O );
              }
            if (signals & TIOCM_DTR)
              {
                fst_dbg(DBG_IOCTL, "Set DTR\n");
                outputs = OPSTS_DTR | port->v24OpSts;
                port->v24OpSts = outputs;
                if ( port->run )
                  fst_issue_cmd ( port, SETV24O );
              }
            else
              {
                fst_dbg(DBG_IOCTL, "Clear DTR\n");
                outputs = port->v24OpSts & ~OPSTS_DTR;
                port->v24OpSts = outputs;
                if ( port->run )
                  fst_issue_cmd ( port, SETV24O );
              }
            fst_dbg(DBG_IOCTL, "Signals are now %x\n", outputs);
            break;
	  case TCFLSH:
	    fst_dbg(DBG_IOCTL, "TCFLSH - ?\n");
	    break;
	  default:
            /*
             * Or maybe it is one of our private ones
             */
	    if (copy_from_user(user_ifr_ptr, (struct ifreq *)arg, sizeof( struct ifreq )))
	      {
		fst_dbg(DBG_ASS, "fst_sync_ioctl: could not read user buffers\n");
		return -EFAULT;
	      }
            retval = fst_ioctl(port->dev, user_ifr_ptr, cmd);
	    if (copy_to_user((struct ifreq *)arg, user_ifr_ptr, 
			     sizeof( struct ifreq )))
	      {
		fst_dbg(DBG_ASS, "fst_sync_ioctl: could not write user buffers\n");
		return -EFAULT;
	      }
	    fst_dbg(DBG_IOCTL, "fst_sync_ioctl: copying user ifr to user space\n");
            return retval;
	  }
        return 0;
}

/* New version for handling 32-bit applications on a 64-bit machine */
#ifdef HAVE_COMPAT_IOCTL
static long
fst_sync_compat_ioctl (struct file *file, unsigned int cmd,
					unsigned long arg )
{
	struct fst_port_info *port;
	struct net_device    *dev;

	static struct fstioc_info_32    my_info_32;	// GETCONF, SETCONF, FSTGETSHELL
	static struct fstioc_info          my_info;	// GETCONF, SETCONF, FSTGETSHELL
	unsigned int           fstioc_info_size_32;    	// GETCONF, SETCONF, FSTGETSHELL

	static struct fstioc_status_32 my_state_32;	// FSTGSTATE
	static struct fstioc_status       my_state;	// FSTGSTATE
	unsigned int               my_state_32_len;	// FSTGSTATE

	static struct fstioc_cmd_32      my_cmd_32;	// FSTCMD's
	static struct fstioc_cmd            my_cmd;	// FSTCMD's

	static struct fst_ints_32   my_fst_ints_32;	// FSTCMD_GET_INTS
	static struct fst_ints         my_fst_ints;	// FSTCMD_GET_INTS

	struct ifreq uifr;
	struct ifreq *ifr = (struct ifreq *)arg;
	struct ifreq my_ifr;
	void *uptr=NULL;

	int retval=0;

	fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: command %x, arg %p\n", cmd, (char *)arg);
        
	memset(&uifr, 0, sizeof(uifr));

	if (ifr != NULL)
		{
		  if (copy_from_user(&uifr, ifr, sizeof(uifr)))
		    {
		      fst_dbg(DBG_ASS, "fst_compat_ioctl: Error in copying the arg structure\n");
		      return -EFAULT;
		    }
		  ifr = &uifr;
		}

	port = file->private_data;
	dev = port_to_dev(port);
	// sometimes the top 32 bits are (unexpectedly) set and it all goes badly wrong, so mask 'em out)
	if(ifr!=NULL)
		uptr = (void *)((unsigned long)(ifr->ifr_data) & 0xFFFFFFFF);

	if(port->compat == 1)
	{
		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: Unexpected port->compat value %d\n", port->compat);
	}
	switch(cmd) {

	// destined for fst_sync_ioctl
	case TCGETS:
	case TCSETS:
	case TCSETSW:
	case TCSETSF:
	case TIOCGSERIAL:
	case TIOCSSERIAL:
	case TIOCMGET:
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
	case TCFLSH:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
		retval = fst_sync_ioctl(NULL, file, cmd, arg);  // nothing to do here so just pass it on
#else
		retval = fst_sync_ioctl(file, cmd, arg);  // nothing to do here so just pass it on
#endif
	break;

	// destined for fst_ioctl
	case FSTCMD:
		/* FSTCMD's arrive in the form of a 32-bit fstioc_cmd_32 structure so we have create a 64-bit
		 * equivalent (fstioc_cmd) before we pass the command onto fst_ioctl.
		 *
		 * On return, fst_ioctl will have updated the status field, so we copy that back to the 32-bit
		 * version of the structure before we copy the whole command back to the user.
		 */

		port->compat=1;
		my_ifr.ifr_data = (char *)&my_cmd;	

		if (copy_from_user(&my_cmd_32, uptr, sizeof( my_cmd_32 )))
		{
			fst_dbg(DBG_ASS, "fst_sync_compat_ioctl: could not access user buffers\n");
			return -EFAULT;
		}
		
		my_cmd.version			= my_cmd_32.version;
		my_cmd.command			= my_cmd_32.command;
		my_cmd.status			= my_cmd_32.status;
		my_cmd.input_data_len		= my_cmd_32.input_data_len;
		my_cmd.output_data_len		= my_cmd_32.output_data_len;
		my_cmd.data_ptr			= (unsigned char *)(long) my_cmd_32.data_ptr;

		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: command 0x%x\n", my_cmd.command);
	        switch (my_cmd.command)
		{

		case FSTCMD_GET_INTS:               
			/* Here, we create a local 64-bit version of the my_fst_ints structure which
			 * fst_ioctl then writes to. We then copy that data back to the 32-bit version and then
			 * copy that data back to the user.
			 */
			my_cmd.data_ptr = (unsigned char *) &my_fst_ints;
			retval = fst_ioctl(dev, &my_ifr, cmd);  

			// Update fields that changed
			my_fst_ints_32.int_186     = (uint32_t)my_fst_ints.int_186;
			my_fst_ints_32.int_txdma   = (uint32_t)my_fst_ints.int_txdma;
			my_fst_ints_32.int_rxdma   = (uint32_t)my_fst_ints.int_rxdma;
			my_fst_ints_32.int_no_work = (uint32_t)my_fst_ints.int_no_work;

			// Copy the data back to user
			if (copy_to_user ( (unsigned char *) (long)my_cmd_32.data_ptr, &my_fst_ints_32, sizeof(struct fst_ints_32)))
			{
 				fst_dbg(DBG_ASS, "fst_sync_compat_ioctl: get_ints: couldn't copy fst_ints\n");
				return -EFAULT;
			}
			/* Need to tweak output_data_len as 32-bit version is only half size of 64-bit version */
			my_cmd_32.output_data_len = sizeof(struct fst_ints_32);			
		break;

		/* For all of these, let fst_ioctl get or set the data fields (if any) */
		case FSTCMD_SET_V24O:
		case FSTCMD_GET_SERIAL:
		case FSTCMD_GET_VERSION:
		case FSTCMD_SET_VERSION:
		case FSTCMD_RESET_INTS:
		case FSTCMD_RESET_STATS:
		case FSTCMD_SET_CHAR:                                    
		case FSTCMD_GET_PRESERVE_SIGNALS:
		case FSTCMD_SET_PRESERVE_SIGNALS:
		case FSTCMD_UPDATE_CLOCK:
		case FSTCMD_SET_CUSTOM_RATE:	
			retval = fst_ioctl(dev, &my_ifr, cmd);  
			/* output_data_len has been updated by driver, so we need to copy it back */
			my_cmd_32.output_data_len = my_cmd.output_data_len;
		break;
		case FSTCMD_SET_READV:  /* We don't support this, so return and inform user of the situation */
			retval = ENOSYS; 
		break;
		default:
			printk("fst_sync_compat_ioctl: unexpected default command: %x\n", my_cmd.command);
		break;
		}//switch(cmd.command)

		/* The status member and output_data_len are the only ones fst_ioctl updates, so just update our
		 * 32-bit version and send the command structure back to the user
		 */		
		my_cmd_32.status = my_cmd.status;
	        
		if ( copy_to_user ( uptr, &my_cmd_32, sizeof(my_cmd_32)))
		{
			fst_dbg(DBG_ASS, "fst_sync_compat_ioctl: couldn't copy status back\n");
			return -EFAULT;
		}  
	break; //FSTCMD

	case FSTGETSHELL:     // drop through
	case FSTGETCONF:
		port->compat=1;
		memset ( &my_info, 0, sizeof ( struct fstioc_info ));

		switch (port->fstioc_info_ver)
		{
			case FST_VERSION_CURRENT:	fstioc_info_size_32 = 436; break;
			case FST_VERSION_V3:		fstioc_info_size_32 = 428; break;
			case FST_VERSION_V2:		fstioc_info_size_32 = 416; break;
			case FST_VERSION_V1:		fstioc_info_size_32 = 408; break;
			default:			fstioc_info_size_32 = 312; break; 	
		}

		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: fstioc_inofo_size_32 is %d\n", fstioc_info_size_32);

		my_ifr.ifr_data = (char *)&my_info; 
		retval = fst_ioctl(dev, &my_ifr, cmd);

		convertTo32(&my_info_32, &my_info);

		if ( copy_to_user ( uptr, &my_info_32, fstioc_info_size_32))
		{
			fst_dbg(DBG_ASS, "fst_sync_compat_ioctl: couldn't copy to user\n");
			return -EFAULT;
		}  
	break;

	case FSTSETCONF:

		port->compat=1;
		memset ( &my_info, 0, sizeof ( struct fstioc_info ));
		// determine right size to copy from user
		switch (port->fstioc_info_ver)
		{
			case FST_VERSION_CURRENT:	fstioc_info_size_32 = 436; break;
			case FST_VERSION_V3:		fstioc_info_size_32 = 428; break;
			case FST_VERSION_V2:		fstioc_info_size_32 = 416; break;
			case FST_VERSION_V1:		fstioc_info_size_32 = 408; break;
			default:			fstioc_info_size_32 = 312; break; 	
		}

		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: FSTSETCONF: fstioc_info_size_32 is %d\n", fstioc_info_size_32);

		if ( copy_from_user ( &my_info_32,  uptr, fstioc_info_size_32))
		{
			fst_dbg(DBG_ASS, "fst_sync_compat_ioctl: FSTSETCONF: failed to copy from user\n");
			return -EFAULT;
		}

		convertTo64(&my_info, &my_info_32);

		my_ifr.ifr_data = (char *)&my_info; 
		retval = fst_ioctl(dev, &my_ifr, cmd);  
	break;

	case FSTGSTATE:

		port->compat=1;
		memset (&my_state, 0, sizeof(struct fstioc_status));

		my_ifr.ifr_data = (char *)&my_state; 
		retval = fst_ioctl(dev, &my_ifr, cmd);  

		my_state_32.carrier_state = (uint32_t)my_state.carrier_state;  
		my_state_32.txq_length    = (uint32_t)my_state.txq_length;  
		my_state_32.rxq_length    = (uint32_t)my_state.rxq_length;  

		deviceStatsTo32(&my_state_32.stats, &my_state.stats);

		// check how much to send back
		if (port->notify_mode == FST_NOTIFY_EXTENDED)
			my_state_32_len = sizeof (struct fstioc_status_32);
		else
			my_state_32_len = FST_NOTIFY_BASIC_SIZE;
		
		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: FSTGTSTATE: length is %d\n", my_state_32_len);

		if(copy_to_user (uptr, &my_state_32, my_state_32_len))
		{
		    return -EFAULT;
		}
	break;

	case FSTWRITE:
	case FSTSNOTIFY:
	case FSTSETMON:
	case FSTSETPORT:
	case FSTSYSREQ:
	case FSTCPURESET:
	case FSTCPURELEASE:
		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: calling fst_ioctl with cmd %x\n", cmd);
		retval = fst_ioctl(dev, ifr, cmd);  // nothing to do here so just pass it on
	break;	
	       
	default:
		fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: Unexpected default cmd %x\n", cmd);
	} //switch(cmd)

	fst_dbg(DBG_IOCTL, "fst_sync_compat_ioctl: Leaving with cmd:%x\n", cmd);
	port->compat=0;  //set this back to zero
	return retval;
}
#endif /* HAVE_COMPAT_IOCTL */

int
fst_sync_async ( int fd, struct file *file, int on )
{
  fst_dbg ( DBG_ASS,"fst_sync_async: %d\n", on );
  return 0;
}

unsigned int
fst_sync_poll ( struct file *file, poll_table *poll_table )
{
  int                   poll_mask = 0;
  int                   txq_length = 0;
  struct fst_port_info *port;
  struct fst_card_info *usb_dev;
  unsigned long         flags;

  port = file->private_data;
  if (port == NULL)
    {
      printk ("fst_sync_poll: Device gone\n");
      return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
    }
  usb_dev = port->card;
  fst_dbg(DBG_INTR, ">>fst_sync_poll: %s\n", port->dev->name);
  poll_wait( file, &port->pollq, poll_table );
  /*
   * Any exceptions to report?
   */
  if (port->exception)
    {
      poll_mask |= POLLRDBAND | POLLPRI;
      port->exception = 0;
    }
  /*
   * Is there something to read?
   */
  if (port->mode == FST_MODE_ASYNC)
    {
      if (port->fifo_rxdata == NULL)
        {
          printk("There is no rx fifo???\n");
          return 0;
        }
      if (port->fifo_rxdata->read_idx != port->fifo_rxdata->write_idx)
	poll_mask |= POLLIN | POLLRDNORM;
    }
  else
    {
      if (port->char_inq)
        poll_mask |= POLLIN | POLLRDNORM;
    }
  /*
   * Is the txq depth below the low water mark?
   */
  spin_lock_irqsave ( &port->card->card_lock, flags );
  if ((txq_length = port->txqe - port->txqs) < 0)
    {
      /*
       * This is the case where one has wrapped and the
       * maths gives us a negative number
       */
      txq_length = txq_length + FST_TXQ_DEPTH ;
    }
  spin_unlock_irqrestore ( &port->card->card_lock, flags );
  if (txq_length < fst_txq_low)
    {
      poll_mask |= POLLOUT | POLLWRNORM;
    }
  fst_dbg( DBG_INTR, "<<fst_sync_poll: %s 0x%08x returning 0x%04x\n",
        port->dev->name, (unsigned int)(long)poll_table, poll_mask );
  return poll_mask;
}

static int char_check_rx_queue_length (struct fst_port_info *port)
{
  int queue_length = 1;
  struct fst_char_rx_frame *q_start;
  /*
   * Parse the rx queue to see how many are on it
   * It should be at least 1
   */
  q_start = port->char_inq;
  if (q_start == NULL)
    {
      fst_dbg(DBG_TTY, "Nothing on Char Rx Queue\n");
      return 0;
    }
  if (q_start == port->char_inq_end)
    {
      fst_dbg(DBG_TTY, "Returning char queue length of 1\n");
      return 1;
    }
  while (q_start != port->char_inq_end)
    {
      queue_length++;
      q_start = q_start->next;
    }
  fst_dbg(DBG_TTY, "Returning char queue length of %d\n", queue_length);
  return queue_length;
}

static int char_purge_one (struct fst_port_info *port)
{
  struct fst_char_rx_frame *rx_frame;
  /*
   * Purge the first frame on the rx queue
   */

  fst_dbg(DBG_TTY, "Purging one frame from the char rx queue\n");
  rx_frame = port->char_inq;
  port->char_inq = rx_frame->next;
  dev_kfree_skb(rx_frame->frame);
  kfree(rx_frame);
  return 0;
}

static int
char_queue_async_rx ( struct fst_port_info *port, char * rx_buff, int len)
{
  unsigned long flags;
#ifdef FST_TTY_CHAR
  st_fst_tty_area *fst_tty;
#endif

  fst_dbg(DBG_RX, "%s: fst_intr_async_rx of %d bytes\n", port->dev->name, len);
#ifdef FST_TTY_CHAR	
        fst_tty = &fst_tty_area[port->minor_dev_no];
        fst_tty->len = len;	  	
	fst_tty->data = rx_buff;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	  schedule_work(&(fst_tty->fst_async_tty_work));
#else
	  schedule_work((void *) (&(fst_tty->fst_async_tty_work)));
#endif
#else
	spin_lock_irqsave(&port->card->fifo_lock, flags);
	write_into_fifo((char *) port->fifo_rxdata, FIFO_PORT_RX, 
                        rx_buff, len);
	spin_unlock_irqrestore(&port->card->fifo_lock, flags);
	if (port->usb_config.flowControl > COM_FLOW_CONTROL_NONE)
          check_rx_fifo_threshold(port);
	fst_dbg(DBG_RX, "%s: Written %d bytes to port rx fifo\n", 
		port->dev->name, len);
        wake_up_interruptible (&port->pollq);
        wake_up_interruptible (&port->readq);

        port_to_stats(port,rx_packets)++;
	port_to_stats(port,rx_bytes) += len;
        fst_dbg(DBG_RX, "Placed data in fifo and stimmed the poll function\n");
#endif
  return 0;
}

static int
char_queue_rx_frame ( struct fst_port_info *port, struct sk_buff *skb)
{
  struct fst_char_rx_frame *rx_frame;
  unsigned long         flags;
  int                       queue_len;

  /*
   * Allocate a queue structure and queue the skb on the ports char interface
   * Note that the flags are GFP_ATOMIC as this could be called from an
   * interrupt.
   */
  fst_dbg(DBG_RX, "char_queue_rx_frame: %s\n", port->dev->name);
  rx_frame = kmalloc(sizeof(struct fst_char_rx_frame), GFP_ATOMIC);
  if (rx_frame == NULL)
    {
      fst_dbg(DBG_ASS,"Could not allocate an fst_char_rx_structure\n");
      return NET_RX_DROP;
    }
  rx_frame->frame = skb;
  spin_lock_irqsave(&port->rxf_lock, flags);
  if (port->char_inq == NULL)
    {
      /*
       * Nothing on the queue yet
       */
      fst_dbg(DBG_RX, "Queue was empty\n");
      port->char_inq = rx_frame;
    }
  else
    {
      fst_dbg(DBG_RX, "Queued at end\n");
      port->char_inq_end->next = rx_frame;
    }
  port->char_inq_end = rx_frame;
  rx_frame->next = NULL;
  rx_frame->prev= NULL;
  queue_len = char_check_rx_queue_length(port);
  if ((port->char_inq_threshold) && (queue_len > port->char_inq_threshold))
    {
      if (port->notify_mode == FST_NOTIFY_EXTENDED)
	{
	  /*
	   * Tell the application the queue of frame waiting is getting
	   * a tad long
	   */
	  fst_notify_state_change ( port->card, port );
	}
    }
  if (queue_len > port->char_inq_max_len)
    {
      /*
       * Something is really wrong.  Perhaps the application doesn't
       * read, so remove the oldest frame
       */
      printk("%s: Char Rx queue depth reached configured maximum (%d), oldest frame purged\n",
		 port->dev->name, port->char_inq_max_len);
      char_purge_one(port);
      port_to_stats(port,rx_dropped)++;
    }
  spin_unlock_irqrestore(&port->rxf_lock, flags);
  wake_up_interruptible (&port->pollq);
  wake_up_interruptible (&port->readq);
  fst_dbg(DBG_RX, "Placed skb on read queue and stimmed the poll function\n");
  return 0;
}

/* Card memory image access */
static struct file_operations fst_mem_fops = {
	read:		fst_sync_read,
	write:		fst_sync_write,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,1,0)
	read_iter:      fst_sync_readv,
	write_iter:     fst_sync_writev,

#else
        aio_read:       fst_sync_readv,
        aio_write:      fst_sync_writev,
#endif
#else
        readv:          fst_sync_readv,
        writev:         fst_sync_writev,
#endif
#ifdef HAVE_UNLOCKED_IOCTL
	unlocked_ioctl:	fst_sync_ioctl,
#else
	ioctl:		fst_sync_ioctl,
#endif
	open:		fst_sync_open,
	release:	fst_sync_release,
	fasync:		fst_sync_async,
	poll:           fst_sync_poll,
#ifdef HAVE_COMPAT_IOCTL	
	compat_ioctl:	fst_sync_compat_ioctl,
#endif	
};
#endif /* end ifdef FST_TTY_CHAR */
/*      Control signal change interrupt event
 */
static void
fst_intr_ctlchg ( struct fst_card_info *card, struct fst_port_info *port, 
		  int signals )
{
        if ( signals & (((port->hwif == X21) || (port->hwif == X21D)) 
			? IPSTS_INDICATE : IPSTS_DCD ))
        {
	        if ( (!netif_carrier_ok(port_to_dev(port))) && (!port->ignoreCarrier))
                {
                  fst_dbg ( DBG_ASS,"%s: DCD active\n", port->dev->name);
                  netif_carrier_on ( port_to_dev(port) );
                  fst_notify_state_change (card, port);
                }
        }
        else
        {
	        if ( (netif_carrier_ok(port_to_dev(port))) && (!port->ignoreCarrier))
                {
                  fst_dbg ( DBG_ASS,"%s: DCD lost\n", port->dev->name);
                  netif_carrier_off ( port_to_dev(port) );
                  fst_notify_state_change (card, port);
#ifdef FST_TTY_CHAR
		  if (fst_tty_area[port->minor_dev_no].state == FST_TTY_ST_DATA)
		    tty_hangup(fst_tty_area[port->minor_dev_no].tty);
#endif
                }
        }
}

/*      Log Rx Errors
 */
static void
fst_log_rx_error ( struct fst_card_info *card, struct fst_port_info *port,
		   unsigned char status)
{
        /* 
	 * Increment the appropriate error counter
	 */
        port_to_stats(port,rx_errors)++;
	if (status & RX_FRAM )
	  {
	    port_to_stats(port,rx_frame_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Rx frame error %x\n", 
			 port_to_dev(port)->name, status);
	  }
	if ( status & RX_OFLO )
	  {
	    port_to_stats(port,rx_fifo_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Rx fifo error %x\n", 
			 port_to_dev(port)->name, status);
	  }
	if (status & RX_CRC )
	  {
	    port_to_stats(port,rx_crc_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Rx crc error %x\n",  
			 port_to_dev(port)->name, status);
	  }
	if ( status & RX_ABRT )
	  {
	    port_to_stats(port,rx_length_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Rx hardware error (abort) %x\n", 
			 port_to_dev(port)->name, status);
	  }
	if (port->notify_mode == FST_NOTIFY_EXTENDED)
	  fst_notify_state_change ( port->card, port );
}

/*      Log Tx Errors
 */
static void
fst_log_tx_error ( struct fst_card_info *card, struct fst_port_info *port,
		   unsigned char status)
{
        /* 
	 * Increment the appropriate error counter
	 */
        port_to_stats(port,tx_errors)++;
	if ( status & TX_URUN )
	  {
	    port_to_stats(port,tx_fifo_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Tx underrun error %d\n", 
			 port_to_dev(port)->name, status);
	  }
	if ( status & TX_ABRT )
	  {
	    port_to_stats(port,tx_aborted_errors)++;
	    if (printk_ratelimit())
	      printk_err("%s: Tx abort error %d\n", 
			 port_to_dev(port)->name, status);
	  }
	if (port->notify_mode == FST_NOTIFY_EXTENDED)
	  fst_notify_state_change ( port->card, port );
}


static unsigned char reverse[256] = {
    0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, 0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
    0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
    0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, 0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
    0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, 0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
    0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
    0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, 0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
    0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
    0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, 0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
    0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, 0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
    0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
    0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, 0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
    0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
    0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, 0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
    0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, 0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
    0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
    0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, 0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff};

void fst_reverse_bits (unsigned char * buff, int length)
{
  int i;

  for (i=0; i< length; i++)
  {
    buff[i] = reverse[buff[i]];
  }
  return;
}

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,34)
#define usb_buffer_alloc usb_alloc_coherent
#define usb_buffer_free usb_free_coherent
#endif

static void deallocate_urbs_and_buffers (struct fst_card_info *dev)
{
  /* free up our allocated buffers */
  fst_dbg(DBG_INIT, "Freeing a trace buffer at %p\n", dev->int_in_buffer);
  usb_buffer_free(dev->udev, dev->int_in_size, 
                  dev->int_in_buffer, 
                  dev->int_urb->transfer_dma);
  fst_dbg(DBG_INIT, "Freeing a read buffer at %p\n", dev->bulk_in_buffer);
  usb_buffer_free(dev->udev, MAX_TRANSFER, 
                  dev->bulk_in_buffer, 
                  dev->bulk_urb->transfer_dma);
  fst_dbg(DBG_INIT, "Freeing a write buffer at %p\n", dev->bulk_out_buffer);
  usb_buffer_free(dev->udev, MAX_TRANSFER, 
                  dev->bulk_out_buffer, 
                  dev->tx_bulk_urb->transfer_dma);
  usb_free_urb(dev->int_urb);
  usb_free_urb(dev->bulk_urb);
  usb_free_urb(dev->tx_bulk_urb);
}

static int allocate_urbs_and_buffers (struct fst_card_info *dev)
{
  dev->int_urb = usb_alloc_urb(0, GFP_ATOMIC);
  if (!dev->int_urb) 
    {
      fst_dbg(DBG_ASS, "Could not allocate interrupt urb\n");
      return  -ENOMEM;
    }

  dev->bulk_urb = usb_alloc_urb(0, GFP_ATOMIC);
  if (!dev->bulk_urb)
    {
      fst_dbg(DBG_ASS, "Could not allocate bulk urb\n");
      usb_free_urb(dev->int_urb);
      return  -ENOMEM;
    }
  dev->tx_bulk_urb = usb_alloc_urb(0, GFP_ATOMIC);
  if (!dev->tx_bulk_urb)
    {
      fst_dbg(DBG_ASS, "Could not allocate tx bulk urb\n");
      usb_free_urb(dev->int_urb);
      usb_free_urb(dev->bulk_urb);
      return  -ENOMEM;
    }

  dev->int_in_buffer = usb_buffer_alloc(dev->udev, 
					dev->int_in_size, GFP_ATOMIC, 
					&dev->int_urb->transfer_dma);
  if (!dev->int_in_buffer) {
      usb_free_urb(dev->int_urb);
      usb_free_urb(dev->bulk_urb);
      usb_free_urb(dev->tx_bulk_urb);
      return -ENOMEM;
  }
  fst_dbg(DBG_INIT, "Allocating a trace buffer at %p\n", dev->int_in_buffer);
  dev->bulk_in_buffer = usb_buffer_alloc(dev->udev, 
                                         MAX_TRANSFER, GFP_ATOMIC, 
                                         &dev->bulk_urb->transfer_dma);
  if (!dev->bulk_in_buffer) {
      usb_buffer_free(dev->udev, dev->int_in_size, 
                      dev->int_urb->transfer_buffer, 
                      dev->int_urb->transfer_dma);
      usb_free_urb(dev->int_urb);
      usb_free_urb(dev->bulk_urb);
      usb_free_urb(dev->tx_bulk_urb);
      return -ENOMEM;
  }
  fst_dbg(DBG_INIT, "Allocating a read buffer at %p\n", dev->bulk_in_buffer);  
  dev->bulk_out_buffer = usb_buffer_alloc(dev->udev, 
                                          MAX_TRANSFER, GFP_ATOMIC, 
                                          &dev->tx_bulk_urb->transfer_dma);
  if (!dev->bulk_out_buffer) {
      usb_buffer_free(dev->udev, dev->int_in_size, 
                      dev->int_urb->transfer_buffer, 
                      dev->int_urb->transfer_dma);
      usb_buffer_free(dev->udev, MAX_TRANSFER, 
                      dev->bulk_urb->transfer_buffer, 
                      dev->bulk_urb->transfer_dma);
      usb_free_urb(dev->int_urb);
      usb_free_urb(dev->bulk_urb);
      usb_free_urb(dev->tx_bulk_urb);
      return -ENOMEM;
  }
  fst_dbg(DBG_INIT, "Allocating a write buffer at %p\n", dev->bulk_out_buffer);  
  return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
static void sync_write_bulk_callback(struct urb *urb)
#else
static void sync_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
#endif
{
	struct fst_card_info *dev;
	struct fst_port_info *port;

	dev = (struct fst_card_info *)urb->context;
	port = dev->ports[0];
	dev->tx_urb_cnt++;

	/* sync/async unlink faults aren't errors */
	if (urb->status && 
	    !(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET ||
	      urb->status == -ESHUTDOWN)) {
		printk("%s - nonzero write bulk status received: %d\n",
		    __FUNCTION__, urb->status);
	}

	if ( urb != dev->tx_bulk_urb )
	  {
	    printk("Completion urb is not current outstanding urb\n");
	  }
	/*
	 * Inc stats
	 */
	if (urb->actual_length)
	  {
	    port_to_stats(port, tx_packets)++;
	    port_to_stats(port, tx_bytes) += (urb->actual_length -8);
	  }
	/*
	 * Schedule another tx attempt
	 */
	if ((port->run) && (!urb->status))
	  {
            fst_q_work_item(&fst_work_txq, dev->card_no);
            tasklet_schedule(&fst_tx_task); 
          }
	else
	  {
	    fst_dbg(DBG_TX, "Not scheduling another send, port closed\n");
	  }
	fst_dbg(DBG_TX, "Finished write bulk callback\n");
}

static void decode_header(char *buff, struct fst_card_info *card, 
			  struct fst_port_info *port)
{
  struct _FLEX_TXRX_PREAMBLE *header;
  unsigned char status_ok_mask = 0xa0;

  header = (struct _FLEX_TXRX_PREAMBLE *)buff;
  flex_txrx_preamble_to_cpu(header);
  if (header->port &1)
    {
      /*
       * Ignore port B status change messages for now
       * They shouldn't happen in the normal course of events
       */
      fst_dbg(DBG_ASS, "%s: Status message for port B %x\n", port_to_dev(port)->name, 
	     header->port);
      return;
    }
  header->port = header->port & 0x0f0;
  switch (header->port)
    {
    case RXDATA_STATUS:
      {
        if (port->mode == FST_MODE_TRANSPARENT)
          status_ok_mask = 0x80;
        if (header->status ^= status_ok_mask)
          {
            printk_err("%s: Rx Data status is %x should be %x\n", 
                       port_to_dev(port)->name, header->status,
                       status_ok_mask);
            fst_log_rx_error ( card, port, header->status^status_ok_mask);
          }
        break;
      }
    case TXDATA_STATUS:
      {
        if (header->status)
          {
            printk_err("%s: Tx Data status is %x\n", 
                     port_to_dev(port)->name, header->status);
            fst_log_tx_error ( card, port, header->status);
          }
        break;
      }
    case SIGNAL_CHANGE:
      {
        printk_err("%s: Signal Change Status status is %x\n", 
                port_to_dev(port)->name, header->status);
	fst_intr_ctlchg ( card, port, header->status );
        break;
      }
    case TXRX_ERROR:
      {
	if (printk_ratelimit())
	  printk_err("%s: TxRx error signalled - header status is %x\n", 
		     port_to_dev(port)->name, header->status);
	fst_q_work_item(&fst_pstatsq, port->card->card_no);
	schedule_work(&port_stats_work_q); 
        break;
      }
    default:
      printk_err("%s: Unknown preamble status %x\n", 
              port_to_dev(port)->name, header->port);
    }
}

/*
 * Mark it for our own raw sockets interface
 */
static __be16 farsync_type_trans(struct sk_buff *skb, struct net_device *dev)
{
	skb->dev = dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
            skb_reset_mac_header(skb);
#else
	skb->mac.raw = skb->data;
#endif
	skb->pkt_type = PACKET_HOST;
	return htons(ETH_P_CUST);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
static void sync_read_bulk_callback(struct urb *urb)
#else
static void sync_read_bulk_callback(struct urb *urb, struct pt_regs *regs)
#endif
{
#ifdef FST_TTY_CHAR
	st_fst_tty_area *fst_tty;
#endif
	struct fst_card_info *dev;
	struct fst_port_info *port;
	int rx_status = NET_RX_SUCCESS;
	__u16 *len_ptr;

	dev = (struct fst_card_info *)urb->context;
	port = dev->ports[0];

	/* sync/async unlink faults aren't errors */
	if (urb->status && 
	    !(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET)) {
		printk("%s - nonzero read bulk status received: %d\n",
		    __FUNCTION__, urb->status);
		return;
	}

	len_ptr = (u16 *) &(dev->bulk_in_buffer[4]);
	dev->actual_length = le16_to_cpu(*len_ptr); 
#if 0
	printk("Length of data in hdr is %d\n", dev->actual_length);
	printk("Length of data from urb is %d\n", urb->actual_length);
	printk("Header is %x %x %x %x %x %x %x %x\n",
	       dev->bulk_in_buffer[0], dev->bulk_in_buffer[1], 
	       dev->bulk_in_buffer[2], dev->bulk_in_buffer[3],
	       dev->bulk_in_buffer[4], dev->bulk_in_buffer[5], 
	       dev->bulk_in_buffer[6], dev->bulk_in_buffer[7]);
#endif
        decode_header(dev->bulk_in_buffer, dev, port);
	if (dev->actual_length != urb->actual_length-8)
	  {
	    fst_dbg(DBG_RX, "Urb reports different length to header %d %d\n",
		   urb->actual_length, dev->actual_length);
	    //post_a_read_bulk_urb(dev);  /* Post a read for data */
	    //return;
	  }
	if (dev->actual_length > port->dev->mtu + MAX_PPP_HEADER)
	  {
	    printk("%s: Received frame longer %d than MTU %d\n", 
		   port->dev->name, dev->actual_length, port->dev->mtu);
	    post_a_read_bulk_urb(dev);  /* Post a read for data */
	    return;
	  }
	if (port->mode == FST_MODE_ASYNC)
          {
            if (dev->actual_length)
#ifdef FST_TTY_CHAR
            { 
              fst_tty = &fst_tty_area[port->minor_dev_no];
              fst_tty->len = dev->actual_length;	  	
	      fst_tty->data = &dev->bulk_in_buffer[8];    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	      schedule_work(&(fst_tty->fst_async_tty_work));
#else
	      schedule_work((void *) (&(fst_tty->fst_async_tty_work)));
#endif	      
	    }
#else
              rx_status = char_queue_async_rx(port, &dev->bulk_in_buffer[8],
                                              dev->actual_length );
#endif
	    post_a_read_bulk_urb(dev);  /* Post a read for data */
	    return;
	  }
	if ((dev->actual_length > 0) && (port->run))
	  {
	    if (( port->rxq.frame = dev_alloc_skb ( port->dev->mtu )) == NULL )
	      {
		printk("intr_rx: can't allocate buffer\n");
		port_to_stats(port, rx_dropped)++;
		port->rxq.error_recovery++;
		return;
	      }
	    port->rxq.flags = 0;
	    port->rxq.count = dev->actual_length;
	    port->rxq.segment_cnt = 1;
            /*
             * If we need to reverse the bits, then this is where we do it
             */
            if (port->receiveMSBFirst)
              {
                fst_reverse_bits(&dev->bulk_in_buffer[8], dev->actual_length);
              }
	    memcpy ( skb_put ( port->rxq.frame, dev->actual_length ),
		     &(dev->bulk_in_buffer[8]),
		     dev->actual_length );
	    port_to_stats(port, rx_bytes) += dev->actual_length;
	    port_to_stats(port, rx_packets)++;
	    fst_dbg(DBG_RX, "Pushing frame up the stack %d bytes\n",
		    dev->actual_length);
	    /* Push upstream */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
            skb_reset_mac_header( port->rxq.frame);
#else
	    port->rxq.frame->mac.raw = port->rxq.frame->data;
#endif
	    port->rxq.frame->dev = port->dev;
	    if (port->proto == FST_RAW)
              port->rxq.frame->protocol = farsync_type_trans(port->rxq.frame, 
                                                             port_to_dev(port));
            else
              {
#ifdef IS_GEN_HDLC
                if (port->char_file)
                  {
                    port->rxq.frame->protocol = htons (ETH_P_WAN_PPP);
                  }
                else
                  {
                    if (port->hdlc_proto == IF_PROTO_HDLC)
                      {
                        port->rxq.frame->protocol = htons ( ETH_P_IP );
                      }
                    else
                      {
                        if (port->hdlc_proto == IF_PROTO_HDLC_ETH)
                          {
                            port->rxq.frame->protocol = htons ( ETH_P_8021Q );
                          }
                        else
                          {
                            port->rxq.frame->protocol = hdlc_type_trans (port->rxq.frame, 
                                                                         port_to_dev(port));
                          }
                      }
                  }
#else
                    port->rxq.frame->protocol = htons (ETH_P_WAN_PPP);
#endif
              }
	    if (port->monitor_mode)
	      gen_mon_packet(port->rxq.frame->data, port->rxq.frame->len, port, FST_MON_RX);
	    if (port->run)
	      {
		/*
		 * Char or network interface?
		 */
		if (port->char_file)		  
		{
#ifdef FST_TTY_CHAR
                
                  st_fst_tty_area *fst_tty;
                  fst_tty = &fst_tty_area[port->minor_dev_no];
                  fst_tty->skb = port->rxq.frame;	    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
                  schedule_work(&(fst_tty->fst_tty_work));
#else
                  schedule_work((void *) (&(fst_tty->fst_tty_work)));
#endif             

#else
                  rx_status = char_queue_rx_frame(port, port->rxq.frame);
#endif  
		}
		else
		  {
                    if (port->run)
                      rx_status = netif_rx ( port->rxq.frame );
                    else
                      {
                        dev_kfree_skb(port->rxq.frame);
                        fst_dbg(DBG_ASS, "Discarding frame, port closed\n");
                      }
		    fst_process_rx_status( rx_status, port->dev->name );
		  }
		if (rx_status == NET_RX_DROP)
		  {
		    port_to_stats(port, rx_dropped)++;
		  }
	      }
	    else
	      {
		fst_dbg(DBG_RX, "Frame received after port closed. Ignored %d\n",
                        dev->actual_length);
		dev_kfree_skb(port->rxq.frame);
	      }
	    port->rxq.frame = NULL;
	  }
	else
	  {
	    fst_dbg(DBG_RX, "Ignoring a zero length urb\n");
	  }
        post_a_read_bulk_urb(dev);  /* Post a read for data */
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
static void sync_read_int_callback(struct urb *urb)
#else
static void sync_read_int_callback(struct urb *urb, struct pt_regs *regs)
#endif
{
	struct fst_card_info *dev;

	fst_dbg(DBG_RX, "In read int callback\n");
	dev = (struct fst_card_info *)urb->context;

	/* sync/async unlink faults aren't errors */
	if (urb->status && 
	    !(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET )) {
		printk("%s - nonzero read int status received: %d\n",
		    __FUNCTION__, urb->status);
		return;
	}
        dev->usb_trace_buffer[dev->usb_trace_ptr++] = dev->int_in_buffer[0];
        if (dev->int_in_buffer[0] == 0)
          {
            printk("fst_usb_trace: %s", dev->usb_trace_buffer);
            dev->usb_trace_ptr = 0;
          }
        if (urb->actual_length)
          post_a_read_int_urb(dev);
}

static int post_a_read_int_urb( struct fst_card_info *usb_dev)
{
  int retval;
  struct urb *urb = NULL;

  urb = usb_dev->int_urb;
  
  /* initialize the urb properly */
  usb_fill_int_urb(urb, usb_dev->udev, 
                   usb_rcvintpipe(usb_dev->udev, 
                                  usb_dev->int_in_endpointAddr),
                   usb_dev->int_in_buffer, usb_dev->int_in_size,
                   sync_read_int_callback, 
                   usb_dev, usb_dev->int_interval);
  urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
  //urb->transfer_flags |= URB_ZERO_PACKET;
  
  /* post a read on the int in port */
  retval = usb_submit_urb(urb, GFP_ATOMIC);
  if (retval) {
    printk_err("%s - failed submitting an int read urb, error %d", __FUNCTION__, retval);
    return retval;
  }

  return 0;
}

static int post_a_read_bulk_urb( struct fst_card_info *usb_dev)
{
  int retval;
  int read_length;
  struct urb *urb = NULL;

  urb = usb_dev->bulk_urb;
  read_length = usb_dev->ports[0]->rx_buffer_size + 8;

  /* initialize the urb properly */
  usb_fill_bulk_urb(urb, usb_dev->udev, 
                    usb_rcvbulkpipe(usb_dev->udev, 
                                    usb_dev->bulk_in_endpointAddr),
                    usb_dev->bulk_in_buffer, read_length, 
                    sync_read_bulk_callback, 
                   usb_dev);
  urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
  //	urb->transfer_flags |= URB_ZERO_PACKET;;
  
  /* post a read on the bulk in port */
  retval = usb_submit_urb(urb, GFP_ATOMIC);
  if (retval) {
    printk_err("%s - failed submitting aread bulk urb, error %d", __FUNCTION__, retval);
    return retval;
  }

  return 0;
}

static void
fst_q_work_item (u64 *queue, int card_index)
{
  unsigned long  flags;
  u64            mask;

  /*
   * Grab the queue exclusively
   */
  spin_lock_irqsave(&fst_work_q_lock, flags);

  /*
   * Making an entry in the queue is simply a matter of setting
   * a bit for the card indicating that there is work to do in the
   * bottom half for the card.  Note the limitation of 64 cards.
   * That ought to be enough
   */
  mask = 1 << card_index;
  *queue |= mask; 
  spin_unlock_irqrestore(&fst_work_q_lock, flags);
}

static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
fst_process_threshold_work_q ( struct work_struct *work )
#else
fst_process_threshold_work_q ( void *work )
#endif
{
  unsigned long         flags;
  u64                   workq;
  int                   i;
  struct fst_port_info *port;

  /*
   * Grab the queue exclusively
   */
  fst_dbg(DBG_TX, "fst_process_threshold_work_q\n");
  spin_lock_irqsave(&fst_work_q_lock, flags);
  workq = fst_threshq;
  fst_threshq = 0;
  spin_unlock_irqrestore(&fst_work_q_lock, flags);
  /*
   * Call the bottom half for each card with work waiting
   */
  for (i=0; i<FST_MAX_CARDS; i++)
    {
      if (workq & 0x01)
        {
          if (fst_cards_list[i] != NULL)
            {
              port = fst_cards_list[i]->ports[0];
              fst_dbg(DBG_TX, "Calling issue command to set signals on %d\n", i);
              if (port->usb_config.flowControl == COM_FLOW_CONTROL_RTSCTS)
                fst_issue_cmd( port, SETV24O);
              if (port->usb_config.flowControl == COM_FLOW_CONTROL_XONXOFF)
                {
                  if (port->v24OpSts & OPSTS_RTS)
                    fst_issue_cmd( port, SETXON);
                  else
                    fst_issue_cmd( port, SETXOFF);
                }
            }
        }
      workq = workq>>1;
    }
}

static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
fst_process_port_stats_work_q ( struct work_struct *work )
#else
fst_process_port_stats_work_q ( void *work )
#endif
{
  unsigned long         flags;
  u64                   workq;
  int                   i;
  struct fst_port_info *port;

  /*
   * Grab the queue exclusively
   */
  fst_dbg(DBG_TX, "fst_process_port_stats_work_q\n");
  spin_lock_irqsave(&fst_work_q_lock, flags);
  workq = fst_pstatsq;
  fst_pstatsq = 0;
  spin_unlock_irqrestore(&fst_work_q_lock, flags);
  /*
   * Call the bottom half for each card with work waiting
   */
  for (i=0; i<FST_MAX_CARDS; i++)
    {
      if (workq & 0x01)
        {
          if (fst_cards_list[i] != NULL)
            {
              port = fst_cards_list[i]->ports[0];
              fst_dbg(DBG_TX, "Calling issue command to get port stats on %d\n", i);
	      fst_issue_cmd( port, GETPORTSTATS);
              fst_dbg(DBG_TX, "Calling issue command to reset port stats on %d\n", i);
	      fst_issue_cmd( port, RESETPORTSTATS);
            }
        }
      workq = workq>>1;
    }
}

/* 400: tasklet changes for >= 5.11 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static void
fst_process_tx_work(struct tasklet_struct *t)
{
  fst_process_tx_work_q(0);
}

static void
fst_process_int_work(struct tasklet_struct *t)
{
  fst_process_int_work_q(0);
}
#endif

static void
fst_process_tx_work_q ( unsigned long work_q)
{
  unsigned long  flags;
  u64            work_txq;
  int            i;

  /*
   * Grab the queue exclusively
   */
  fst_dbg(DBG_TX, "fst_process_tx_work_q\n");
  spin_lock_irqsave(&fst_work_q_lock, flags);
  work_txq = fst_work_txq;
  fst_work_txq = 0;
  spin_unlock_irqrestore(&fst_work_q_lock, flags);

  /*
   * Call the bottom half for each card with work waiting
   */
  for (i=0; i<FST_MAX_CARDS; i++)
    {
      if (work_txq & 0x01)
        {
          if (fst_cards_list[i] != NULL)
            {
              fst_dbg(DBG_TX, "Calling tx bh for card %d\n", i);
              do_bottom_half_tx(fst_cards_list[i]);
            }
        }
      work_txq = work_txq>>1;
    }
}

static void
fst_process_int_work_q ( unsigned long work_q)
{
  unsigned long  flags;
  u64            work_intq;
  int            i;

  /*
   * Grab the queue exclusively
   */
  fst_dbg(DBG_INTR, "fst_process_int_work_q\n");
  spin_lock_irqsave(&fst_work_q_lock, flags);
  work_intq = fst_work_intq;
  fst_work_intq = 0;
  spin_unlock_irqrestore(&fst_work_q_lock, flags);

  /*
   * Call the bottom half for each card with work waiting
   */
  for (i=0; i<FST_MAX_CARDS; i++)
    {
      if (work_intq & 0x01)
        {
          if (fst_cards_list[i] != NULL)
            {
              fst_dbg(DBG_INTR,"Calling rx & tx bh for card %d\n", i);
              do_bottom_half_rx(fst_cards_list[i]);
	      /* 
               * Calling the tx BH directly could mean that
               * it runs twice concurrently.  So we have to
               * schedule it
               */
	      fst_q_work_item(&fst_work_txq, i);
	      tasklet_schedule(&fst_tx_task); 
            }
        }
      work_intq = work_intq>>1;
    }
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
static int fst_proc_info(struct seq_file *m, void *v)
{
	struct seq_file *buffer = m;
	struct fst_card_info *card;
	int c;
	int port_count = 0;

	seq_printf(buffer,
		   "FarSync %s Driver version %s - Patch Level %s - Build %s\n",
		   FST_DRIVER_TYPE, FST_USER_VERSION, FST_PATCH_LEVEL,
		   FST_ADDITIONAL);
	seq_printf(buffer, "%d Cards found\n", fst_ncards);
	for (c = 0; c < FST_MAX_CARDS; c++) {
		card = fst_cards_list[c];
		if (card)
		  {
		    seq_printf(buffer,
			       "\t%s-%s:(%s) %s IRQ%d,\t%d ports, State: %s\n",
			       port_to_dev(card->ports[0])->name,
			       port_to_dev(card->ports[card->nports - 1])->name,
			       card->ports[0]->usb_card_info.szSerialNo, 
			       type_strings[card->type],
			       card->irq, card->nports, state_strings[card->state]);
		    port_count += card->nports;
		  }
	}
	seq_printf(buffer, "Total number of ports = %d\n", port_count);
	return 0;
}
#else
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
fst_proc_info ( char *buffer, char **start, off_t offset, int length, 
                int *oef, void *data )
#else
fst_proc_info ( char *buffer, char **start, off_t offset, int length )
#endif
{
	int                   len=0;
	struct fst_card_info *card;
	int                   c;
        int                   port_count=0;

	len = sprintf ( buffer,"FarSync Flex %s Driver version %s - Patch Level %s - Build %s\n", 
                        FST_DRIVER_TYPE, FST_USER_VERSION, FST_PATCH_LEVEL,
                        FST_ADDITIONAL);
	len += sprintf ( buffer + len,"%d Cards found\n", 
			fst_ncards );
	for ( c = 0 ; c < FST_MAX_CARDS ; c++ ) {
          card = fst_cards_list[c];
          if (card)
            {
              len += sprintf ( buffer + len, 
                               "\t%s-%s: %s (%s),\t%d ports,  State: %s \n",
                               port_to_dev(card->ports[0])->name,
                               port_to_dev(card->ports[card->nports-1])->name,
                               type_strings[card->type],
                               card->ports[0]->usb_card_info.szSerialNo, 
                               card->nports,
                               state_strings[card->state] );
              port_count += card->nports;
            }
	}
        len += sprintf ( buffer + len, "Total number of ports = %d\n",
                         port_count);
#ifdef FST_TTY_CHAR
        len += sprintf ( buffer + len, "Total number of async connects = %d\n",
                         fst_tty_cnt);
#endif
	return len;
}
#endif

/*      Process the result of trying to pass a received frame up the stack
 */
static void fst_process_rx_status ( int rx_status , char *name)
{
  switch (rx_status)
    {
    case NET_RX_SUCCESS:
      {
	/*
	 * Nothing to do here
	 */
	break;
      }

    case NET_RX_DROP:
      {
	printk_err("%s: Received packet dropped\n", name);
	break;
      }
    }
}

#if 0
static int gen_mon_packet(struct sk_buff *skb, struct fst_port_info *port,
                          int tx_rx_ind)
{
  /*
   * We need to duplicate the skb and send it up to
   * the monitor application
   */
  struct sk_buff *mon_skb;
  int rx_status = NET_RX_SUCCESS;
  struct fstioc_mon header;

  /* 
   * Allocate SKB.  Length of frame to monitor + length of the
   * header we prepend.
   */
  if (( mon_skb = dev_alloc_skb ( skb->len+sizeof(struct fstioc_mon) )) == NULL )
    {
      fst_dbg( DBG_INTR,"gen_mon_packet: can't allocate skb\n");
      return 1;
    }

  /*
   * Setup the header
   */
  header.version = FSTIOC_MON_VERSION;
  header.tx_rx_ind = tx_rx_ind;
  header.sequence = port->sequence++;
  header.timestamp = jiffies*1000/HZ;
  header.length = skb->len;

  memcpy(skb_put (mon_skb, sizeof(struct fstioc_mon)), &header, 
         sizeof(struct fstioc_mon));
	
  /* Push upstream */
  fst_dbg(DBG_INTR, "Pushing monitor frame up the stack\n");
  mon_skb->protocol = htons ( ETH_P_DIAG );
  mon_skb->pkt_type = PACKET_HOST;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
  skb_reset_mac_header(mon_skb);
#else
  mon_skb->mac.raw = mon_skb->data;
#endif
  mon_skb->dev = port_to_dev(port);
  memcpy(skb_put(mon_skb, skb->len), skb->data, skb->len);
  fst_dbg(DBG_INTR, "Monitor packet length is %d\n", mon_skb->len);
  if (port->run)
    rx_status = netif_rx ( mon_skb );
  else
    {
      dev_kfree_skb(mon_skb);
      fst_dbg(DBG_ASS, "Discarding monitor frame, port closed\n");
    }
  fst_process_rx_status( rx_status, port_to_dev(port)->name );
  if (rx_status == NET_RX_DROP)
    {
      fst_dbg(DBG_ASS, "Could not deleiver monitor message");
      return 1;
    }
  return 0;
}
#else
static int gen_mon_packet(char *pkt, int pkt_len, struct fst_port_info *port,
                          int tx_rx_ind)
{
  /*
   * We need to duplicate the skb and send it up to
   * the monitor application
   */
  struct sk_buff *mon_skb;
  int rx_status = NET_RX_SUCCESS;
  struct fstioc_mon header;

  /* 
   * Allocate SKB.  Length of frame to monitor + length of the
   * header we prepend.
   */
  if (( mon_skb = dev_alloc_skb ( pkt_len+sizeof(struct fstioc_mon) )) == NULL )
    {
      fst_dbg( DBG_INTR,"gen_mon_packet: can't allocate skb\n");
      return 1;
    }

  /*
   * Setup the header
   */
  header.version = FSTIOC_MON_VERSION;
  header.tx_rx_ind = tx_rx_ind;
  header.sequence = port->sequence++;
  header.timestamp = jiffies*1000/HZ;
  header.length = pkt_len;

  memcpy(skb_put (mon_skb, sizeof(struct fstioc_mon)), &header, 
         sizeof(struct fstioc_mon));
	
  /* Push upstream */
  fst_dbg(DBG_INTR, "Pushing monitor frame up the stack\n");
  mon_skb->protocol = htons ( ETH_P_DIAG );
  mon_skb->pkt_type = PACKET_HOST;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
  skb_reset_mac_header(mon_skb);
#else
  mon_skb->mac.raw = mon_skb->data;
#endif
  mon_skb->dev = port_to_dev(port);
  memcpy(skb_put(mon_skb, pkt_len), pkt, pkt_len);
  fst_dbg(DBG_INTR, "Monitor packet length is %d\n", mon_skb->len);
  if (port->run)
    rx_status = netif_rx ( mon_skb );
  else
    {
      dev_kfree_skb(mon_skb);
      fst_dbg(DBG_ASS, "Discarding monitor frame, port closed\n");
    }
  fst_process_rx_status( rx_status, port_to_dev(port)->name );
  if (rx_status == NET_RX_DROP)
    {
      fst_dbg(DBG_ASS, "Could not deleiver monitor message");
      return 1;
    }
  return 0;
}
#endif

static void
fst_notify_state_change ( struct fst_card_info *card, struct fst_port_info *port )
{
  struct sk_buff *skb;
  int rx_status = NET_RX_SUCCESS;

    fst_dbg(DBG_INTR, "fst_notify_state_change: %s\n", port_to_dev(port)->name);
    if (port->notify_mode)
      {
        /*
         * Notify mode is on.  Basic or Extended?
         */
        if (port->notify_mode == FST_NOTIFY_EXTENDED)
          {
            /* 
             * Just signal select that something has happened
             */
            port->exception++;
            wake_up_interruptible (&port->pollq);
            return;
          }
        /*
         * Basic mode is send up a zero length message
         * Note that we can't do this for the async interface
         * How do you write a zero length message into a fifo?
         */
        if (port->mode == FST_MODE_ASYNC)
          return;
        /*
         * Otherwise proceed.  Allocate the SKB 
         */
        if (( skb = dev_alloc_skb ( 0 )) == NULL )
          {
            fst_dbg( DBG_INTR,"notify_state_change: can't allocate skb\n");
          }
        
        /*
         * We know the length we need to receive, = 0
         */
        skb_put( skb, 0 );
	
        /* Push upstream */
        fst_dbg(DBG_INTR, "Pushing frame up the stack\n");
        if ( port->proto == FST_HDLC || port->proto == FST_PPP )
          {
            /* Mark for further processing by sPPP module */
            skb->protocol = htons ( ETH_P_WAN_PPP );
          }
        else
          {
            /* DEC customer specific protocol (since nothing defined for
             * marking raw data), at least one other driver uses this value
             * for this purpose.
             */
            skb->protocol = htons ( ETH_P_CUST );
            skb->pkt_type = PACKET_HOST;
          }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
        skb_reset_mac_header(skb);
#else
        skb->mac.raw = skb->data;
#endif
        skb->dev = port_to_dev(port);
        if (port->char_file)
	{
#ifdef FST_TTY_CHAR
	
   	  st_fst_tty_area *fst_tty;
          fst_tty = &fst_tty_area[port->minor_dev_no];
          fst_tty->skb = skb;	  
	 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	  schedule_work(&(fst_tty->fst_tty_work));
#else
	  schedule_work((void *) (&(fst_tty->fst_tty_work)));
#endif
#else	
          rx_status = char_queue_rx_frame(port, skb);
#endif
        }
        else
          {
            if (port->run)
              {
                rx_status = netif_rx ( skb );
              }
            else
              {
                dev_kfree_skb(skb);
                fst_dbg(DBG_ASS, "Discarding state change message, port closed\n");
              }
            fst_process_rx_status( rx_status, port_to_dev(port)->name );
          }
        if (rx_status == NET_RX_DROP)
          {
            fst_dbg(DBG_ASS, "Could not deliver state change message");
          }
      }
    else
      {
        fst_dbg(DBG_INTR, "Notify mode not on\n");
      }

}

#ifndef FST_TTY_CHAR
static int
fst_get_tx_queue_len (struct fst_port_info *port)
{
  struct fst_card_info *card = port->card;
  unsigned long flags;
  int txq_length = 0;

  spin_lock_irqsave ( &card->card_lock, flags );
  if ((txq_length = port->txqe - port->txqs) < 0)
    {
      /*
       * This is the case where the next free has wrapped but the
       * last used hasn't
       */
      txq_length = txq_length + FST_TXQ_DEPTH;
    }
  spin_unlock_irqrestore ( &card->card_lock, flags );
  return txq_length;
}
#endif

#define CMD_BUF_LEN 4
/*      Issue a Mailbox command for a port.
 *      Note we issue them on a fire and forget basis, not expecting to see an
 *      error and not waiting for completion.
 */
static void
fst_issue_cmd ( struct fst_port_info *port, unsigned short cmd )
{
        struct fst_card_info *usb_dev;
        int retval;
	char *sigs = (char *) &port->scratch;

        usb_dev = port->card;
	switch (cmd)
	  {
	    case STARTPORT:
	     {
	       fst_dbg(DBG_CMD, "Sending start port\n");
	       retval = usb_control_msg(usb_dev->udev, 
					usb_sndctrlpipe(usb_dev->udev, 0),
					USS_CMD_START_PORT, 0x40, 0, 0, 
					NULL, 0, 10000);
	       if (retval==0) 
		 {
		   fst_dbg(DBG_CMD, "Start port response received\n");
		 }
	       else
		 {
		   fst_dbg(DBG_ASS, "Start port error errno = %d\n", retval);
		 }
	       break;
	     }
	     
	    case STOPPORT:
	      {
		fst_dbg(DBG_CMD, "Sending stop port\n");
		retval = usb_control_msg(usb_dev->udev, 
					 usb_sndctrlpipe(usb_dev->udev, 0),
					 USS_CMD_STOP_PORT, 0x40, 0, 0, 
					 NULL, 0, 10000);
		if (retval==0) 
		  {
		    fst_dbg(DBG_CMD, "Stop port response received\n");
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Stop port error errno = %d\n", retval);
		  }
		break;
	    }
	    case ASSERTSIGNALS:
	      {
		sigs[0]=0x03;
		sigs[1]=0x00;
		sigs[2]=0x00;
		sigs[3]=0x00;

		fst_dbg(DBG_CMD, "Raising control signals\n");
		
		retval = usb_control_msg(usb_dev->udev, 
					 usb_sndctrlpipe(usb_dev->udev, 0),
					 USS_CMD_SET_SIGNALS, 0x40, 0, 0, 
					 sigs, CMD_BUF_LEN, 10000);
		
		if (retval==CMD_BUF_LEN) 
		  {
		    fst_dbg(DBG_CMD, "Set signals up response received\n");
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Set signals up error errno = %d\n", retval);
		  }
                port->v24OpSts = 0x3;
		break;
	      }
	    case RELEASESIGNALS:
	      {
		sigs[0]=0x00;
		sigs[1]=0x00;
		sigs[2]=0x00;
		sigs[3]=0x00;
		fst_dbg(DBG_CMD, "Lowering control signals\n");
		retval = usb_control_msg(usb_dev->udev, 
					 usb_sndctrlpipe(usb_dev->udev, 0),
					 USS_CMD_SET_SIGNALS, 0x40, 0, 0, 
					 sigs, CMD_BUF_LEN, 10000);

		if (retval==CMD_BUF_LEN) 
		  {
		    fst_dbg(DBG_CMD, "Set signals down response received\n");
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Set signals down error errno = %d\n", retval);
		  }
                port->v24OpSts = 0;
		break;
	      }
	    case GETSIGNALS:
	      {
		int *data = (int *) &port->scratch;

		*data=0;
		fst_dbg(DBG_CMD, "Getting control signals\n");
		retval = usb_control_msg(usb_dev->udev, 
					 usb_rcvctrlpipe(usb_dev->udev, 0),
					 USS_CMD_GET_SIGNALS, 0xC0, 0, 0, 
					 data, CMD_BUF_LEN, 10000);
		le32_to_cpus(&data);					 
		if (retval==4) 
		  {
		    fst_dbg(DBG_CMD, "Get signals response received 0x%p\n", data);
		    port->v24IpSts = *data & 0x0f;
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Get signals error errno = %d\n", retval);
		  }
		break;
	      }
	    case SETSIGNALS:
	      {
		int *data = (int *) &port->scratch;

		*data = port->v24OpSts;
		fst_dbg(DBG_CMD, "Setting control signals\n");
		cpu_to_le32s(&data);
		retval = usb_control_msg(usb_dev->udev, 
					 usb_sndctrlpipe(usb_dev->udev, 0),
					 USS_CMD_SET_SIGNALS, 0x40, 0, 0, 
					 data, CMD_BUF_LEN, 10000);
		le32_to_cpus(&data);
		if (retval==CMD_BUF_LEN) 
		  {
		    fst_dbg(DBG_CMD, "Set signals response received\n");
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Set signals up error errno = %d\n", retval);
		  }
		break;
	      }

	  case RECONFIG:
	    {
	      printk("%s: Sending Reconfig async port\n",  port->dev->name);
	    }
	    break;

	  case FLUSHBUFFERS:
	    {
	      int *data = (int *) &port->scratch;

	      *data =0x03;
	      fst_dbg(DBG_CMD, "%s: Sending flush buffer\n",  port->dev->name);

	      cpu_to_le32s(&data);
	      retval = usb_control_msg(usb_dev->udev, 
				       usb_sndctrlpipe(usb_dev->udev, 0),
				       USS_CMD_FLUSH_BUFFERS, 0x40, 0, 0, 
				       data, CMD_BUF_LEN, 10000);
	      le32_to_cpus(&data);		       
	      if (retval==CMD_BUF_LEN) 
		{
		  fst_dbg(DBG_CMD, "Flush buffers response received\n");
		}
	      else
		{
		  fst_dbg(DBG_ASS, "Flush Buffers error errno = %d\n", retval);
		}
	    }
	    break;

	    case SETPRESERVE:
	      {
		int *data = (int *) &port->scratch;

		*data = 1;
		fst_dbg(DBG_CMD, "Setting preserve control signals\n");
		cpu_to_le32s(&data);
		retval = usb_control_msg(usb_dev->udev, 
					 usb_sndctrlpipe(usb_dev->udev, 0),
					 USS_CMD_SET_PRESERVE_SIGNALS, 0x40, 0, 0, 
					 data, CMD_BUF_LEN, 10000);
		le32_to_cpus(&data);			 
		if (retval==CMD_BUF_LEN) 
		  {
		    fst_dbg(DBG_CMD, "Set Preserve Signals response received\n");
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Set Preserve Signals up error errno = %d\n", retval);
		  }
	      }
              break;

	    case GETPRESERVE:
	      {
		int *data = (int *) &port->scratch;

		*data=0;
		fst_dbg(DBG_CMD, "Getting Preserve Signals\n");
		cpu_to_le32s(&data);
		retval = usb_control_msg(usb_dev->udev, 
					 usb_rcvctrlpipe(usb_dev->udev, 0),
					 USS_CMD_GET_PRESERVE_SIGNALS, 0xC0, 0, 0, 
					 data, 4, 10000);
		le32_to_cpus(&data);			 
		if (retval==4) 
		  {
		    fst_dbg(DBG_CMD, "Get Preserve Signals response received 0x%p\n", data);
		    if (data)
                      port->v24OpSts |= 0x0100;
                    else
                      port->v24OpSts &= !0x0100;
		  }
		else
		  {
		    fst_dbg(DBG_ASS, "Get Preserve Signals error errno = %d\n", retval);
		  }
	      }
              break;

	  case SETXON:
	    {
	      fst_dbg(DBG_CMD, "%s: Sending XON Command\n",  port->dev->name);

	      retval = usb_control_msg(usb_dev->udev, 
				       usb_sndctrlpipe(usb_dev->udev, 0),
				       USS_CMD_SEND_XON, 0x40, 0, 0, 
				       NULL, 0, 10000);
	      if (retval==0) 
		{
		  fst_dbg(DBG_CMD, "XON response received\n");
		}
	      else
		{
		  fst_dbg(DBG_ASS, "Set XON error errno = %d\n", retval);
		}
	    }
	    break;

	  case SETXOFF:
	    {
	      fst_dbg(DBG_CMD, "%s: Sending XOFF Command\n",  port->dev->name);

	      retval = usb_control_msg(usb_dev->udev, 
				       usb_sndctrlpipe(usb_dev->udev, 0),
				       USS_CMD_SEND_XOFF, 0x40, 0, 0, 
				       NULL, 0, 10000);
	      if (retval==0) 
		{
		  fst_dbg(DBG_CMD, "XOFF response received\n");
		}
	      else
		{
		  fst_dbg(DBG_ASS, "Set XOFF error errno = %d\n", retval);
		}
	    }
	    break;

	  case GETPORTSTATS:
	    {
	      IREX *stats = (IREX *) &port->scratch;
	      unsigned char tx_status = 0;
	      unsigned char rx_status = 0;

	      fst_dbg(DBG_CMD, "%s: Sending GETPORTSTATS Command\n",  port->dev->name);
	      flex_irex_to_le(stats);
	      retval = usb_control_msg(usb_dev->udev, 
				       usb_rcvctrlpipe(usb_dev->udev, 0),
				       USS_CMD_GET_PORT_STATS, 0xC0, 0, 0, 
				       stats, sizeof(IREX), 10000);
	      flex_irex_to_cpu(stats);
	      if (retval==sizeof(IREX)) 
		{
		  fst_dbg(DBG_CMD, "GETPORTSTATS response received\n");
                  fst_dbg(DBG_CMD, "Status count = %d\n", stats->InterfaceRecord.StatusCount);
                  fst_dbg(DBG_CMD, "V24In = %x  ", stats->InterfaceRecord.V24In);
                  fst_dbg(DBG_CMD, "V24Out = %x\n", stats->InterfaceRecord.V24Out);
                  if (stats->InterfaceRecord.StatusArray[0])
		    {
		      fst_dbg(DBG_CMD, "%s: Rx CRC\n", port->dev->name );
		      rx_status |= RX_CRC;
		    }
		  if (stats->InterfaceRecord.StatusArray[1])
		    {
		      fst_dbg(DBG_CMD, "%s: Rx Overflow\n", port->dev->name);
		      rx_status |= RX_OFLO;
		    }
		  if (stats->InterfaceRecord.StatusArray[2])
		    {
		      fst_dbg(DBG_CMD, "%s: Rx frame less than 4 bytes\n", port->dev->name);
		      rx_status |= RX_FRAM;
		    }
		  if (stats->InterfaceRecord.StatusArray[3])
		    {
		      fst_dbg(DBG_CMD, "%s: Rx ending on non octet boundary\n", port->dev->name);
		      rx_status |= RX_FRAM;
		    }
		  if (stats->InterfaceRecord.StatusArray[4])
		    {
		      fst_dbg(DBG_CMD, "%s: Rx Aborted frame\n", port->dev->name);
		      rx_status |= RX_ABRT;
		    }
		  if (stats->InterfaceRecord.StatusArray[5])
		    {
		      fst_dbg(DBG_CMD, "%s: Tranmitter Interrupt underrun\n", port->dev->name );
		      tx_status |= TX_URUN;
		    }
		  if (stats->InterfaceRecord.StatusArray[6])
		    {
		      fst_dbg(DBG_CMD, "%s: Receiver Interrupt overrun\n", port->dev->name);
		      rx_status |=  RX_OFLO;
		    }
		  if (stats->InterfaceRecord.StatusArray[7])
		    {
		      fst_dbg(DBG_CMD, "%s: DCD lost during frame reception\n", port->dev->name);
		    }
		  if (stats->InterfaceRecord.StatusArray[8])
		    {
		      fst_dbg(DBG_CMD, "%s: CTS lost while transmitting\n", port->dev->name);
		    }
		  if (stats->InterfaceRecord.StatusArray[9])
		    {
		      fst_dbg(DBG_CMD, "%s: DSR dropped\n", port->dev->name);
		    }
		  if (stats->InterfaceRecord.StatusArray[10])
		    {
		      fst_dbg(DBG_CMD, "%s: Hardware Adapter Failure\n", port->dev->name );
		    }
		  if (rx_status)
		    fst_log_rx_error(usb_dev, port, rx_status);
		  if (tx_status)
		    fst_log_tx_error(usb_dev, port, tx_status);
		}
	      else
		{
		  fst_dbg(DBG_CMD, "Set GETPORTSTATS error errno = %d\n", retval);
		}
	    }
	    break;

	  case RESETPORTSTATS:
	    {
	      fst_dbg(DBG_CMD, "%s: Sending RESETPORTSTATS Command\n",  port->dev->name);

	      retval = usb_control_msg(usb_dev->udev, 
				       usb_sndctrlpipe(usb_dev->udev, 0),
				       USS_CMD_RESET_PORT_STATS, 0x40, 0, 0, 
				       NULL, 0, 10000);
	      if (retval==0) 
		{
		  fst_dbg(DBG_CMD, "RESETPORTSTATS response received\n");
		}
	      else
		{
		  printk("Set RESETPORTSTATS error errno = %d\n", retval);
		}
	    }
	    break;

	  case SETLINESPEED:
	    {
	      fst_dbg(DBG_CMD, "%s: Setting the line speed\n",  port->dev->name);

	      retval = usb_control_msg(usb_dev->udev, 
				       usb_sndctrlpipe(usb_dev->udev, 0),
				       USS_CMD_SET_LINESPEED, 0x40, 0, 0, 
				       &port->usb_config.lineSpeed, 4, 10000);
	      if (retval==0) 
		{
		  fst_dbg(DBG_CMD, "set LINESPEED response received\n");
		}
	      else
		{
		  fst_dbg(DBG_ASS, "Set LINESPEED error errno = %d\n", retval);
		}
	    }
	    break;

          default:
            printk_err("Invalid Isuue Command %d\n", cmd);
	  }
}

/*      Port output signals control
 */
static inline void
fst_op_raise ( struct fst_port_info *port, unsigned int outputs )
{
        outputs |= port->v24OpSts;
        port->v24OpSts =  outputs;
        if ( port->run )
	  fst_issue_cmd (port, ASSERTSIGNALS);
        /*
         * If we are in ignore carrier mode, then now is the
         * time to tell the network layer that the carrier is on
         */
        if (port->ignoreCarrier)
          {
            if ( ! netif_carrier_ok (port_to_dev(port) ))
              {
                fst_dbg( DBG_OPEN,"Net carrier on\n");
#ifdef FST_NET_DEV
#ifndef IS_GEN_HDLC
                /* Poke sPPP to renegotiate */
                if ( port->proto == FST_HDLC || port->proto == FST_PPP )
                  {
                    sppp_reopen ( port_to_dev(port) );
                  }
#endif
#endif
                netif_carrier_on ( port_to_dev(port) );
              }
          }
}

static inline void
fst_op_lower ( struct fst_port_info *port, unsigned int outputs )
{
        int preserve = port->v24OpSts & 0x100;

        fst_dbg(DBG_ASS, "Preserve signals = %d\n", preserve);
        outputs = ~(outputs & port->v24OpSts);
        port->v24OpSts = outputs & 0x03;
	if (preserve)
	  fst_dbg(DBG_ASS, "%s: Leaving signals alone\n", port->dev->name);
	else
	  {
	    /*
	     * Need to lower control signals here, if we are not
	     * in the preserve signals mode
	     */
	    fst_dbg(DBG_ASS, "%s: Lowering signals\n", port->dev->name);
	    fst_issue_cmd (port, RELEASESIGNALS);

        /*
         * If we are in ignore carrier mode, then now is the 
         * time to tell the network layer that carrier is off
         */
		if (port->ignoreCarrier)
        {
            netif_carrier_off ( port_to_dev(port) );
        }
	  }
}


/*
 *      Setup port Rx buffers
 */
static void
fst_rx_config ( struct fst_port_info *port )
{
        int pi;
        struct fst_card_info *card;

        pi   = port->index;
        card = port->card;
}


/*
 *      Setup port Tx buffers
 */
static void
fst_tx_config ( struct fst_port_info *port )
{
        int pi;
        struct fst_card_info *card;

        pi   = port->index;
        card = port->card;
}

/*      Check that the shared memory configuration is one that we can handle
 *      and that some basic parameters are correct
 */
static void
check_started_ok ( struct fst_card_info *card )
{
  return;

}

static void do_bottom_half_tx (struct fst_card_info *card)
{
  struct fst_port_info *port;
  int retval = 0;
  struct urb *urb = NULL;
  char *buf = NULL;
  size_t writesize;
  int pi;
  int txq_length;
  struct sk_buff *skb;
  unsigned long flags;
  int tx_length;
  __u16 *len_ptr;


  /*
   *  Find a free buffer for the transmit
   *  Step through each port on this card
   */

  fst_dbg(DBG_TX, "do_bottom_half_tx\n");
  for ( pi = 0 ; pi < card->nports ; pi++ )
    {
      port = card->ports[pi];
      if ( (port->dev == NULL) || (!port->run) )
        {
          continue;
        }

      if (card->tx_urb_cnt <= 0)
	{
	  fst_dbg(DBG_TX, "Max URBs in progress limit reached\n");
	  return;
	}

      spin_lock_irqsave ( &card->card_lock, flags );
      if ((txq_length = port->txqe - port->txqs) < 0)
	{
	  /*
	   * This is the case where one has wrapped and the
	   * maths gives us a negative number
	   */
	  txq_length = txq_length + FST_TXQ_DEPTH ;
	}
      spin_unlock_irqrestore ( &card->card_lock, flags );
      if (txq_length > 0)
	{
	  /*
	   * There is something to send
	   */
	  skb = port->txq[port->txqs].frame;
	  
	  /*
	   * Decrement the segment count.  It tells us how many more bits
	   * to the frame there are left to do
	   */
	  port->txq[port->txqs].segment_cnt--;
	  tx_length = skb->len;
	  port->txq[port->txqs].current_seg++;

	  writesize = min((size_t)tx_length+8, (size_t)port->tx_buffer_size+8);
	  /*
	   * copy the data and set the required indicators on the
	   * card.
	   */
	  
	  /* verify that we actually have some data to write */
	  if (tx_length == 0)
	    goto exit;

	  urb = card->tx_bulk_urb;	  	  
	  buf = card->bulk_out_buffer;
	  
	  memcpy(&buf[8], skb->data, tx_length);
          /*
           * If we need to reverse the bits, this is where we do it
           */
          if (port->transmitMSBFirst)
            {
              fst_reverse_bits(&buf[8], tx_length);
            }
	  /*
	   * Set the length in the pre-able
	   */
	  len_ptr = (__u16 *)&buf[4];          	  
	  *len_ptr = cpu_to_le16((__u16)tx_length);
	  /* initialize the urb properly */
	  usb_fill_bulk_urb(urb, card->udev, 
                            usb_sndbulkpipe(card->udev, 
                                            card->bulk_out_endpointAddr),
			   buf, writesize, sync_write_bulk_callback, card);
	  urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	  urb->transfer_flags |= URB_ZERO_PACKET;;
	  
	  /* send the data out the bulk port */
	  retval = usb_submit_urb(urb, GFP_ATOMIC);
	  if (retval) {
	    printk_err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
	    port->txqs++;
	    if (port->txqs == FST_TXQ_DEPTH)
	        port->txqs = 0;
	    goto exit;
	  }
	  else
	    {
	      card->tx_urb_cnt--;
	    }
	  wake_up_interruptible (&port->writeq);
	  spin_lock_irqsave ( &card->card_lock, flags );
	  port->txqs++;
	  if (port->txqs == FST_TXQ_DEPTH)
	    {
	      port->txqs = 0;
	    }
	  if (port->txqs == port->txqe)
	    {
	      fst_dbg(DBG_TX, "Card %d Port %d: Tx queue Now empty\n",
		  card->card_no, port->index);
	      fst_notify_state_change (card, port);
              wake_up_interruptible (&port->pollq);
	    }
	  spin_unlock_irqrestore ( &card->card_lock, flags );
	  /*
	   * If we have flow control on, can we now release it?
	   */
	  if (port->start)
	    {
	      if (txq_length < fst_txq_low)
		{
                  fst_dbg(DBG_ASS, "%s: Start the network layer\n", 
                          port->dev->name);
                  if (!port->char_file)
                    {
                      netif_wake_queue ( port->dev );
                    }
#ifdef FST_TTY_CHAR
                  else
                    {
                      st_fst_tty_area *fst_tty;
                      fst_tty = &fst_tty_area[port->index];
                      fst_dbg(DBG_TTY, "%s: Waking up tty writes\n", 
                              fst_tty->name);
                      tty_wakeup(fst_tty->tty);
                    }
#endif
                  port->start = 0;
		}
	    }
exit:
	  dev_kfree_skb ( skb );
	  return;
	}
      else
	{
	  /*
	   * Nothing to send so break out of the while loop
	   */
	  break;
	}
      if ( ++port->txpos >= port->num_tx_buffers )
	port->txpos = 0;
    }
  return;
}

static void do_bottom_half_rx (struct fst_card_info *card)
{
  fst_dbg(DBG_ASS, "Do bottom half rx\n");
  return;
}
static int
check_combination ( int num, int size)
{
  /*
   * Check that the number and size do not exceed the following table
   *
   *   num of buffs            max size   
   *     2                       32K
   *     4                       16K
   *     8                        8K
   *    16                        4K
   *    32                        2K
   *    64                        1K
   *   128                      0.5K
   */
  int err = 0;
  switch(num)
    {
    case 2:
      {
        if (size > 32*1024)
          err++;
        break;
      }
    case 4:
      {
        if (size > 16*1024)
          err++;
        break;
      }
    case 8:
      {
        if (size > 8*1024)
          err++;
        break;
      }
    case 16:
      {
        if (size > 4*1024)
          err++;
        break;
      }
    case 32:
      {
        if (size > 2*1024)
          err++;
        break;
      }
    case 64:
      {
        if (size > 1*1024)
          err++;
        break;
      }
    case 128:
      {
        if (size > 1024/2)
          err++;
        break;
      }
    default:
      {
        printk_err("Number of buffers must be 2,4,8,16,32,64, or 128 : %d\n",
                   num);
        err++;
      }
    }
  if (err)
    {
      return 1;
    }
  return 0;
}

static int
validate_buffer_config (struct fstioc_info *info)
{
  /*
   * We need some checks on buffer configurations because the
   * card itself doesn't do any
   */
  if ( (info->numTxBuffers == 0) || (info->numRxBuffers == 0) )
    {
      printk_err("Num Tx Buffers or Num Rx Buffers is zero\n");
      return 1;
    }
  if ( (info->numTxBuffers == 1) || (info->numRxBuffers == 1) )
    {
      printk_err("Num Tx Buffers or Num Rx Buffers is 1\n");
      return 1;
    }    
  if ( (info->numTxBuffers > 128) || (info->numRxBuffers > 128) )
    {
      printk_err("Num Tx Buffers or Num Rx Buffers > 128\n");
      return 1;
    }
  if ( (info->txBufferSize == 0) || (info->rxBufferSize == 0) )
    {
      printk_err("Tx Buffer Size or Rx Buffer Size is zero\n");
      return 1;
    }
  if ( (info->txBufferSize > 32*1024) || (info->rxBufferSize > 32*1024) )
    {
      printk_err("Tx Buffer Size or Rx Buffer Size > 32*1024\n");
      return 1;
    }
  if ( check_combination(info->numTxBuffers, info->txBufferSize) )
    {
      printk_err("Invalid num/size combination on Tx Buffers\n");
      return 1;
    }
  if ( check_combination(info->numRxBuffers, info->rxBufferSize) )
    {
      printk_err("Invalid num/size combination on Rx Buffers\n");
      return 1;
    }
  return 0;
}

 static int purge_tx_queue (struct fst_card_info *card, 
			    struct fst_port_info *port)
{
  int txq_length;
  int i;
  int start;
  unsigned long flags;

  /*
   * Purge all frames on the tx queue
   */

  fst_dbg(DBG_TX, "Purging all frame from the tx queue\n");
  /*
   * Is there anything there?
   */
  spin_lock_irqsave ( &card->card_lock, flags );
  start = port->txqs;
  if ((txq_length = port->txqe - port->txqs) < 0)
    {
      /*
       * This is the case where the next free has wrapped but the
       * last used hasn't
       */
      txq_length = txq_length + FST_TXQ_DEPTH;
    }
  port->txqs = port->txqe;  
  spin_unlock_irqrestore ( &card->card_lock, flags );
  if (txq_length == 0)
    {
      return 0;  /* Nothing there */
    }
  /*
   * de-queue the buffer's and deallocate the resources
   */
  printk("%s: Purging %d frames from the transmit queue\n", port->dev->name,
	 txq_length);
  for (i=0; i<txq_length; i++)
    {
      if (port->txq[start].frame)
	{
          fst_dbg(DBG_TX, "Purging buffer %d at %p\n", start, port->txq[start].frame);
          dev_kfree_skb (port->txq[start].frame);
          port->txq[start].frame = NULL;
	}
      else
	{
	  printk("txq purge found NULL pointer\n");
	}
      start++;
      if (start == FST_TXQ_DEPTH)
	start = 0;
    }
  return 0;
}

static int
parse_custom_clock_rate_structure(struct fst_port_info *port, 
				  struct fstioc_custom_rate_config *custom_clock_rate)
{
  /*
   * Check that the supplied buffer looks like it may contain a valid
   * set of values
   */
  if (custom_clock_rate->version != FST_CUSTOM_RATE_CONFIG_VERSION)
    {
      printk("Custom clock rate structure is not the version expected %d %d\n",
	      FST_CUSTOM_RATE_CONFIG_VERSION, custom_clock_rate->version);
      return 1;
    }
  if (!((custom_clock_rate->multiplier == (FST_CUSTOM_CLOCK_MULTIPLIER_1))  ||
	(custom_clock_rate->multiplier == (FST_CUSTOM_CLOCK_MULTIPLIER_16))))
    {
      printk("Custom clock rate multiplier is not expected value (1 or 16) %d\n",
	      custom_clock_rate->multiplier);
      return 2;
    }
  if (custom_clock_rate->clock_type > FST_CUSTOM_RATE_CLOCK_LOW_MASTER)
    {
      printk("Custom clock type is not in the rage expected (0 - 2) %d\n",
	      custom_clock_rate->clock_type);
      return 3;
    }
  if (strlen(custom_clock_rate->rate_info) != FST_CUSTOM_RATE_CONFIG_LENGTH-1)
    {
      printk("Custom clock rate info is not the length expected %d %d\n",
	     (int) strlen(custom_clock_rate->rate_info), FST_CUSTOM_RATE_CONFIG_LENGTH-1);
      return 4;
    }
  /*
   * Validation passed
   */
  return 0;
}

#if 1
unsigned int to_int(char c)
{
  if (c >= '0' && c <= '9') return      c - '0';
  if (c >= 'A' && c <= 'F') return 10 + c - 'A';
  if (c >= 'a' && c <= 'f') return 10 + c - 'a';
  return -1;
}

static int strtohex (unsigned char * source, unsigned char * dest, int size)
{
  int count = 0;
  int loops = 0;

  loops = (int)strlen(source)/2;
  fst_dbg(DBG_IOCTL, "strtohex of length %d\n", loops);
  for(count = 0; count < loops; count++) {
    dest[count] = 16* to_int(source[2*count]) + to_int(source[2*count+1]);
  }
  dest[count] = 0xff;
  return 0;
}
#endif


static int
set_custom_clock_rate(struct fst_port_info *port, 
		      struct fstioc_custom_rate_config *custom_clock_rate)
{
  struct fstioc_req *sys_request = (struct fstioc_req *) &port->scratch;
  char szRateInfo[(FST_CUSTOM_RATE_CONFIG_LENGTH*2)+1];
  unsigned char buf[FST_CUSTOM_RATE_CONFIG_LENGTH];
  unsigned char checksum[16];
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0)
  struct crypto_shash *tfm;
  struct shash_desc *desc;
#else
  struct crypto_hash *tfm;
  struct hash_desc desc;
#endif
#else
  struct crypto_tfm *tfm;
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,8,0)
  struct scatterlist sg;
#endif
#endif
  char szChecksum[3];
  char* pszClockType = NULL;
  int i;
  int retval;

#define MD5_SUM_LENGTH 23

  /*
   * Prepare the string and construct the usb command and send it
   */
  memset(sys_request, 0, sizeof(struct fstioc_req));
  memset(buf, 0x00, sizeof(buf));
  sys_request->msg_type = custom_clock_rate->multiplier;
  sys_request->msg_len = sizeof(struct fstioc_req);
  sys_request->i_reg_idx = port->index;

  fst_dbg(DBG_IOCTL, "Supplied rate = %d and rate info is %s\n", custom_clock_rate->rate, 
	 custom_clock_rate->rate_info);

  switch (custom_clock_rate->clock_type)
    {
    case FST_CUSTOM_RATE_CLOCK_LOW_SLAVE:
      pszClockType = "0000"; break;
    case FST_CUSTOM_RATE_CLOCK_LOW_MASTER:
      pszClockType = "0101"; break;
    case FST_CUSTOM_RATE_CLOCK_SLAVE:
    default:
      pszClockType = "0100"; break;
    }

  sprintf(szRateInfo, "%08X", ntohl(custom_clock_rate->rate));
  strcat(szRateInfo, "0");
  strcat(szRateInfo, custom_clock_rate->rate_info);
  strcat(szRateInfo, pszClockType);
  fst_dbg(DBG_IOCTL, "Temp1 ascii szRateInfo=%s\n", szRateInfo);
  fst_dbg(DBG_IOCTL, "Length of string to convert is %d\n", (int) strlen(szRateInfo));
  strtohex(szRateInfo, buf, sizeof(buf)-1);
  fst_dbg(DBG_IOCTL, "Temp2 ascii szRateInfo=%s\n", szRateInfo);
  fst_dbg(DBG_IOCTL, "Temp hex buf: ");
  for (i=0; i< sizeof(buf); i++)
    {
      fst_dbg(DBG_IOCTL, "%02x ", buf[i]);
    }
  fst_dbg(DBG_IOCTL, "\n");
  fst_dbg(DBG_IOCTL, "Temp3 ascii szRateInfo=%s\n", szRateInfo);

  /*
   * Use the Kernel Crypto API to calculate the MD5Sum of the updated rate string
   */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,8,0)
  tfm = crypto_alloc_shash("md5", 0, 0);
  if (IS_ERR(tfm))
    {
      printk("error in crypto_alloc_shash\n");
      return EIO;
    }
  desc = kmalloc(sizeof(struct shash_desc) + crypto_shash_descsize(tfm), 
			GFP_KERNEL);
  if (!desc)
    {
      printk("Error allocation memory for custom rate string\n");
      return EIO;
    }
  desc->tfm = tfm;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,2,0)
  desc->flags = 0;
#endif
  if (crypto_shash_init(desc))
    {
      printk("error in crypto_shash_init\n");
      kfree(desc);
      return EIO;
    }
  if (crypto_shash_update(desc, buf,  MD5_SUM_LENGTH))
    {
      printk("error in crypto_shash_update\n");
      kfree(desc);
      return EIO;
    }
  memset(checksum, 0, sizeof(checksum));
  if (crypto_shash_final(desc, checksum))
    {
      printk("error in crypto_shash_final\n");
      kfree(desc);
      return EIO;
    }
  kfree (desc);
#else
  tfm = crypto_alloc_hash("md5", 0, CRYPTO_ALG_ASYNC);
  desc.tfm = tfm;
  desc.flags = 0;
  crypto_hash_init(&desc);
  sg_init_one(&sg, buf, MD5_SUM_LENGTH);
  crypto_hash_update(&desc, &sg,  MD5_SUM_LENGTH);
  memset(checksum, 0, sizeof(checksum));
  crypto_hash_final(&desc, checksum);
#endif
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)
  tfm = crypto_alloc_tfm("md5", CRYPTO_TFM_REQ_MAY_SLEEP);
  crypto_digest_init(tfm);
  sg_init_one(&sg, buf, MD5_SUM_LENGTH);
  crypto_digest_update(tfm, &sg, MD5_SUM_LENGTH);
  memset(checksum, 0, sizeof(checksum));
  crypto_digest_final(tfm, checksum);
#else
  printk("This version of the kernel does not support setting custom clock rates\n");
  return EIO;
#endif
#endif
  fst_dbg(DBG_IOCTL, "Checksum: ");
  for (i=0; i< sizeof(checksum); i++)
    {
      fst_dbg(DBG_IOCTL, "%02x ", checksum[i]);
    }
  fst_dbg(DBG_IOCTL,"\n");

  sprintf(szChecksum, "%02X", checksum[0]);
  strcat(szRateInfo, szChecksum);
  fst_dbg(DBG_IOCTL,"Temp7 ascii szRateInfo=%s\n", szRateInfo);
  strtohex(szRateInfo, sys_request->u_msg, FST_CUSTOM_RATE_CONFIG_LENGTH);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,8,0)
  crypto_free_shash(tfm);
#else
  crypto_free_hash(tfm);
#endif
#else
  crypto_free_tfm(tfm);
#endif
  /*
   * Now we can send the request to Flex
   */
  flex_fstioc_req_to_le(sys_request);
  retval = usb_control_msg(port->card->udev, 
			   usb_sndctrlpipe(port->card->udev,0),
			   USS_CMD_SEND_SECURE_MSG, 0x40, 0, 0, 
			   sys_request, 
			   sizeof(struct fstioc_req), 10000);
  flex_fstioc_req_to_cpu(sys_request);
  if (retval != sizeof(struct fstioc_req))
    {
      printk("%s: Error in sending USB Control Message, retval is %d\n", 
	     port->dev->name, retval);
      return retval;
    }
  /*
   * We are always going to wait for a reply
   */
  fst_dbg(DBG_IOCTL, "Waiting for a reply\n");
  flex_fstioc_req_to_le(sys_request);
  retval = usb_control_msg(port->card->udev, 
			   usb_rcvctrlpipe(port->card->udev,0),
			   USS_CMD_RECV_SECURE_RSP, 0xC0, 0, 0, 
			   sys_request, 
			   sizeof(struct fstioc_req), 10000);
  flex_fstioc_req_to_cpu(sys_request);
  if (retval  != sizeof(struct fstioc_req))
    {
      printk("Set Custom Clock Rate: Retval is %d\n", retval);
      return retval;
    }
  fst_dbg(DBG_IOCTL, "Returning from set_custom_clock_rate\n");
  return 0;
}

static int
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
fst_siocdevprivate(struct net_device *dev, struct ifreq *ifr, void __user *data, int cmd)
#else
fst_ioctl ( struct net_device *dev, struct ifreq *ifr, int cmd )
#endif
{
	struct fst_card_info *card;
	struct fst_port_info *port;
	struct fstioc_control_request *control_request;
	struct fstioc_info info;
	struct fstioc_status state;
	struct fstioc_req *sys_request;
	struct fstioc_custom_rate_config custom_clock_rate;
	int retval = 0;
	int len;
	struct fstioc_cmd my_cmd;
	struct fstioc_char_data char_data;
	char fstioc_info_ver;
	char readv_mode;
	int info_size;
	unsigned char signals;
	char *sigs;
	unsigned long copied = 0;
	int speed;

#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
	fst_dbg( DBG_IOCTL, "ioctl: %x, %p\n", cmd, data);
#else
	fst_dbg( DBG_IOCTL,"ioctl: %d, %p\n", cmd-SIOCDEVPRIVATE, ifr->ifr_data );
#endif	
	port = dev_to_port(dev);
	card = port->card;
	sigs = (char *) &port->scratch;
	sys_request = (struct fstioc_req *) &port->scratch;

	if (!capable ( CAP_NET_ADMIN ))
	{
		if (!fst_allow_ioctl)
		{
			return -EPERM;
		}
	}

	switch ( cmd )
	{
		case FSTCPURESET:
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_CARD)))
			{
				card->state = FST_RESET;
				return 0;
			}
			else
			{
				printk("Permission for FSTCPURESET not allowed\n");
				return -EPERM;
			}

		case FSTCPURELEASE:
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_CARD)))
			{
				card->state = FST_STARTING;
				return 0;
			}
			else
			{
				printk("Permission for FSTCPURELEASE not allowed\n");
				return -EPERM;
			}

		case FSTWRITE:                  /* Code write (download) */
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_CARD)))
			{
				/* First copy in the header with the length and offset of data
				* to write
				*/
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( !data )
#else
				if ( ifr->ifr_data == NULL )
#endif	
				{
					return -EINVAL;
				}
				control_request = kmalloc ( sizeof ( struct fst_card_info ),  GFP_KERNEL);
				if (control_request == NULL)
				{
					printk_err ("%s: insufficient memory for FSTWRITE data\n",
						  port->dev->name);
					return -ENOMEM;
				}
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( copy_from_user ( control_request, data,
#else
				if ( copy_from_user ( control_request, ifr->ifr_data,
#endif	      
					sizeof ( struct fstioc_control_request )))
				{
					kfree (control_request);
					return -EFAULT;
				}
				/*
				* Check the version of the request
				* We only support version 1 so far
				*/
				if (control_request->uVersion != 1)
				{
					fst_dbg(DBG_ASS, "Invalid version number in fstwrite\n");
					kfree (control_request);
					return -EINVAL;
				}
				if (control_request->bDirection)
				{
					/*
					* To device
					*/
					retval = usb_control_msg(card->udev, 
												usb_sndctrlpipe(card->udev,  0),
												control_request->byRequest, 0x40, 
												control_request->wValue,
												control_request->wIndex, 
												&control_request->Data, 
												control_request->wDataLength, 10000);
					if (retval > 0)
					{
						fst_dbg(DBG_IOCTL, "Set control response received\n");
					}
					else
					{
						fst_dbg(DBG_ASS, "Set Control error errno = %d \n", 
								retval);
					}
				}
				else
				{
					/*
					* From device
					*/
					retval = usb_control_msg(card->udev, 
												usb_rcvctrlpipe(card->udev,  0),
												control_request->byRequest, 0xc0, 
												control_request->wValue,
												control_request->wIndex, 
												&control_request->Data, 
												control_request->wDataLength, 10000);
					if (retval > 0)
					{
						fst_dbg(DBG_IOCTL, "Get control response received\n");
						control_request->wDataLength = retval;
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
						if ( copy_to_user ( data, control_request, 
#else
						if ( copy_to_user ( ifr->ifr_data, control_request, 
#endif
								sizeof (struct fstioc_control_request)))
						{
							kfree (control_request);
							return -EFAULT;
						}
					}
					else
					{
						fst_dbg(DBG_ASS, "Get Control error errno = %d \n", 
								retval);
					}
				}   
				kfree (control_request);
				return 0;
			}
			else
			{
				printk("Permission for FSTWRITE not allowed\n");
				return -EPERM;
			}

		case FSTGETSHELL:
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_GET)))
			{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( data == NULL )
#else
				if ( ifr->ifr_data == NULL )
#endif
				{
					return -EINVAL;
				}

				info_size = flex_get_config(card, port, &info);
				if (info_size)
				{
					if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
						memcpy ( data, &info, info_size);
					else if ( copy_to_user ( data, &info, info_size))
#else
						memcpy ( ifr->ifr_data, &info, info_size);
					else if ( copy_to_user ( ifr->ifr_data, &info, info_size))
#endif
					{
						return -EFAULT;
					}
					return 0;
				}
				else
				{
					return -EFAULT;
				}
			}
			else
			{
				printk("Permission for FSTCPUGETSHELL not allowed\n");
				return -EPERM;
			}

        case FSTGETCONF:
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_GET)))
			{
				/* If card has just been started check the shared memory config
				 * version and marker
				 */
				if ( card->state == FST_STARTING )
				{
					check_started_ok ( card );

					/* If everything checked out enable card interrupts */
					if ( card->state == FST_RUNNING )
					{
						/* Nothing else to do */
					}
				}

				if ( card->state != FST_RUNNING )
				{
					return -ENODEV;
				}

#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( data == NULL )
#else
				if ( ifr->ifr_data == NULL )
#endif
				{
					return -EINVAL;
				}
				memset(&info, 0, sizeof(struct fstioc_info));
				info_size = flex_get_config(card, port, &info);
				if (info_size)
				{
					if(port->compat==1)
					{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
						memcpy ( data, &info, info_size);
#else
						memcpy ( ifr->ifr_data, &info, info_size);
#endif
					}
					else
					{
#ifndef RHEL_RELEASE_CODE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0)
						if (!access_ok(ifr->ifr_data, sizeof(info)))
#else
						if (!access_ok(data, sizeof(info)))
#endif
#else
						if (!access_ok(VERIFY_WRITE, ifr->ifr_data, sizeof(info)))
#endif
#elif (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(8,1)))
#if (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(9,0))
						if (!access_ok(ifr->ifr_data, sizeof(info)))
#else
						if (!access_ok(data, sizeof(info)))
#endif
#else
						if (!access_ok(VERIFY_WRITE, ifr->ifr_data, sizeof(info)))
#endif
						{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							printk("%s: fstgetconf: Can't access address %p\n", dev->name, data);
#else
							printk("%s: fstgetconf: Can't access address %p\n", dev->name, ifr->ifr_data);
#endif
							return -EFAULT;
						}
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
						if ( copy_to_user ( data, &info, info_size))
#else
						if ( copy_to_user ( ifr->ifr_data, &info, info_size))
#endif
						{
							return -EFAULT;
						}
					}
				}
				return 0;
			}
			else
			{
				printk("Permission for FSTGETCONF not allowed\n");
				return -EPERM;
			}

        case FSTSETCONF:
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_SET)))
			{
				if(card->state != FST_RUNNING)
				{
					printk_err("Attempt to configure card %d in non-running state (%d)\n", card->card_no, card->state);
					return -EIO;
				}
				memset ( &info, 0, sizeof ( struct fstioc_info ));
				if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
					retval = __copy_from_user ( &info,  data, sizeof ( info ));
#else
					retval = __copy_from_user ( &info,  ifr->ifr_data, sizeof ( info ));
#endif
				else
				{
#ifndef RHEL_RELEASE_CODE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0)
					if (!access_ok(ifr->ifr_data, sizeof(info)))
#else
					if (!access_ok(data, sizeof(info)))
#endif
#else
					if (!access_ok(VERIFY_READ, ifr->ifr_data, sizeof(info)))
#endif
#elif (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(8,1)))
#if (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(9,0))
					if (!access_ok(ifr->ifr_data, sizeof(info)))
#else
					if (!access_ok(data, sizeof(info)))
#endif
#else
					if (!access_ok(VERIFY_READ, ifr->ifr_data, sizeof(info)))
#endif
					{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
						printk("%s: fstsetconf: Can't access address %p\n", dev->name, data);
#else
						printk("%s: fstsetconf: Can't access address %p\n", dev->name, ifr->ifr_data);
#endif
						return -EFAULT;
					}
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
					retval = copy_from_user ( &info,  data,
#else
					retval = copy_from_user ( &info,  ifr->ifr_data,
#endif
												sizeof ( info ));
				}
				if (retval)
				{
					return -EFAULT;
				}
				if (validate_buffer_config(&info))
				{
					return -EFAULT;
				}
				if (flex_set_config(card, &info))
				{
					return 0;
				}
				else
				{
					return -EFAULT;
				}
			}
			else
			{
				printk("Permission for FSTSETCONF not allowed\n");
				return -EPERM;
			}

        case FSTSNOTIFY:
			/*
			* The application layer above wants to toggle the state notify
			* mechanism.
			*/
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_SET)))
			{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if (copy_from_user(&port->notify_mode, data, sizeof( int )))
#else
				if (copy_from_user(&port->notify_mode, ifr->ifr_data, sizeof( int )))
#endif
				{
					return -EFAULT;
				}
				fst_dbg(DBG_IOCTL, "%s: Setting Notify mode to %d\n", 
						port->dev->name, port->notify_mode);
						return 0;
			}
			else
			{
				printk("Permission for FSTSNOTIFY not allowed\n");
				return -EPERM;
			}

        case FSTSETMON:
			/*
			* The application layer above wants to monitor tx and rx data
			* mechanism.
			*/
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_SET)))
			{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if (copy_from_user(&port->monitor_mode, data, sizeof( int )))
#else
				if (copy_from_user(&port->monitor_mode, ifr->ifr_data, sizeof( int )))
#endif
				{
					return -EFAULT;
				}
				fst_dbg(DBG_IOCTL, "Card %d port %d: Setting Monitor mode to %d\n", 
						card->card_no, port->index, port->monitor_mode);
				return 0;
			}
			else
			{
				printk("Permission for FSTSETMON not allowed\n");
				return -EPERM;
			}

        case FSTGSTATE:
			/*
			* The application  layer above wants the carrier state
			*/
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_GET)))
			{
				state.carrier_state =  netif_carrier_ok ( port->dev );
				/*
				* and the txq is a simple length
				*/
				if ((state.txq_length = port->txqe - port->txqs) < 0)
				{
					/*
					* This is the case where the next free has wrapped but the
					* last used hasn't
					*/
					state.txq_length = state.txq_length + FST_TXQ_DEPTH;
				}
#ifdef FST_TTY_CHAR
                state.rxq_length = 0;
#else
				state.rxq_length = char_check_rx_queue_length (port);
#endif
				if (port->notify_mode == FST_NOTIFY_EXTENDED)
				{
					/*
					 * Update the stats
					 */
#ifdef IS_GEN_HDLC
					memcpy(&state.stats, hdlc_stats(port->dev), 
							sizeof(struct fst_device_stats));
#else
					memcpy(&state.stats, &port->stats, 
							sizeof(struct fst_device_stats));
#endif
					len = sizeof (struct fstioc_status);
				}
				else
				{
					len = FST_NOTIFY_BASIC_SIZE;
				}
				if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
					memcpy ( data, &state, len);
				else if ( copy_to_user ( data, &state, len))
#else
					memcpy ( ifr->ifr_data, &state, len);
				else if ( copy_to_user ( ifr->ifr_data, &state, len))
#endif
				{
					return -EFAULT;
				}
				return 0;
			}
			else
			{
				printk("Permission for FSTGSTATE not allowed\n");
				return -EPERM;
			}

        case FSTSYSREQ:
			/*
			 * Pass the command through to the card.  If the command generates a
			 * response, wait for it.  The application suspends until the command
			 * completes.
			 */
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & FST_ALLOW_IOCTL_SET)))
			{
				fst_dbg(DBG_IOCTL, "FSTSYSREQ received\n");
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( copy_from_user(sys_request, data, sizeof(struct fstioc_req)))
#else
				if ( copy_from_user(sys_request, ifr->ifr_data, sizeof(struct fstioc_req)))
#endif
				{
					fst_dbg(DBG_ASS, "Could not copy sysrequest from user\n");
					return -EFAULT;
				}
				fst_dbg(DBG_IOCTL, "The request is as follows:\n");
				fst_dbg(DBG_IOCTL, "msg_type = %x\n", sys_request->msg_type);
				fst_dbg(DBG_IOCTL, "Sending the card the request\n");

				flex_fstioc_req_to_le(sys_request);
				retval = usb_control_msg(card->udev, 
											usb_sndctrlpipe(card->udev,0),
											USS_CMD_SEND_SECURE_MSG, 0x40, 0, 0, 
											sys_request, 
											sizeof(struct fstioc_req), 10000);
				flex_fstioc_req_to_cpu(sys_request);
				if (retval != sizeof(struct fstioc_req))
				{
					printk("%s: Error in sending USB Control Message, retval is %d\n", 
							port->dev->name, retval);
					return retval;
				}
				/*
				* We are always going to wait for a reply
				*/
				fst_dbg(DBG_IOCTL, "Waiting for a reply\n");
				flex_fstioc_req_to_le(sys_request);
				retval = usb_control_msg(card->udev, 
											usb_rcvctrlpipe(card->udev,0),
											USS_CMD_RECV_SECURE_RSP, 0xC0, 0, 0, 
											sys_request, 
											sizeof(struct fstioc_req), 10000);
				flex_fstioc_req_to_cpu(sys_request);

				if (retval  != sizeof(struct fstioc_req))
				{
					printk("fstsysreq: Retval is %d\n", retval);
					return retval;
				}
				/*
				 * Copy the buffer back and we are complete
				 */
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				if ( copy_to_user ( data, sys_request, 
#else
				if ( copy_to_user ( ifr->ifr_data, sys_request, 
#endif		
									sizeof ( struct fstioc_req )))
				{
					fst_dbg(DBG_IOCTL, "Error copying data back to user\n");
					return -EFAULT;
				}
				fst_dbg(DBG_IOCTL, "Returning from FSTSYSREQ\n");
				return 0;
			}
			else
			{
				printk("Permission for FSTSYSREQ not allowed\n");
				return -EPERM;
			}

		case FSTCMD:
			/*
			* A general purpose command/response mechanism
			* First, get it.
			*/
			if ((capable(CAP_NET_ADMIN) || (fst_allow_ioctl & 
							(FST_ALLOW_IOCTL_SET | FST_ALLOW_IOCTL_GET))))
			{
				if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
					memcpy (&my_cmd, data, sizeof( my_cmd ));
				else if (copy_from_user(&my_cmd, data, sizeof( my_cmd )))
#else
					memcpy (&my_cmd, ifr->ifr_data, sizeof( my_cmd ));
				else if (copy_from_user(&my_cmd, ifr->ifr_data, sizeof( my_cmd )))
#endif
				{
					fst_dbg(DBG_ASS, "fstcmd: Could not access user buffers\n");
					return -EFAULT;
				}
				if (my_cmd.version != 1)
				{
					fst_dbg(DBG_ASS, "fstcmd: bad version %d\n", my_cmd.version);
					return -EINVAL;
				}
				switch (my_cmd.command)
				{
					case FSTCMD_GET_SERIAL:
						if ( copy_to_user ( my_cmd.data_ptr, 
											card->ports[0]->usb_card_info.szSerialNo, 
											strlen(card->ports[0]->usb_card_info.szSerialNo)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy serial back\n");
							return -EFAULT;
						}
						my_cmd.output_data_len = strlen(card->ports[0]->usb_card_info.szSerialNo); 
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1)			
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy(data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd, sizeof(my_cmd)))
#else
							memcpy(ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd, sizeof(my_cmd)))
#endif
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						fst_dbg(DBG_IOCTL, "Returning serial number %s", 
											card->ports[0]->usb_card_info.szSerialNo);              
						return 0;
              
					case FSTCMD_SET_V24O:
						if (my_cmd.input_data_len <1)
						{
							fst_dbg(DBG_ASS, "(%s) Cannot set V24 signals , no data provided %d\n", 
												port->dev->name, my_cmd.input_data_len);
							return -EFAULT;
						}
						if ( copy_from_user ( &signals, my_cmd.data_ptr, 1))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not set v24 signals\n");
							return -EFAULT;
						}
						if (signals > (OPSTS_RTS + OPSTS_DTR + OPSTS_DSRS + OPSTS_SS +
										OPSTS_LL + OPSTS_DCD + OPSTS_RL))
						{
							fst_dbg(DBG_ASS, "fstcmd: v24 signals not valid %d\n", signals);
							return -EFAULT;
						}
						sigs[0]=signals;
						sigs[1]=0x00;
						sigs[2]=0x00;
						sigs[3]=0x00;
						fst_dbg(DBG_IOCTL, "Setting V24 Signals %x\n", sigs[0]);
						retval = usb_control_msg(card->udev, 
													usb_sndctrlpipe(card->udev, 0),
													USS_CMD_SET_SIGNALS, 0x40, 0, 0, 
													sigs, CMD_BUF_LEN, 10000);
						if (retval==CMD_BUF_LEN) 
						{
							fst_dbg(DBG_IOCTL, "Set signals down response received\n");
							card->ports[0]->v24OpSts = signals & 0x03;            
						}
						else
						{
							fst_dbg(DBG_ASS, "Set signals down error errno = %d\n", retval);
							return -EFAULT;
						}
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy ( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd, sizeof(my_cmd)))
#else
							memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd, sizeof(my_cmd)))
#endif
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_GET_VERSION:
						if ( copy_to_user ( my_cmd.data_ptr, &port->fstioc_info_ver, 1))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy fstioc_info version back\n");
							return -EFAULT;
						}
						my_cmd.output_data_len = sizeof(fstioc_info_ver);
						/*
						* And invert the status field to show we have 
						* Understood and complete the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy ( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd, sizeof(my_cmd)))
#else
							memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd, sizeof(my_cmd)))
#endif
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_SET_VERSION:
						if (my_cmd.input_data_len > 0)
						{
							if ( copy_from_user ( &fstioc_info_ver, 
													my_cmd.data_ptr, 
													1))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not get fstioc_info version\n");
								return -EFAULT;
							}
							if ((fstioc_info_ver < FST_VERSION_OLD) ||
								(fstioc_info_ver > FST_VERSION_CURRENT))
							{
								fst_dbg(DBG_ASS, "fstcmd: setting an invalid fstioc_info version %d \n", fstioc_info_ver);
								return -EFAULT;
							}
							my_cmd.output_data_len = 0;
							/*
							* And invert the status field to show we have done 
							* the command
							*/
							my_cmd.status = ~my_cmd.status;
							if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
								memcpy ( data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( data, &my_cmd,
#else
								memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
														sizeof(my_cmd)))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
								return -EFAULT;
							}
							port->fstioc_info_ver = fstioc_info_ver;
						}
						else
						{
							fst_dbg(DBG_ASS, "(%s) Cannot copy fstioc_info version, data buffer too small %d %d\n", 
											port->dev->name, 1, 
											my_cmd.input_data_len);
							return -EFAULT;
						}
						return 0;

					case FSTCMD_GET_INTS:
						if(port->compat==1)
							memcpy ( my_cmd.data_ptr, &fst_int_counter[port->card->card_no], 
													sizeof(struct fst_ints)); 
						else
							copied = copy_to_user ( my_cmd.data_ptr,
													&fst_int_counter[port->card->card_no], 
													sizeof(struct fst_ints));
						if ( copied )
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy all the int stats back %lu\n", copied);
							return -EFAULT;
						}
						my_cmd.output_data_len = sizeof(struct fst_ints);  
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy ( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd, sizeof(my_cmd)))
#else
							memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd, sizeof(my_cmd)))
#endif
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
              
					case FSTCMD_RESET_INTS:
						memset(&fst_int_counter[port->card->card_no], 0, sizeof(struct fst_ints));
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy ( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd,
#else
							memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
									sizeof(my_cmd)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_RESET_STATS:
#ifdef IS_GEN_HDLC
						memset(hdlc_stats(port->dev), 0, sizeof(struct fst_device_stats));
#else
						memset(&port->stats, 0, sizeof(struct fst_device_stats));
#endif
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy ( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd,
#else
							memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
					    else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
													sizeof(my_cmd)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_SET_READV:
						if (my_cmd.input_data_len > 0)
						{
							if ( copy_from_user ( &readv_mode, 
													my_cmd.data_ptr, 
													1))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not get readv_mode\n");
								return -EFAULT;
							}
							fst_dbg(DBG_IOCTL, "Setting readv mode to %d\n", 
													readv_mode);
							if ((readv_mode < FST_READV_NORMAL) ||
										(readv_mode > FST_READV_SYNC2))
							{
								fst_dbg(DBG_ASS, "fstcmd: setting an invalid readv mode  %d \n", readv_mode);
								return -EFAULT;
							}
							my_cmd.output_data_len = 0;
							/*
							* And invert the status field to show we have done 
							* the command
							*/
							my_cmd.status = ~my_cmd.status;
							if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
								memcpy ( data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( data, &my_cmd,
#else
								memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
															sizeof(my_cmd)))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
								return -EFAULT;
							}
							port->readv_mode = readv_mode;
						}
						else
						{
							fst_dbg(DBG_ASS, "(%s) Cannot copy readv_mode, data buffer too small %d %d\n", 
										  port->dev->name, 1, 
										  my_cmd.input_data_len);
							return -EFAULT;
						}
						return 0;

					case FSTCMD_SET_CHAR:
						if (my_cmd.input_data_len <= sizeof(struct fstioc_char_data))
						{
							if ( copy_from_user ( &char_data, 
													my_cmd.data_ptr, 
													sizeof(struct fstioc_char_data)))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not get char data\n");
								return -EFAULT;
							}
							fst_dbg(DBG_IOCTL, "Setting char data to %d %d\n", 
											char_data.queue_len, char_data.threshold);
							port->char_inq_max_len = char_data.queue_len;
							port->char_inq_threshold = char_data.threshold;
							my_cmd.output_data_len = 0;
							/*
							* And invert the status field to show we have done 
							* the command
							*/
							my_cmd.status = ~my_cmd.status;
							if(port->compat==1)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
								memcpy ( data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( data, &my_cmd,
#else
								memcpy ( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
							else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
														sizeof(my_cmd)))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
								return -EFAULT;
							}
						}
						else
						{
							fst_dbg(DBG_ASS, "(%s) Cannot copy char_data, data buffer too small %d %d\n", 
												port->dev->name, 1, 
												my_cmd.input_data_len);
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_GET_PRESERVE_SIGNALS:
						fst_issue_cmd (port, GETPRESERVE);
						if (port->v24OpSts & 0x100)
							signals = 1;
						else
							signals = 0;
						if ( copy_to_user ( my_cmd.data_ptr, &signals, 1))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not return preserve signals\n");
							return -EFAULT;
						}
						my_cmd.output_data_len = sizeof(signals);	
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd, sizeof(my_cmd)))
#else
							memcpy( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd, sizeof(my_cmd)))
#endif		  
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
              
					case FSTCMD_SET_PRESERVE_SIGNALS:
						fst_issue_cmd ( port, SETPRESERVE );
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd,
#else
							memcpy( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif		    
												sizeof(my_cmd)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					case FSTCMD_UPDATE_CLOCK:
						if (my_cmd.input_data_len > 0)
						{
							if ( copy_from_user ( &speed, 
													my_cmd.data_ptr, 
													sizeof(int)))
							{
								fst_dbg(DBG_ASS, "fstcmd: could not get speed setting\n");
								return -EFAULT;
							}
						}
						else
						{
							fst_dbg(DBG_ASS, "(%s) Cannot copy speed setting, data buffer too small %d %d\n", 
														port->dev->name, 1, 
														my_cmd.input_data_len);
							return -EFAULT;
						}
						fst_dbg(DBG_IOCTL, "%s: Update clock speed to %d\n", port->dev->name, speed);
						port->usb_config.lineSpeed = speed;
						fst_issue_cmd ( port, SETLINESPEED );
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd,
#else
							memcpy( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
													sizeof(my_cmd)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;

					case FSTCMD_SET_CUSTOM_RATE:
						if (!port->features)
						{
							printk("%s: custom clock rates not supported on this port\n",
										port->dev->name);
							return -EIO;
						}
						if (my_cmd.input_data_len >= (unsigned int)sizeof(struct fstioc_custom_rate_config))
						{
							if ( copy_from_user ( &custom_clock_rate, 
													my_cmd.data_ptr, 
													sizeof(struct fstioc_custom_rate_config)))
							{
								printk("fstcmd: could not get custom rate config structure\n");
								return -EFAULT;
							}
						}
						else
						{
							printk("(%s) Cannot copy custom_rate_config, data buffer too small %lu %u\n", 
										port->dev->name, sizeof(struct fstioc_custom_rate_config), 
										my_cmd.input_data_len);
							return -EFAULT;
						}
						retval = parse_custom_clock_rate_structure(port, &custom_clock_rate);
						if (retval)
						{
							printk("(%s) Invalid custom_rate_config structure, %d\n", 
 											port->dev->name, retval);
							return -EINVAL;
						}
						retval = set_custom_clock_rate(port, &custom_clock_rate);
						if (retval)
						{
							printk("(%s) Error in setting custom clock rate, %d\n", 
 											port->dev->name, retval);
							return -EINVAL;
						}
						/*
						* Everything worked
						*/
						my_cmd.output_data_len = 0;
						/*
						* And invert the status field to show we have done 
						* the command
						*/
						my_cmd.status = ~my_cmd.status;
						if(port->compat==1) 
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
							memcpy( data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( data, &my_cmd,
#else
							memcpy( ifr->ifr_data, &my_cmd, sizeof(my_cmd));
						else if ( copy_to_user ( ifr->ifr_data, &my_cmd,
#endif
														sizeof(my_cmd)))
						{
							fst_dbg(DBG_ASS, "fstcmd: could not copy status back\n");
							return -EFAULT;
						}
						return 0;
		  
					default:
						fst_dbg(DBG_ASS, "Unrecognised FST Command %d\n", my_cmd.command);
						return -EINVAL;
				}
			}
			else
			{
				printk("Permission for FSTCMD not allowed\n");
				return -EPERM;
			}

		default:
			/* Not one of ours. Pass it through to sPPP package */
			if ( (port->proto == FST_HDLC || port->proto == FST_PPP) 
					&& (port->run) && (!port->char_file) )
			{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
				return -EINVAL;
#else
#ifdef FST_NET_DEV
				fst_dbg(DBG_ASS, "Passing this ioctl on to sppp %d \n", cmd );
				return sppp_do_ioctl ( dev, ifr, cmd );
#endif
#endif
			}
			else
				return -EINVAL;
	}

	return -EINVAL;
}

#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
static int
fst_ioctl ( struct net_device *dev, struct ifreq *ifr, int cmd )
{
  return fst_siocdevprivate(dev, ifr, ifr->ifr_data, cmd);
}

static int
fst_ifs_ioctl(struct net_device *dev, struct if_settings *ifs)
{
	struct fst_card_info *card;
	struct fst_port_info *port;
	//int retval = 0;

	fst_dbg(DBG_IOCTL, "***ioctl, %x\n", ifs->type);

	port = dev_to_port(dev);
	card = port->card;

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;
#if 0
	switch ( ifs->type )
	{
		case IF_GET_IFACE:
			return fst_get_iface ( card, port, ifs );

		case IF_IFACE_SYNC_SERIAL:
		case IF_IFACE_V35:
		case IF_IFACE_V24:
		case IF_IFACE_X21:
		case IF_IFACE_X21D:
		case IF_IFACE_T1:
		case IF_IFACE_E1:
		case IF_IFACE_SHDSL:
		case IF_IFACE_RS530_449:
		case IF_IFACE_RS485:
		case IF_IFACE_RS485_FDX:
		case IF_IFACE_UX35C:
			return fst_set_iface ( card, port, ifs );

		case IF_PROTO_RAW:
			port->proto = FST_RAW;
			return 0;

		case IF_GET_PROTO:
			if (port->proto == FST_RAW)
			{
				ifs->type = IF_PROTO_RAW;
				return 0;
			}
			retval = hdlc_ioctl ( dev, ifs );
			if (retval == -EINVAL)
			{
				/*
				* Protocol not set yet
				*/
				port->proto = FST_GEN_HDLC;
				ifs->type = FST_GEN_HDLC;
				retval = 0;
			}
			return retval;

		default:
			port->proto = FST_GEN_HDLC;
			fst_dbg(DBG_IOCTL, "Passing this type to hdlc %x\n",
					ifs->type);
			port->hdlc_proto = ifs->type;
			return hdlc_ioctl ( dev, ifs );
	}
#endif
	return 0;
}
#endif

static void
fst_openport ( struct fst_port_info *port )
{
	int txq_length;

        /* Only init things if card is actually running. This allows open to
         * succeed for downloads etc.
         */
        fst_dbg(DBG_OPEN, "Opening port %s\n", port->dev->name);
        fst_dbg(DBG_OPEN, "Tx buffers = %d of size %d\n", port->num_tx_buffers,
               port->tx_buffer_size);
        fst_dbg(DBG_OPEN, "Rx buffers = %d of size %d\n", port->num_rx_buffers,
               port->rx_buffer_size);

        last_segment_cnt = 0;
        if ( port->card->state == FST_RUNNING )
        {
          if ( port->run )
            {
              fst_dbg ( DBG_OPEN,"open: found port already running\n");
            }
          else
            {
              fst_rx_config ( port );
              fst_tx_config ( port );
              fst_issue_cmd ( port, STARTPORT );
              port->run = 1;
              fst_op_raise (port, OPSTS_RTS | OPSTS_DTR);
              txq_length = port->txqe - port->txqs;
              port->txqe = 0;
              port->txqs = 0;
              port->rxq.error_recovery = 0;
              port->char_inq = NULL;
              port->char_inq_end = NULL;
            }
        }
}

static void
fst_closeport ( struct fst_port_info *port )
{
        fst_dbg(DBG_OPEN, "Closing port %s\n", port->dev->name);
        if ( port->card->state == FST_RUNNING )
        {
          /*
           * Need to lower control signals here
           */
	  fst_op_lower ( port, OPSTS_RTS | OPSTS_DTR );
          fst_issue_cmd ( port, STOPPORT );
	  if (port->mode == FST_MODE_ASYNC)
	    {
	      fst_reset_rx_fifo(port);
	    }
	  fst_issue_cmd (port, FLUSHBUFFERS);
          port->run = 0;
          purge_tx_queue(port->card, port);
        }
        else
          {
            fst_dbg ( DBG_OPEN,"close: card not running\n");
          }
}


static int
fst_open ( struct net_device *dev )
{
        struct fst_card_info *card;
        struct fst_port_info *port;
        int orig_mtu;
#ifdef FST_NET_DEV
        int err;
#endif

        port = netdev_priv(dev);
        card = port->card;
	fst_dbg(DBG_OPEN, "fst_open on port %s\n", dev->name);

        if (card->state != FST_RUNNING)
          {
            fst_dbg(DBG_ASS, "%s: Cannot open port if card is not loaded\n", dev->name);
            return -ENODEV;
          }
        if ((port->run) && (port->char_file))
          {
	    if (!port->monitor_mode)
	      {
		fst_dbg(DBG_ASS, "%s: Port is already open\n", dev->name);
		return -EBUSY;
	      }
	    else
	      {
		fst_dbg(DBG_ASS, "%s: Opening network port to monitor char interface\n",
			dev->name);
	      }
          }

	if (!try_module_get(THIS_MODULE))
          return -EBUSY;

        switch ( port->proto )
        {
        case FST_HDLC:
        case FST_PPP:

                orig_mtu = dev->mtu;
#ifdef FST_NET_DEV
                /* Attach to sync PPP module */
                port->pppdev.dev = dev;
                sppp_attach ( &port->pppdev );
                if ( orig_mtu < dev->mtu )
                        dev->change_mtu ( dev, orig_mtu );

                /* Claw back the ioctl routine. We promise to be good and call
                 * the sync PPP routines once we've eliminated our functions.
                 */
                dev->do_ioctl = fst_ioctl;
                err = sppp_do_ioctl ( dev, NULL, port->proto == FST_HDLC
                                                ? SPPPIOCCISCO : SPPPIOCPPP );
                if ( err )
		{
			sppp_detach ( dev );
                        module_put(THIS_MODULE);
			return err;
		}
                err = sppp_open ( dev );
                if ( err )
		{
			sppp_detach ( dev );
                        module_put(THIS_MODULE);
			return err;
		}
#endif
                break;

        case FST_MONITOR:
        case FST_RAW:
                break;

        default:
                fst_dbg ( DBG_OPEN,"open: Unknown proto %d\n", port->proto );
                break;
        }
        if (!port->char_file)
          fst_openport ( port );
        netif_wake_queue ( dev );
        return 0;
}

static int
fst_close ( struct net_device *dev )
{
        struct fst_port_info *port;
        struct fst_card_info *card;
	unsigned long flags;

        port = netdev_priv(dev);
        if (port == NULL)
        {
          fst_dbg(DBG_ASS, "fst_close: port is NULL\n");
        }
	card = port->card;
        if (card == NULL)
        {
          fst_dbg(DBG_ASS, "fst_close: card is NULL\n");
        }
	fst_dbg(DBG_OPEN, "Port Close: %s\n", dev->name);
	netif_stop_queue ( dev );

	if (!port->char_file)
        {
	if (card->tx_urb_cnt != MAX_URBS_OUTSTANDING)
	  {
            if (card->tx_bulk_urb)
              {
                fst_dbg(DBG_OPEN, "Cancel a Tx urb\n");
                usb_kill_urb(card->tx_bulk_urb); /* Stop sending data */
              }
	    fst_dbg(DBG_ASS, "Closing port with urbs outstanding\n");
	  }
	    /*
             * Only close the port if really not in use
             */
          fst_closeport ( port );
	  spin_lock_irqsave ( &card->card_lock, flags );
          while (port->txqe != port->txqs)
            {
              fst_dbg(DBG_OPEN, "Purging tx queue\n");
              fst_dbg(DBG_OPEN, "%s: Closing port with data in tx queue\n", dev->name);
              dev_kfree_skb(port->txq[port->txqs].frame);
              port->txq[port->txqs].frame = NULL;
              port->txqs++;
              if (port->txqs == FST_TXQ_DEPTH)
                {
                  port->txqs = 0;
                }
            }
	  spin_unlock_irqrestore ( &card->card_lock, flags );
        }
	  
        switch ( port->proto )
        {
        case FST_HDLC:
        case FST_PPP:
#ifdef FST_NET_DEV
                fst_dbg(DBG_OPEN, "Closing sppp connection\n");
                sppp_close ( dev );
                sppp_detach ( dev );
#endif
                break;

        case FST_MONITOR:
        case FST_RAW:
                break;

        default:
                fst_dbg ( DBG_OPEN,"close: Unknown proto %d\n", port->proto );
                break;
        }

        module_put(THIS_MODULE);
        return 0;
}

/* 401: ndo_tx_timeout changes for >= 5.6 (and 5.3 for suse) */
#ifndef RHEL_RELEASE_CODE
#ifndef CONFIG_SUSE_VERSION
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
static void fst_tx_timeout ( struct net_device *dev, unsigned int txqueue )
#else
static void fst_tx_timeout ( struct net_device *dev )
#endif
#else
//suse 15.3 does not have ndo_tx_timeout def that agrees with kernel version!
#if ((CONFIG_SUSE_VERSION >= 15) && (CONFIG_SUSE_PATCHLEVEL >= 3))
static void fst_tx_timeout ( struct net_device *dev, unsigned int txqueue )
#else
static void fst_tx_timeout ( struct net_device *dev )
#endif
#endif
#elif (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(8,3)))
static void fst_tx_timeout ( struct net_device *dev, unsigned int txqueue )
#else
static void fst_tx_timeout ( struct net_device *dev )
#endif
{
        struct fst_port_info *port;
        struct fst_card_info *card;

	printk("fst_tx_timeout on %s\n", dev->name);
        port = netdev_priv(dev);
        card = port->card;
        port->stats.tx_errors++;
        port->stats.tx_aborted_errors++;
	fst_dbg(DBG_ASS, "Tx timeout card %d port %d\n", 
	    card->card_no, port->index);
        //fst_issue_cmd ( port, ABORTTX );
	purge_tx_queue(card, port);
        netif_wake_queue ( dev );
	port->start = 0;
	if (port->notify_mode == FST_NOTIFY_EXTENDED)
	  fst_notify_state_change ( port->card, port );
}


static int
fst_start_xmit ( struct sk_buff *skb, struct net_device *dev )
{
        struct fst_card_info *card;
        struct fst_port_info *port;
        unsigned long flags;
	int txq_length;
        struct sk_buff *new_skb;
        int count = 0;

        port = netdev_priv(dev);
        if (port == NULL)
        {
          fst_dbg(DBG_ASS, "fst_start_xmit: port is null\n");
          dev_kfree_skb ( skb );
          port->stats.tx_errors++;
          port->stats.tx_dropped++;
          return 0;
        }
        if (dev == NULL)
        {
          fst_dbg(DBG_ASS, "fst_start_xmit: dev is null\n");
          dev_kfree_skb ( skb );
          port->stats.tx_errors++;
          port->stats.tx_dropped++;
          return 0;
        }
        card = port->card;
        if (card == NULL)
        {
            printk("fst_start_xmit: card is null\n");
            dev_kfree_skb ( skb );
            port->stats.tx_errors++;
            port->stats.tx_dropped++;
            return 0;
        }
	fst_dbg(DBG_TX, "fst_start_xmit: length = %d\n", skb->len);

        /* Drop packet with error if we don't have carrier */
        if (( ! netif_carrier_ok ( dev )) && (!port->ignoreCarrier))
        {
                dev_kfree_skb ( skb );
                port->stats.tx_errors++;
                port->stats.tx_carrier_errors++;
		fst_dbg(DBG_ASS, "Tried to transmit but no carrier on card %d port %d\n",
		    card->card_no, port->index);
		if (port->notify_mode == FST_NOTIFY_EXTENDED)
		  fst_notify_state_change ( port->card, port );
                return 0;
        }

        /* Drop it if it's too big! MTU failure ? */
        if ( skb->len > FST_MAX_MTU )
        {
                printk_err("Tx Packet too large %d vs %d\n", skb->len,
                           FST_MAX_MTU );
                dev_kfree_skb ( skb );
                port->stats.tx_errors++;
		port->stats.tx_dropped++;
		if (port->notify_mode == FST_NOTIFY_EXTENDED)
		  fst_notify_state_change ( port->card, port );
                return 0;
        }

        /* Drop it if it's bigger than the tx_buffer_size */
        if ( skb->len > port->tx_buffer_size )
        {
                printk_err("Tx Packet too large %d vs buffer %d\n", skb->len,
                           port->tx_buffer_size );
                dev_kfree_skb ( skb );
                port->stats.tx_errors++;
		port->stats.tx_dropped++;
		if (port->notify_mode == FST_NOTIFY_EXTENDED)
		  fst_notify_state_change ( port->card, port );
                return 0;
        }

        /* Drop it if it's port not running */
        if ( !port->run )
        {
                printk_err("Tx When port closed\n");
                dev_kfree_skb ( skb );
                port->stats.tx_errors++;
		port->stats.tx_dropped++;
		if (port->notify_mode == FST_NOTIFY_EXTENDED)
		  fst_notify_state_change ( port->card, port );
                return 0;
        }
	count = skb->len;
	/* Drop it if it is too small */
	if (skb->len <= 0)
	  {
	    printk("fst_start_xmit: packet too small %d\n", skb->len);
            dev_kfree_skb ( skb );
            port->stats.tx_errors++;
            port->stats.tx_dropped++;
	    return 0;
	  }
	if (skb->len == 0)
	  printk("Transmitting a zero length frame\n");

	/*
	 * We are always going to queue the packet
	 * so that the bottom half is the only place we tx from
	 * Check there is room in the port txq
	 */
        spin_lock_irqsave ( &card->card_lock, flags );
	if ((txq_length = port->txqe - port->txqs) < 0)
	  {
	    /*
	     * This is the case where the next free has wrapped but the
	     * last used hasn't
	     */
	    txq_length = txq_length + FST_TXQ_DEPTH;
	  }
	spin_unlock_irqrestore ( &card->card_lock, flags );
	if (txq_length >= fst_txq_high)
	  {
	    /*
	     * We have got enough buffers in the pipeline.  Ask the network
	     * layer to stop sending frames down
	     */
            port->start = 1;   /* I'm using this to signal stop sent up */
	    if (!port->char_file)
	      {
                fst_dbg(DBG_ASS, "%s: Stop the network layer\n", dev->name);
                netif_stop_queue ( dev );
              }
              else
              {
                /*
                 * If the char interface say we can't do it at all
                 */
                dev_kfree_skb ( skb );
                return 0;
              }
	  }
	if (txq_length == FST_TXQ_DEPTH - 1)
	  {
	    /*
	     * This shouldn't have happened but such is life
	     */
	    dev_kfree_skb ( skb );
	    port->stats.tx_errors++;
            port->stats.tx_dropped++;
	    fst_dbg(DBG_ASS, "Tx queue overflow card %d port %d\n",
		card->card_no, port->index);
	    if (port->notify_mode == FST_NOTIFY_EXTENDED)
	      fst_notify_state_change ( port->card, port );
	    return 0;
	  }

	/*
	 * queue the buffer
	 */
        new_skb = skb;
        if (new_skb == NULL)
          {
            fst_dbg(DBG_ASS, "Error in queuing atm pdu\n");
            return 0;
          }
        if (port->monitor_mode)
          {
            gen_mon_packet(new_skb->data, new_skb->len, port, FST_MON_TX);
          }
        spin_lock_irqsave ( &card->card_lock, flags );
	port->txq[port->txqe].frame = new_skb;
        port->txq[port->txqe].count = new_skb->len;
	port->txq[port->txqe].segment_cnt = 
          ((new_skb->len-1)/port->tx_buffer_size) + 1 ;
	port->txq[port->txqe].current_seg = 0;
	port->txq[port->txqe].flags = 0;
	port->txqe++;
	if (port->txqe == FST_TXQ_DEPTH)
	  port->txqe = 0;
	spin_unlock_irqrestore ( &card->card_lock, flags );
	
	/* Scehdule the bottom half which now does transmit processing */
        fst_q_work_item(&fst_work_txq, card->card_no);
	tasklet_schedule(&fst_tx_task); 
        if (port->char_file)
          return count;
        else
          return 0;
}

static struct net_device_stats *
fst_get_stats ( struct net_device *dev )
{
        struct fst_port_info *port;

        if (( port = netdev_priv(dev)) != NULL )
                return &port->stats;
        else
                return NULL;
}

/*
 * Sysdev stuff for adding the serial number attribute to the char device
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
static ssize_t serial_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct fst_card_info *card = dev_get_drvdata(dev);
#else
static ssize_t serial_show(struct class_device *dev, char *buf)
{
	struct fst_card_info *card = class_get_devdata(dev);
#endif

	ssize_t ret = 0;

	/* no lock needed for this */

	strcpy(buf, card->ports[0]->usb_card_info.szSerialNo);
	ret = strlen(buf);
	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
static DEVICE_ATTR(serial, S_IRUGO, serial_show, NULL);
#else
static CLASS_DEVICE_ATTR(serial, 0644, serial_show, NULL);
#endif
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
static const struct net_device_ops fst_ops = {
	.ndo_open       = fst_open,
	.ndo_stop       = fst_close,
	.ndo_start_xmit = fst_start_xmit,
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5,15,0)) || (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(9,0))))
	.ndo_siocwandev	= fst_ifs_ioctl,
	.ndo_siocdevprivate = fst_siocdevprivate,
#else
	.ndo_do_ioctl   = fst_ioctl,
#endif	
	.ndo_tx_timeout = fst_tx_timeout,
	.ndo_get_stats  = fst_get_stats,
};
#endif

static void fst_setup_netdev (struct net_device *dev)
{
	struct fst_port_info *port;

        port = netdev_priv(dev);
	if (port == NULL)
	  return;

	/* Fill in remainder of the net device info 
	 * Since this is a PCI setup this is purely
	 * informational. Give them the buffer addresses
	 * and basic card I/O.
	 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
        dev->netdev_ops         = &fst_ops;
#else
	dev->get_stats          = fst_get_stats;
	dev->priv               = port;
	dev->open               = fst_open;
	dev->stop               = fst_close;
	dev->do_ioctl           = fst_ioctl;
	dev->tx_timeout         = fst_tx_timeout;
	dev->hard_start_xmit    = fst_start_xmit;
#endif
	dev->mtu                = FST_DEF_MTU;
	dev->tx_queue_len       = FST_TX_QUEUE_LEN;
	dev->type               = ARPHRD_MYTYPE;
	dev->addr_len           = 8;
	dev->watchdog_timeo     = FST_TX_TIMEOUT;
	dev->flags              = IFF_POINTOPOINT|IFF_NOARP;
#ifdef FST_NET_DEV
	dev->priv               = port;
#endif
}

/*
 *      Card setup having checked hardware resources.
 *      Should be pretty bizarre if we get an error here (kernel memory
 *      exhaustion is one possibility). If we do see a problem we report it
 *      via a printk and leave the corresponding interface and all that follow
 *      disabled.
 */

static void 
fst_init_card ( struct fst_card_info *card, char *dev_name )
{
        int i;
        int err;
        struct net_device *dev;
	struct fst_port_info *port;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
        char char_dev_name[16];
#endif
#ifdef FST_TTY_CHAR
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
	struct device *tty_device = NULL;
#endif
#endif

	/*
	 * Assign the minor device number
	 */
        for (i=0; i < FST_MAX_CARDS*FST_MAX_PORTS; i++)
	  {
	    if (fst_ports_list[i] == NULL)
	      {
		fst_minor = i;
		break;
	      }
	  }
        /* We're working on a number of ports based on the card ID. If the
         * firmware detects something different later (should never happen)
         * we'll have to revise it in some way then.
         */
        spin_lock_init ( &card->card_lock );
        spin_lock_init ( &card->fifo_lock);
        init_waitqueue_head (&card->fifo_waitq);
        for ( i = 0 ; i < card->nports ; i++ )
        {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
               dev = alloc_netdev (sizeof(struct fst_port_info), 
				   dev_name,
				   NET_NAME_UNKNOWN,
                                   fst_setup_netdev);
#else
               dev = alloc_netdev (sizeof(struct fst_port_info), 
				   dev_name,
                                   fst_setup_netdev);
#endif
                if ( dev == NULL )
                {
                        printk_err ("Cannot allocate net_device for port %d\n",
                                        i );
                        /* No point going any further */
                        card->nports = i;
                        break;
                }
		port = netdev_priv(dev);
		card->ports[i] = port;
		memset(port, 0, sizeof(struct fst_port_info));
#ifdef FST_NET_DEV
                port->if_ptr = &card->ports[i]->pppdev;
#endif
		SET_NETDEV_DEV(dev, &card->udev->dev);
                port->card   = card;
                port->index  = i;
                port->proto  = FST_HDLC;
                port->run    = 0;
                port->num_tx_buffers = REQUIRED_TX_BUFFERS;
                port->num_rx_buffers = REQUIRED_RX_BUFFERS;
                port->tx_buffer_size = REQUIRED_TX_BUFFER_SIZE;
                port->rx_buffer_size = REQUIRED_RX_BUFFER_SIZE;
                port->fstioc_info_ver = fst_iocinfo_version;
                port->char_file = NULL;
                port->char_inq = NULL;
                port->char_inq_end = NULL;
                port->hwif = X21;
		port->v24OpSts = 0;
		port->minor_dev_no = fst_minor;
                spin_lock_init ( &port->rxf_lock );
                init_waitqueue_head ( &port->pollq );
                init_waitqueue_head ( &port->readq );
                init_waitqueue_head ( &port->writeq );
                port->dev = dev;
		fst_ports_list[fst_minor] = port;
                if (( err = register_netdev ( dev )) < 0 )
                {
                        printk_err ("Cannot register %s (errno %d)\n",
                                        dev->name, -err );
                        kfree ( dev );
                        card->nports = i;
                        break;
                }
		fst_dbg(DBG_INIT, "Allocated device name was %s\n", dev->name);
		if (fst_alloc_rx_fifo(port))
		  {
		    printk_err("%s: Cannot allocate rx fifo\n", port->dev->name);
                        kfree ( dev );
                        card->nports = i;
                        break;
		  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
		dev->ml_priv = port;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
                fst_dbg(DBG_INIT, "fst_major is %x: fst_minor is %d\n", 
                    fst_major, fst_minor);
#ifdef FST_TTY_CHAR
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
		fst_dbg(DBG_INIT, "Register the tty device %d\n", fst_minor);
		tty_port_init(&fst_tty_area[fst_minor].tty_port);
		tty_device = tty_port_register_device(&fst_tty_area[fst_minor].tty_port,
						      serial_drv, fst_minor, NULL);

		fst_tty_area[fst_minor].tty_port.tty = fst_tty_area[fst_minor].tty;
                strcpy(char_dev_name, "");
#else
                strcpy(char_dev_name, "tty_");
#endif
#else
                strcpy(char_dev_name, "");
#endif
                strcat(char_dev_name, dev->name);
		card->sync_cdev = CLASS_DEVICE_CREATE(farsync_class, 
						MKDEV(fst_major, fst_minor),
						NULL, char_dev_name,fst_minor);
		fst_dbg(DBG_INIT, "CLASS_DEVICE_CREATE: 0x%p\n", card->sync_cdev);   
                if (card->sync_cdev < 0)
                {
                       printk_err ("Cannot create udev entry for %s\n",
				   port_to_dev(card->ports[0])->name);
                }
		else
		{
		  int rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
		  dev_set_drvdata(card->sync_cdev, card);
		  /* register the attributes */
		  rc = device_create_file(card->sync_cdev, 
                                          &dev_attr_serial);
                 fst_dbg(DBG_INIT, "device_create_file: %d\n", rc);   
#else
		  class_set_devdata(card->sync_cdev, card);
		  /* register the attributes */
		  rc = class_device_create_file(card->sync_cdev, 
						&class_device_attr_serial);
#endif

		  if (rc)
		    {
		      printk("Couldn't register sync serial no attribute\n");
		    }
		}
#endif
		init_async_parameters(port, 8, COM_STOP_BITS_1, 
                                      COM_NO_PARITY, COM_FLOW_CONTROL_NONE);

		port->char_inq_max_len = FST_CHAR_INQ_MAX_LEN;
        }
}

/* Get a minor range for your devices from the usb maintainer */
#define USB_MINOR_BASE	192
#define USB_VENDOR_ID	0x1afd
#define USB_PRODUCT_ID	0x0721
#define WRITES_IN_FLIGHT	8

/* table of devices that work with this driver */
static struct usb_device_id flex_table [] = {
	{ USB_DEVICE(USB_VENDOR_ID, USB_PRODUCT_ID) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, flex_table);

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with the driver core
 */
static struct usb_class_driver flex_class = {
	.name =		"flex%d",
	.fops =		&fst_mem_fops,
	.minor_base =	USB_MINOR_BASE,
};

static int flex_get_config ( struct fst_card_info *usb_dev, 
                             struct fst_port_info *port,
                             struct fstioc_info *info)
{
        int retval   = 0;
	int retval2  = 0;
        int retval3  = 0;
	int retval4  = 1;
        int *clocks   = (int *) &port->scratch;
	int *features = (int *) &port->scratch;

	fst_dbg(DBG_IOCTL, "Getting port config %lu %lu\n",
	       sizeof(IREX), sizeof(struct fstioc_req));
	flex_serconf_to_le(&port->usb_config);
	fst_dbg(DBG_IOCTL, "Sending GET_CONFIG URB\n");
	retval = usb_control_msg(usb_dev->udev, usb_rcvctrlpipe(usb_dev->udev,
                                                                0),
				 USS_CMD_GET_CONFIG, 0xc0, 0, 0, 
                                 &port->usb_config, 
                                 sizeof(FLEX_SERCONF), 10000);
	fst_dbg(DBG_IOCTL, "Result was %d\n", retval);
	flex_serconf_to_cpu(&port->usb_config);
	flex_card_info_to_le(&port->usb_card_info);
	fst_dbg(DBG_IOCTL, "Sending GET_CARD_INFO URB\n");
	retval2 = usb_control_msg(usb_dev->udev, usb_rcvctrlpipe(usb_dev->udev,
								 0),
				  USS_CMD_GET_CARD_INFO, 0xc0, 0, 0, 
				  &port->usb_card_info, 
				  sizeof(FLEX_CARD_INFO), 10000);
	fst_dbg(DBG_IOCTL, "Result was %d\n", retval2);
	if (info)
	  {
	    info->cardRevMajor = port->usb_card_info.uMajorRev;
	    info->cardRevMinor = port->usb_card_info.uMinorRev;
	    info->cardRevBuild = port->usb_card_info.uBuildState;
	  }
	flex_card_info_to_cpu(&port->usb_card_info);
	cpu_to_le32s(clocks);
	fst_dbg(DBG_IOCTL, "Sending GET_CLOCK_STATUS URB\n");
	retval3 = usb_control_msg(usb_dev->udev, 
				 usb_rcvctrlpipe(usb_dev->udev, 0),
				 USS_CMD_GET_CLOCK_STATUS, 0xC0, 0, 0, 
				 clocks, 4, 10000);
	fst_dbg(DBG_IOCTL, "Result was %d\n", retval3);
	le32_to_cpus(&clocks);
	/*
	 * If we are a Flex 2 or above then check the features
	 */
 	if ((port->usb_card_info.uMajorRev > FIRST_FLEX_V2_MAJOR_ID) || 
	    (port->usb_card_info.uMinorRev >= FIRST_FLEX_V2_MINOR_ID))
	  {
	    cpu_to_le32s(&features);
	    fst_dbg(DBG_IOCTL, "Sending GET_FEATURE_MASK URB\n");
	    retval4 = usb_control_msg(usb_dev->udev, 
				      usb_rcvctrlpipe(usb_dev->udev, 0),
				      USS_CMD_GET_FEATURE_MASK, 0xC0, 0, 0, 
				      features, 4, 10000);
	    fst_dbg(DBG_IOCTL, "Result was %d\n", retval4);
	    le32_to_cpus(&features);
	  }
	if ((retval>0) && (retval2>0) && (retval3>0) && (retval4>0) && (info != NULL))
         {
            fst_dbg(DBG_IOCTL, "Get config response received\n");
      	    info->valid
                = (( usb_dev->state == FST_RUNNING ) ? FSTVAL_ALL : FSTVAL_CARD )
#if FST_DEBUG
                | FSTVAL_DEBUG
#endif
                ;
            info->valid = 1;
            if (*clocks)
              info->clockStatus = 0;
            else
              info->clockStatus = 1;
	    info->synthAbility = 0;
	    if (*features)
	      {
		/*
		 * This means it is a flex v2 or later, so synthisiser
		 * is supported
		 */
		info->synthAbility = 1;
	      }
	    port->features = *features;
	    info->features = (unsigned int) *features;
            info->nports = usb_dev->nports;
            info->iocinfo_version = port->fstioc_info_ver;
            info->type = usb_dev->type;
            info->state = usb_dev->state;
            info->index = 0;
            info->smcFirmwareVersion = port->usb_card_info.uSoftwareVersion;

	    info->lineInterface = port->hwif;	
            info->proto = port->proto;
            info->valid = FSTVAL_ALL;
            info->internalClock = port->usb_config.clock;
            info->lineSpeed = port->usb_config.lineSpeed;
	    info->termination = port->usb_config.termination;
            info->ignoreCarrier = port->ignoreCarrier;
	    switch (port->usb_config.encoding)
	      {
	      case FSCMN_ENCODING_NRZ:
		info->coding = CODING_NRZ;
		break;
	      case FSCMN_ENCODING_NRZ_CLK_REC:
		info->coding = CODING_NRZ_CLK_REC;
		break;
	      case FSCMN_ENCODING_NRZI:
		info->coding = CODING_NRZI;
		break;
	      case FSCMN_ENCODING_FM0:
		info->coding = CODING_FM0;
		break;
	      case FSCMN_ENCODING_FM1:
		info->coding = CODING_FM1;
		break;
	      case FSCMN_ENCODING_MANCHESTER:
		info->coding = CODING_MANCHESTER;
		break;
	      case FSCMN_ENCODING_DMAN:
		info->coding = CODING_DIFF_MANCHESTER;
		break;
	      default:
		info->coding = CODING_NRZ;
		break;
	      }
            info->estLineSpeed = 0;
	    fst_issue_cmd(port, GETSIGNALS);
            info->v24IpSts = port->v24IpSts;
            info->v24OpSts = port->v24OpSts;
	    info->extendedClocking = 0;
            info->cableStatus = 0;
            info->cardMode = usb_dev->card_mode;
#if FST_DEBUG
            info->debug = fst_debug_mask;
#else
            info->debug = 0;
#endif
            info->transparentMode = port->usb_config.mode;
            info->transmitMSBFirst = port->transmitMSBFirst;
            info->receiveMSBFirst = port->receiveMSBFirst;

	    info->asyncAbility = 1;
            info->invertClock = port->usb_config.rxclock;
            info->numTxBuffers = port->usb_config.numTxBuffer; 
            info->numRxBuffers = port->usb_config.numRxBuffer;
            info->txBufferSize = port->usb_config.sizeTxBuffer;
            info->rxBufferSize = port->usb_config.sizeRxBuffer;
	    info->enableNRZIClocking = port->usb_config.enableNRZIClocking;
            /*
             * Extended Clocking
             */
	    if (port->usb_config.extendedClocking)
	      {
		info->extendedClocking = 0x80;
		if (port->usb_config.internalTxClock)
		  info->extendedClocking |= 0x01;
		if (port->usb_config.internalRxClock)
		  info->extendedClocking |= 0x02;
	      }
	    if (port->usb_config.extendedConfig & (EXT_CONFIG_TTT | EXT_CONFIG_RTT))
	      {
		info->extendedClocking |= 0x80;
		if (port->usb_config.extendedConfig & EXT_CONFIG_TTT)
		  info->extendedClocking |= 0x04;
		if (port->usb_config.extendedConfig & EXT_CONFIG_RTT)
		  info->extendedClocking |= 0x08;
	      }
	    if ((!port->usb_config.extendedClocking) && 
		(!port->usb_config.extendedConfig))
              {
                info->extendedClocking = 0;
              }
	    info->oneBitInsert = port->usb_config.extendedConfig & EXT_CONFIG_OIN;
	    fst_dbg(DBG_IOCTL, "Flex was configured for One Bit Insertion as %x\n", info->oneBitInsert);
            /*
             * Update the stats
             */
#ifdef IS_GEN_HDLC
            memcpy(&info->stats, hdlc_stats(port->dev), 
                   sizeof(struct fst_device_stats));
#else
            memcpy(&info->stats, &port->stats, 
                   sizeof(struct fst_device_stats));
#endif
            /*
             * Copy in the async config
             */
            info->async_conf.flow_control =  port->usb_config.flowControl;
            info->async_conf.stop_bits =  port->usb_config.stopbits;
            info->async_conf.parity = port->usb_config.parity;
            info->async_conf.word_length = port->usb_config.dataBits;
            info->async_conf.xon_char =  port->usb_config.xonchar;
            info->async_conf.xoff_char = port->usb_config.xoffchar;
#if 0
            printk("Async parameters returned as bits %d parity %d stop %d fc %d xon %d xof %d\n",
		   info->async_conf.word_length,
		   info->async_conf.parity,
		   info->async_conf.stop_bits,
		   info->async_conf.flow_control,
		   info->async_conf.xon_char,
		   info->async_conf.xoff_char);
#endif
            /*
             * and now return the size of the structure according
             * to the version set
             */
            switch (port->fstioc_info_ver)
              {
              case FST_VERSION_CURRENT:
                return fstioc_info_sz_current;
                
              case FST_VERSION_V1:
                return fstioc_info_sz_ver1;

              case FST_VERSION_V2:
                return fstioc_info_sz_ver2;
                
	      case FST_VERSION_V3:
                return fstioc_info_sz_ver3;
                
              default:
                return fstioc_info_sz_old;
              }
          }
        else
          {
            if ((retval != sizeof(port->usb_config)) || 
                (retval2 != sizeof(port->usb_card_info)) ||
                (retval3 != 4))
            {
              printk("Get Config error errno = %d %d %d %p\n", 
                      retval, retval2, retval3, info);
             }
            return 0;
          }
}

static int flex_set_config ( struct fst_card_info *usb_dev, 
                             struct fstioc_info *info)
{
        int retval;

	fst_dbg(DBG_IOCTL, "Setting port config\n");
	if (info->cardMode != usb_dev->card_mode)
	  {
	    char *mode = (char *) &usb_dev->ports[0]->scratch;
	    if (info->cardMode)
	      {
		mode[0] = 1;
		mode[1] = 0;
		mode[2] = 0;
		mode[3] = 0;
	      }
	    else
	      {
		mode[0] = 0;
		mode[1] = 0;
		mode[2] = 0;
		mode[3] = 0;
	      }
	    retval = usb_control_msg(usb_dev->udev, 
				     usb_sndctrlpipe(usb_dev->udev, 0),
				     USS_CMD_SET_IDENTIFY_MODE, 0x40, 0, 0, 
				     mode, CMD_BUF_LEN, 10000);
	    usb_dev->card_mode = info->cardMode;
	    if (retval==CMD_BUF_LEN) 
	      {
		fst_dbg(DBG_IOCTL, "Set identify mode response received\n");
	      }
	    else
	      {
		fst_dbg(DBG_ASS, "Set identify mode error errno = %d\n", retval);
	      }
	  }
	switch (info->lineInterface)
	  {
	  case V24:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_V24;
	    break;
	  case X21:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_X21;
	    break;
	  case V35:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_V35;
	    break;
	  case X21D:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_X21D;
	    break;
	  case NOCABLE:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_NO_CABLE;
	    break;
	  case RS530_449:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_RS530;
	    break;
	  case RS485:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_RS485;
	    break;
	  case RS485_FDX:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_RS485_FDX;
	    break;
	  default:
	    usb_dev->ports[0]->usb_config.iface = FSCMN_INTERFACE_V24;
	    break;
	  }
        usb_dev->ports[0]->hwif = info->lineInterface;
        usb_dev->ports[0]->proto = info->proto;
        usb_dev->ports[0]->usb_config.clock = info->internalClock;
        usb_dev->ports[0]->usb_config.lineSpeed = info->lineSpeed;
        usb_dev->ports[0]->ignoreCarrier = info->ignoreCarrier;
	    switch ( info->coding)
	      {
	      case CODING_NRZ:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_NRZ;
		break;
	      case CODING_NRZ_CLK_REC:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_NRZ_CLK_REC;
		break;
	      case CODING_NRZI:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_NRZI;
		break;
	      case CODING_FM0:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_FM0;
		break;
	      case CODING_FM1:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_FM1;
		break;
	      case CODING_MANCHESTER:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_MANCHESTER;
		break;
	      case CODING_DIFF_MANCHESTER:
		usb_dev->ports[0]->usb_config.encoding = FSCMN_ENCODING_DMAN;
		break;
	      default:
		usb_dev->ports[0]->usb_config.encoding =  FSCMN_ENCODING_NRZ;
		break;
	      }

        usb_dev->ports[0]->usb_config.termination = info->termination;
        usb_dev->ports[0]->mode = info->transparentMode;
        usb_dev->ports[0]->usb_config.mode = info->transparentMode;

        usb_dev->ports[0]->transmitMSBFirst = info->transmitMSBFirst;
        usb_dev->ports[0]->receiveMSBFirst = info->receiveMSBFirst;
        usb_dev->ports[0]->usb_config.bitOrder = 0;

        usb_dev->ports[0]->usb_config.start = info->txRxStart;
        if (usb_dev->ports[0]->fstioc_info_ver > FST_VERSION_V1)
          {
            usb_dev->ports[0]->usb_config.dataBits = info->async_conf.word_length;
            usb_dev->ports[0]->usb_config.parity = info->async_conf.parity;
            usb_dev->ports[0]->usb_config.stopbits = info->async_conf.stop_bits;
            usb_dev->ports[0]->usb_config.flowControl = info->async_conf.flow_control;
            usb_dev->ports[0]->usb_config.timeoutEnable = 0; /* Not used yet */
            usb_dev->ports[0]->usb_config.timeoutLength = 0; /* Not used yet */
#if 0
            printk("Async parameters set as bits %d parity %d stop %d fc %d xon %d xof %d\n",
		   info->async_conf.word_length,
		   info->async_conf.parity,
		   info->async_conf.stop_bits,
		   info->async_conf.flow_control,
		   info->async_conf.xon_char,
		   info->async_conf.xoff_char);
#endif
          }
#if FST_DEBUG
        fst_debug_mask = info->debug;
#endif
        usb_dev->ports[0]->usb_config.rxclock = info->invertClock;
        usb_dev->ports[0]->usb_config.numTxBuffer  = usb_dev->ports[0]->num_tx_buffers = info->numTxBuffers; 
        usb_dev->ports[0]->usb_config.numRxBuffer  = usb_dev->ports[0]->num_rx_buffers = info->numRxBuffers;
        usb_dev->ports[0]->usb_config.sizeTxBuffer = usb_dev->ports[0]->tx_buffer_size = info->txBufferSize;
        usb_dev->ports[0]->usb_config.sizeRxBuffer = usb_dev->ports[0]->rx_buffer_size = info->rxBufferSize;        usb_dev->ports[0]->usb_config.preserveSignals = usb_dev->ports[0]->v24OpSts & 0x100;
        usb_dev->ports[0]->usb_config.asyncReceiveMode = FSCMN_ASYNC_RX_CHAR;
	usb_dev->ports[0]->usb_config.enableNRZIClocking = info->enableNRZIClocking;
        /*
         * Extended Clocking feature
         */
	if ((info->extendedClocking) && 
	    ((info->lineInterface == V24) || 
	     (info->lineInterface == V35) || 
	     (info->lineInterface == RS530_449)))
	  {
            usb_dev->ports[0]->usb_config.clock = 0;
            usb_dev->ports[0]->usb_config.extendedClocking = 1;
            if (info->extendedClocking & 1)
              {
                usb_dev->ports[0]->usb_config.internalTxClock = 1;
              }
            else
              {
                usb_dev->ports[0]->usb_config.internalTxClock = 0;
              }
            if (info->extendedClocking & 2)
              {
                usb_dev->ports[0]->usb_config.internalRxClock = 1;
              }
            else
              {
                usb_dev->ports[0]->usb_config.internalRxClock = 0;
              }
          }
	else
	  {
	    /*
             * Reset extended clocking because wrong line type
             */
	    fst_dbg(DBG_IOCTL,"%s: Extended Clocking not allowed on interface type %d\n",
		   usb_dev->ports[0]->dev->name, info->lineInterface);
	    usb_dev->ports[0]->usb_config.extendedClocking = 0;
	    usb_dev->ports[0]->usb_config.internalTxClock = 0;
	    usb_dev->ports[0]->usb_config.internalRxClock = 0;
	  }
        /*
         * Extended Config feature
         */
	usb_dev->ports[0]->usb_config.extendedConfig = 0;
	if ((info->extendedClocking) && 
	    ((info->lineInterface == V24) || 
	     (info->lineInterface == V35) || 
	     (info->lineInterface == RS530_449)))
	  {
            usb_dev->ports[0]->usb_config.clock = 0;
            if (info->extendedClocking & 4)
              {
                usb_dev->ports[0]->usb_config.extendedConfig = EXT_CONFIG_TTT;
              }
            if (info->extendedClocking & 8)
              {
                usb_dev->ports[0]->usb_config.extendedConfig = EXT_CONFIG_RTT;
              }
           }
	else
	  {
	    /*
             * Reset extended config because wrong line type
             */
	    fst_dbg(DBG_IOCTL,"%s: Terminal Timing not allowed on interface type %d\n",
		   usb_dev->ports[0]->dev->name, info->lineInterface);
	    usb_dev->ports[0]->usb_config.extendedConfig = 0;
	  }
        fst_dbg(DBG_IOCTL, "Extended Clocking is %x \n", info->extendedClocking); 
	/*
	 * Check for OneBitInsertio
	 */
	if (info->oneBitInsert)
	  {
	    usb_dev->ports[0]->usb_config.extendedConfig |= EXT_CONFIG_OIN;
	  }
	fst_dbg(DBG_IOCTL, "extendedConfig is %x\n", usb_dev->ports[0]->usb_config.extendedConfig);
	flex_serconf_to_le(&usb_dev->ports[0]->usb_config);
	retval = usb_control_msg(usb_dev->udev, usb_sndctrlpipe(usb_dev->udev,
                                                                0),
				 USS_CMD_SET_CONFIG, 0x40, 0, 0, 
                                 &usb_dev->ports[0]->usb_config, 
                                 sizeof(FLEX_SERCONF), 10000);
	flex_serconf_to_cpu(&usb_dev->ports[0]->usb_config);				 
	if (retval>0) 
          {
            fst_dbg(DBG_IOCTL, "Set config response received\n");
            return 1;
          }
        else
          {
            printk("Set Config error errno = %d\n", retval);
            return 0;
          }
}

/*
 *      Initialise usb interface when detected.
 *      Returns 0 to indicate success, or errno otherwise.
 */
static int flex_probe ( struct usb_interface *interface, const struct usb_device_id *id )
{
        //static int firsttime_done = 0;
	struct fst_card_info *dev;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
        int retval=0;
        int instance=0;
	char dev_name[16];
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)
#ifdef SPECIFY_DEV_NAME
        struct net_device *netdev;
	struct usb_device *usbdev;
#endif
#endif
	
        /* set the defaults */
        strcpy(dev_name, FST_PORT_NAME);
        strcat(dev_name, "%d");  

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)
#ifdef SPECIFY_DEV_NAME	
	/* 
	 * Check this Flex's serial number against those in the fst_serials array.
	 * If a match is found, the associated dev_num (syncX) for that serial number
	 * is used in the alloc_netdev call instead of %d (which gives the next avaiable
	 * integer)
	 */	
	usbdev = interface_to_usbdev(interface);
	if((fst_num_serials %2) == 0)
	{  
	  for (i=0; i<fst_num_serials; i+=2) 
	  {
	    if((fst_serials[i+1] != NULL) && (strcmp(usbdev->serial, fst_serials[i+1]) == 0))
	    {
	      printk("Serial number %s matched to device %s\n", fst_serials[i+1],fst_serials[i]);
	      printk("Checking device %s against installed ones\n", fst_serials[i]);
   	      strcpy(dev_name, FST_PORT_NAME);
              strcat(dev_name, fst_serials[i]); 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	      netdev = dev_get_by_name(dev_name);
#else
	      netdev = dev_get_by_name(&init_net, dev_name);
#endif
	      if(netdev!=NULL)
	      {
	        printk("%s already in use so returning error\n", dev_name);
		dev_put(netdev);
		return EBUSY;
	      }
            }	    
    	  } /* for */
        }
#endif
#endif
	/*
	 * We are going to be clever and allow certain cards not to be
	 * configured.  An exclude list can be provided in /etc/modules.conf
	 */
	if (fst_excluded_cards != 0)
	  {
	    /*
	     * There are cards to exclude
	     *
	     */
	    for (i=0; i< fst_excluded_cards; i++)
	      {
		if (1 == fst_excluded_list[i])
		  {
		    fst_dbg(DBG_ASS, "Flex USB device %d not assigned\n", 1);
		    return -EBUSY;
		  }
	      }
	  }

        /* Allocate driver private data */
        dev = kmalloc ( sizeof ( struct fst_card_info ),  GFP_KERNEL);
        if (dev == NULL)
        {
          printk_err ("Flex USB device found but insufficient memory for"
                      " driver storage\n");
          return -ENOMEM;
        }
        fst_dbg(DBG_USB, "dev allocated at %p\n", dev);
        memset ( dev, 0, sizeof ( struct fst_card_info ));

	dev->family      = 0;
	dev->nports      = 1;
        dev->state       = FST_RUNNING;
	dev->udev        = usb_get_dev(interface_to_usbdev(interface));
	dev->interface   = interface;

        fst_dbg ( DBG_USB, "type %d nports %d irq %d\n", dev->type,
                        dev->nports, dev->irq );

	sema_init(&dev->limit_sem, WRITES_IN_FLIGHT);
        dev->tx_urb_cnt = MAX_URBS_OUTSTANDING;
	/* 
         * set up the endpoint information 
         * use only the first bulk-in and bulk-out endpoints 
         */
	iface_desc = interface->cur_altsetting;
	fst_dbg (DBG_USB, "The number of endpoints on this device are %d\n",
             iface_desc->desc.bNumEndpoints);
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
          endpoint = &iface_desc->endpoint[i].desc;
          /* 
           * Get Bulk end point addresses 
           */
          if (!dev->bulk_in_endpointAddr &&
              ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
               == USB_DIR_IN) &&
              ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
               == USB_ENDPOINT_XFER_BULK)) {
            /* we found a bulk in endpoint */
            buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
            dev->bulk_in_size = buffer_size;
            dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
            dev->bulk_in_buffer = NULL;
          }
          
          if (!dev->bulk_out_endpointAddr &&
              ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
               == USB_DIR_OUT) &&
              ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
               == USB_ENDPOINT_XFER_BULK)) {
            /* we found a bulk out endpoint */
            dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
          }
          /* 
           * Get Interrupt end point addresses 
           */
          if (!dev->int_in_endpointAddr &&
              ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
               == USB_DIR_IN) &&
              ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
               == USB_ENDPOINT_XFER_INT)) {
            /* we found a int in endpoint */
            buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
            dev->int_in_size = buffer_size;
            dev->int_in_endpointAddr = endpoint->bEndpointAddress;
            dev->int_in_buffer = NULL;
          }
          
          if (!dev->int_out_endpointAddr &&
              ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
               == USB_DIR_OUT) &&
              ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
               == USB_ENDPOINT_XFER_INT)) {
            /* we found a int out endpoint */
            dev->int_out_endpointAddr = endpoint->bEndpointAddress;
            dev->int_interval = endpoint->bInterval;
            if ((dev->int_interval != 2) && (dev->int_interval != 16))
              {
                printk_err("Interval value on int endpoint not as expected %d\n",
                    dev->int_interval);
                dev->int_interval = 4;
              }
            dev->int_interval = 4 /* Larger values can cause a problem */;
          }
	}
	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
          printk_err("Could not find both bulk-in and bulk-out endpoints");
          goto error;
	}
	if (!(dev->int_in_endpointAddr && dev->int_out_endpointAddr)) {
          printk_err("Could not find both int-in %d and int-out %d endpoints",
              dev->int_in_endpointAddr, dev->int_out_endpointAddr);
          //goto error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* allocate urbs */
	if ((retval = allocate_urbs_and_buffers(dev)))
	  goto error;

        /* Remainder of card setup */
	dev->fifo_complete = 1;
        fst_init_card ( dev, dev_name );
        flex_get_config(dev, dev->ports[0], NULL);
	if ((dev->ports[0]->usb_card_info.uMajorRev > FIRST_FLEX_V2_MAJOR_ID) ||
	    (dev->ports[0]->usb_card_info.uMinorRev >= FIRST_FLEX_V2_MINOR_ID))
	  {
	    if (dev->ports[0]->usb_card_info.uMajorRev >= FIRST_FLEX_V3_MAJOR_ID)
	      {
		fst_dbg(DBG_USB, "Flex Type Version 3 (minor rev = %d)\n", 
			dev->ports[0]->usb_card_info.uMinorRev);
		dev->type = FST_TYPE_FLEX3;
	      }
	      else
		{
		  fst_dbg(DBG_USB, "Flex Type Version 2 (minor rev = %d\n", 
			  dev->ports[0]->usb_card_info.uMinorRev);
		  dev->type = FST_TYPE_FLEX2;
		}
	  }
	else
	  {
	    fst_dbg(DBG_USB, "Flex Type Version 1 (minor rev = %d\n", 
		    dev->ports[0]->usb_card_info.uMinorRev);
	    dev->type = FST_TYPE_FLEX1;
	  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,17,0)
	strcpy((char* const)dev->ports[0]->dev->dev_addr, dev->ports[0]->usb_card_info.szSerialNo);
#else
	strcpy(dev->ports[0]->dev->dev_addr, dev->ports[0]->usb_card_info.szSerialNo);
#endif
        printk ( KERN_INFO "%s-%s: (%s) %s, %d ports\n",
                 dev->ports[0]->dev->name,
                 dev->ports[dev->nports-1]->dev->name,
                 dev->ports[0]->usb_card_info.szSerialNo,
                 type_strings[dev->type], dev->nports );

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &flex_class);
	if (retval) {
          /* something prevented us from registering this driver */
          printk_err("Not able to get a minor for this device.");
          usb_set_intfdata(interface, NULL);
          goto error;
	}
	/* let the user know what node this device is now attached to */
	instance = interface->minor % USB_MINOR_BASE;
	printk("USB flex device now attached to interface minor %d\n", instance);
        fst_cards_list[instance] = dev;
        dev->ports[0]->card = dev;
	dev->card_no = instance;
	fst_ncards++;    /* Record instance and bump it */
        fst_dbg(DBG_USB, "Setting the interface to start switch to 80MHz\n");
        i = usb_set_interface(dev->udev, 0, 0); 
        if(i)
          printk("Error in call to usb_set_interface %d\n", i);
        dev->actual_length = 0;
        dev->usb_trace_ptr = 0;
        post_a_read_int_urb(dev);       /* Start reading trace from card */
        post_a_read_bulk_urb(dev);      /* Post a read for data */
        return 0;                       /* Success */

                                        /* Failure. Release resources */
error:
        if (dev)
          {
            flex_delete(dev);
          }
        return retval;
}


/*
 *      Cleanup and close down a card
 */
static void flex_disconnect ( struct usb_interface *interface )
{
        struct fst_card_info *dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
	int minor = 0; //interface->minor % USB_MINOR_BASE;
#endif
        int i;

	dev = usb_get_intfdata(interface);
	minor = dev->ports[0]->minor_dev_no;
	fst_dbg(DBG_USB, "%s: Closing usb device for minor: %d\n", 
                  dev->ports[0]->dev->name, minor);
	/* give back our minor */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
        fst_dbg(DBG_INIT, "device_remove_file: 0x%p\n", dev->sync_cdev);   
	device_remove_file(dev->sync_cdev,
				 &dev_attr_serial);
        device_destroy(farsync_class, MKDEV(fst_major, minor));
#else
	class_device_remove_file(dev->sync_cdev,
				 &class_device_attr_serial);
        class_device_destroy(farsync_class, MKDEV(fst_major, minor));
#endif
#endif
	usb_deregister_dev(interface, &flex_class);
        for ( i = 0 ; i < dev->nports ; i++ )
        {
          if (dev->ports[i]->char_file)
            {
              fst_dbg(DBG_USB, "port is %p %p\n", 
                      dev->ports[i]->char_file->private_data, 
                      dev->ports[i]);
              if (dev->ports[i]->char_file->private_data == dev->ports[i])
                {
                  fst_dbg(DBG_USB, "Zapped files pointer to us\n");
                  dev->ports[i]->char_file->private_data = NULL;
                }
            }
#ifdef FST_TTY_CHAR
	  fst_tty_area[minor].state = FST_TTY_ST_DISC;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
	  fst_tty_area[minor].state = FST_TTY_ST_DISC;
	  printk("Calling tty unregister device for port %d\n", minor);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
	  tty_port_destroy(&fst_tty_area[minor].tty_port);
#endif
	  tty_unregister_device(serial_drv, minor);
#endif
#endif
          if (dev->ports[i]->fifo_rxdata)
            {
              /*
               * Deallocate the rx fifo
               */
              u16 used_space;
              used_space = (u16) ((dev->ports[i]->fifo_rxdata->write_idx - 
                                   dev->ports[i]->fifo_rxdata->read_idx) 
                                  % dev->ports[i]->fifo_rxdata->fifo_length);
              fst_dbg(DBG_ASS, "%s: Releasing Fifo memory at %p\n", 
                      dev->ports[i]->dev->name,
                      dev->ports[i]->fifo_rxdata);
              fst_dbg(DBG_ASS, "%s: %d bytes left in fifo\n", 
                      dev->ports[i]->dev->name, used_space);
              kfree(dev->ports[i]->fifo_rxdata);
              dev->ports[i]->fifo_rxdata = NULL;
            }
          unregister_netdev(dev->ports[i]->dev);
	  fst_ports_list[dev->ports[i]->minor_dev_no] = NULL;
          while (dev->ports[i]->txqe != dev->ports[i]->txqs)
            {
              printk("Purging tx queue %d %d\n", dev->ports[i]->txqe,
                     dev->ports[i]->txqs);
              printk("%s: Closing port with data in tx queue %p\n", 
                     dev->ports[i]->dev->name, 
                     dev->ports[i]->txq[dev->ports[i]->txqs].frame);
              dev_kfree_skb(dev->ports[i]->txq[dev->ports[i]->txqs].frame);
              dev->ports[i]->txqs++;
              if (dev->ports[i]->txqs == FST_TXQ_DEPTH)
                {
                  dev->ports[i]->txqs = 0;
                }
            }
	  fst_dbg(DBG_OPEN, "Any char opens on the device %p %d\n", dev->ports[i],
		 dev->ports[i]->char_opens);
          if (!dev->ports[i]->char_opens)
            {
	      fst_dbg(DBG_OPEN, "I can free the net device\n");
	      free_netdev(dev->ports[i]->dev);
	    }
	  else
	    {
	      fst_dbg(DBG_ASS, "*** Delaying freeing of the net device %d at %p\n", 
		     dev->ports[i]->minor_dev_no,
		     dev->ports[i]->dev);
	      fst_net_list[dev->ports[i]->minor_dev_no] = dev->ports[i]->dev;
	    }
        }
        fst_cards_list[dev->card_no] = NULL;
        fst_ncards--;
	/* decrement our usage count */
	deallocate_urbs_and_buffers(dev);
	usb_set_intfdata(interface, NULL);
	flex_delete(dev);
}

static struct usb_driver fst_driver = {
	.name =		"fsflex",
	.probe =	flex_probe,
	.disconnect =	flex_disconnect,
	.id_table =	flex_table,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
static int farsync_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, fst_proc_info, NULL);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
static const struct proc_ops farsync_proc_fops = {
	.proc_open = farsync_proc_open,
	.proc_read = seq_read,
	.proc_lseek = seq_lseek,
	.proc_release = single_release,	
};
#else
static const struct file_operations farsync_proc_fops = {
	.open = farsync_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
#endif
#endif

static int __init
fst_init(void)
{
        int i;

        printk ( KERN_INFO "\n");
        printk ( KERN_INFO "Flex USB OEM driver " FST_USER_VERSION
                 "-" FST_PATCH_LEVEL "-" FST_PLATFORM FST_ADDITIONAL
		 " \n(C) 2001-2023 FarSite Communications Ltd.\n");
	fst_dbg(DBG_ASS, "The value of debug mask is %x\n", 
                fst_debug_mask);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	INIT_WORK(&thresh_work_q, fst_process_threshold_work_q);
	INIT_WORK(&port_stats_work_q, fst_process_port_stats_work_q);
#else
	INIT_WORK(&thresh_work_q, fst_process_threshold_work_q, NULL);
	INIT_WORK(&port_stats_work_q, fst_process_port_stats_work_q, NULL);
#endif
        for (i=0; i<FST_MAX_CARDS; i++)
          fst_cards_list[i] = NULL;
	for (i=0; i<FST_MAX_CARDS*FST_MAX_PORTS; i++)
	  {
	    fst_ports_list[i] = NULL;
            fst_net_list[i] = NULL;
	  }
        spin_lock_init ( &fst_work_q_lock );
        /*
         * Create the /proc entry for /proc/fsflex
         */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	proc_create("fsflex", 0, NULL, &farsync_proc_fops);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
        create_proc_read_entry ("fsflex", 0, NULL, fst_proc_info, NULL);
#else
        create_proc_info_entry ("fsflex", 0, NULL, fst_proc_info);
#endif
#endif
        /* Register the char driver */
#ifdef FST_TTY_CHAR
        fst_tty_init();
        fst_major = FST_TTY_MAJOR;
#else
        fst_sync_init();
        i = register_chrdev ( fst_major, "fsflex", &fst_mem_fops );
        if ( i < 0 ) {
          printk_err ("Unable to register major device %d. Aborting\n",
                      fst_major );
          return i;
        }
        /* If we were using dynamic allocation record the result */
        if ( fst_major == 0 ) {
          fst_major = i;
        }
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
        fst_dbg(DBG_ASS, "Creating the device class for fsflex\n");
        farsync_class = class_create(THIS_MODULE, "fsflex");
#endif
	i = usb_register ( &fst_driver );
	if (i)
          {
		printk_err("usb_register failed. Error number %d", i);
                return i;
          }
        printk_info("fst_iocinfo_version set to %d\n", fst_iocinfo_version);
	printk_info("fst_allow_ioctl set to %d\n", fst_allow_ioctl);
        return 0;
}

static void __exit
fst_cleanup_module(void)
{
        usb_deregister ( &fst_driver );
        /* Remove the char device driver entry */
#ifdef FST_TTY_CHAR
        fst_tty_uninit();
#else
        (void) unregister_chrdev ( fst_major, "fsflex" );
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
	class_destroy(farsync_class);
#endif
        remove_proc_entry ("fsflex", NULL);
	printk_info("Unloading module fsflex\n");
}

/* 
* Converts the FLEX_SERCONF structure from native to Little Endian 
*/
static void flex_serconf_to_le(FLEX_SERCONF *usb_config)
{
    cpu_to_le32s(&usb_config->uVersion);
    cpu_to_le32s(&usb_config->iface);
    cpu_to_le32s(&usb_config->termination);
    cpu_to_le32s(&usb_config->lineSpeed);
    cpu_to_le32s(&usb_config->clock);
    cpu_to_le32s(&usb_config->encoding);
    cpu_to_le32s(&usb_config->mode);
    cpu_to_le32s(&usb_config->syncChar);
    cpu_to_le32s(&usb_config->bitOrder);
    cpu_to_le32s(&usb_config->rxclock);
    cpu_to_le32s(&usb_config->start);
    cpu_to_le32s(&usb_config->numTxBuffer);
    cpu_to_le32s(&usb_config->numRxBuffer);
    cpu_to_le32s(&usb_config->sizeTxBuffer);
    cpu_to_le32s(&usb_config->sizeRxBuffer);
    cpu_to_le32s(&usb_config->dataBits);
    cpu_to_le32s(&usb_config->parity);
    cpu_to_le32s(&usb_config->stopbits);
    cpu_to_le32s(&usb_config->flowControl);
    cpu_to_le32s(&usb_config->timeoutEnable);
    cpu_to_le32s(&usb_config->timeoutLength);
    cpu_to_le32s(&usb_config->monitorMode);
    cpu_to_le32s(&usb_config->extendedClocking);
    cpu_to_le32s(&usb_config->internalTxClock);
    cpu_to_le32s(&usb_config->internalRxClock);
    cpu_to_le32s(&usb_config->enableNRZIClocking);
    cpu_to_le32s(&usb_config->xonchar);
    cpu_to_le32s(&usb_config->xoffchar);
    cpu_to_le32s(&usb_config->preserveSignals);
    cpu_to_le32s(&usb_config->asyncReceiveMode);
    cpu_to_le32s(&usb_config->crcReset);
    cpu_to_le32s(&usb_config->extendedConfig);
}

/*
* Converts the FLEX_SERCONF struct from Little Endian to native
*/
static void flex_serconf_to_cpu(FLEX_SERCONF *usb_config)
{
    le32_to_cpus(&usb_config->uVersion);
    le32_to_cpus(&usb_config->iface);
    le32_to_cpus(&usb_config->termination);
    le32_to_cpus(&usb_config->lineSpeed);
    le32_to_cpus(&usb_config->clock);
    le32_to_cpus(&usb_config->encoding);
    le32_to_cpus(&usb_config->mode);
    le32_to_cpus(&usb_config->syncChar);
    le32_to_cpus(&usb_config->bitOrder);
    le32_to_cpus(&usb_config->rxclock);
    le32_to_cpus(&usb_config->start);
    le32_to_cpus(&usb_config->numTxBuffer);
    le32_to_cpus(&usb_config->numRxBuffer);
    le32_to_cpus(&usb_config->sizeTxBuffer);
    le32_to_cpus(&usb_config->sizeRxBuffer);
    le32_to_cpus(&usb_config->dataBits);
    le32_to_cpus(&usb_config->parity);
    le32_to_cpus(&usb_config->stopbits);
    le32_to_cpus(&usb_config->flowControl);
    le32_to_cpus(&usb_config->timeoutEnable);
    le32_to_cpus(&usb_config->timeoutLength);
    le32_to_cpus(&usb_config->monitorMode);
    le32_to_cpus(&usb_config->extendedClocking);
    le32_to_cpus(&usb_config->internalTxClock);
    le32_to_cpus(&usb_config->internalRxClock);
    le32_to_cpus(&usb_config->enableNRZIClocking);
    le32_to_cpus(&usb_config->xonchar);
    le32_to_cpus(&usb_config->xoffchar);
    le32_to_cpus(&usb_config->preserveSignals);
    le32_to_cpus(&usb_config->asyncReceiveMode);
    le32_to_cpus(&usb_config->crcReset);
    le32_to_cpus(&usb_config->extendedConfig);
}

/*
* Convert the Flex Interface Record definition into Little Endian, ready for USB
*/
static void flex_irex_to_le(IREX *stats)
{
    int i;
    cpu_to_le32s(&stats->InterfaceRecord.RxFrameCount);
    cpu_to_le32s(&stats->InterfaceRecord.TxMaxFrSizeNow);
    cpu_to_le32s(&stats->InterfaceRecord.StatusCount);
    for (i = 0; i < SA_Max_Stat; i++)
    {
        cpu_to_le32s(&stats->InterfaceRecord.StatusArray[i]);
    }
    cpu_to_le32s(&stats->StatusCount);
    cpu_to_le32s(&stats->OpenedCount);
    cpu_to_le32s(&stats->TxRequestCount);
    cpu_to_le32s(&stats->TxCompleteCount);
    cpu_to_le32s(&stats->RxPostedCount);
    cpu_to_le32s(&stats->RxCompleteCount);
}

/*
* Converts the IREX object from Little Endian to native byte order
*/
static void flex_irex_to_cpu(IREX *stats)
{
    int i;
    le32_to_cpus(&stats->InterfaceRecord.RxFrameCount);
    le32_to_cpus(&stats->InterfaceRecord.TxMaxFrSizeNow);
    le32_to_cpus(&stats->InterfaceRecord.StatusCount);
    for (i = 0; i < SA_Max_Stat; i++)
    {
        le32_to_cpus(&stats->InterfaceRecord.StatusArray[i]);
    }
    le32_to_cpus(&stats->StatusCount);
    le32_to_cpus(&stats->OpenedCount);
    le32_to_cpus(&stats->TxRequestCount);
    le32_to_cpus(&stats->TxCompleteCount);
    le32_to_cpus(&stats->RxPostedCount);
    le32_to_cpus(&stats->RxCompleteCount);
}

/*
* Converts the struct fstioc_req from native byte order to Little Endian
*/
static void flex_fstioc_req_to_le(struct fstioc_req *sys_request)
{
    cpu_to_le16s(&sys_request->msg_type);
    cpu_to_le16s(&sys_request->msg_len);
    cpu_to_le16s(&sys_request->ret_code);
    cpu_to_le16s(&sys_request->i_reg_idx);
    cpu_to_le16s(&sys_request->value);
}

/*
* Converts the struct fstioc_req from native to Little Endian byte order
*/
static void flex_fstioc_req_to_cpu(struct fstioc_req *sys_request)
{
    le16_to_cpus(&sys_request->msg_type);
    le16_to_cpus(&sys_request->msg_len);
    le16_to_cpus(&sys_request->ret_code);
    le16_to_cpus(&sys_request->i_reg_idx);
    le16_to_cpus(&sys_request->value);
}

/*
* Converts flex_card_info from native to Little Endian (USB format)
*/
static void flex_card_info_to_le(FLEX_CARD_INFO *usb_card_info)
{
    unsigned short i;
    cpu_to_le32s(&usb_card_info->uVersion);
    cpu_to_le32s(&usb_card_info->uMajorRev);
    cpu_to_le32s(&usb_card_info->uMinorRev);
    cpu_to_le32s(&usb_card_info->uBuildState);
    cpu_to_le32s(&usb_card_info->uCPUSpeed);
    cpu_to_le32s(&usb_card_info->uMode);
    cpu_to_le32s(&usb_card_info->uSoftwareVersion);
    cpu_to_le32s(&usb_card_info->uArmVersion);
    cpu_to_le32s(&usb_card_info->uflashManfId);
    cpu_to_le32s(&usb_card_info->uflashDevId);
    cpu_to_le32s(&usb_card_info->uSeroccoVersion);
    for (i = 0; i < 17; i++)
    {
        cpu_to_le32s(&usb_card_info->spare[i]);
    }
}

/*
* Converts the flex_card_info back from USB's Little Endian format into native
*/
static void flex_card_info_to_cpu(FLEX_CARD_INFO *usb_card_info)
{
    unsigned short i;
    le32_to_cpus(&usb_card_info->uVersion);
    le32_to_cpus(&usb_card_info->uMajorRev);
    le32_to_cpus(&usb_card_info->uMinorRev);
    le32_to_cpus(&usb_card_info->uBuildState);
    le32_to_cpus(&usb_card_info->uCPUSpeed);
    le32_to_cpus(&usb_card_info->uMode);
    le32_to_cpus(&usb_card_info->uSoftwareVersion);
    le32_to_cpus(&usb_card_info->uArmVersion);
    le32_to_cpus(&usb_card_info->uflashManfId);
    le32_to_cpus(&usb_card_info->uflashDevId);
    le32_to_cpus(&usb_card_info->uSeroccoVersion);
    for (i = 0; i < 17; i++)
    {
        le32_to_cpus(&usb_card_info->spare[i]);
    }
}

/*
* Convert the header back from USB's Little Endian format to native
*/
static void flex_txrx_preamble_to_cpu(FSCMN_TXRX_PREAMBLE *header)
{
    le32_to_cpus(&header->timestamp);
    le16_to_cpus(&header->length);
}

module_init ( fst_init );
module_exit ( fst_cleanup_module );

/* End $Id: fsflex-2.6-sync.c 4596 2019-06-04 08:49:07Z kevinc $ */
